Package: cat2cat
Title: Handling an Inconsistently Coded Categorical Variable in a
        Longitudinal Dataset
Version: 0.6.1
Authors@R: c(
    person("Maciej", "Nasinski", email = "nasinski.maciej@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-5236-8040"))
    )
Maintainer: Maciej Nasinski <nasinski.maciej@gmail.com>
Description: 
    Unifying an inconsistently coded categorical variable between two different time points in accordance with a mapping table. 
    The main rule is to replicate the observation if it could be assigned to a few categories. 
    Then using frequencies or statistical methods to approximate the probabilities of being assigned to each of them. 
    This procedure was invented and implemented in the paper by 'Nasinski', 'Majchrowska', and 'Broniatowska' (2020) <doi:10.24425/cejeme.2020.134747>.
Depends: R (>= 3.6)
License: GPL (>= 2) | file LICENSE
URL: https://github.com/Polkas/cat2cat,
        https://polkas.github.io/cat2cat/
BugReports: https://github.com/Polkas/cat2cat/issues
Encoding: UTF-8
Imports: MASS
Suggests: caret, dplyr, e1071, fixest, forcats, knitr, magrittr,
        randomForest, rmarkdown, spelling, testthat (>= 3.0.0), tidyr
LazyData: true
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
Language: en-US
NeedsCompilation: no
Packaged: 2026-05-17 20:59:10 UTC; maciejnasinski
Author: Maciej Nasinski [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5236-8040>)
Repository: CRAN
Date/Publication: 2026-05-17 21:40:02 UTC
Built: R 4.6.0; ; 2026-05-17 22:13:00 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,cat2cat_ml_run)
export(cat2cat)
export(cat2cat_agg)
export(cat2cat_ml_run)
export(cat_apply_freq)
export(cross_c2c)
export(dummy_c2c)
export(get_freqs)
export(get_mappings)
export(plot_c2c)
export(prune_c2c)
export(summary_c2c)
importFrom(MASS,lda)
importFrom(graphics,barplot)
importFrom(graphics,hist)
importFrom(graphics,par)
importFrom(stats,complete.cases)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,pt)
importFrom(stats,setNames)
importFrom(utils,head)

# cat2cat 

<a href='https://github.com/polkas/py-cat2cat'>
<img src='https://raw.githubusercontent.com/Polkas/cat2cat/master/man/figures/cat2cat_logo.png'  style="display:block;margin-left:auto;margin-right:auto;width:200px;" width="200px" alt="cat2cat logo"/>
</a>

<hr>

<div>
<a href="https://github.com/polkas/py-cat2cat/actions">
<img src="https://github.com/polkas/py-cat2cat/workflows/ci/badge.svg" alt="Build Status">
</a>
<a href="https://codecov.io/gh/Polkas/py-cat2cat">
<img src="https://codecov.io/gh/Polkas/py-cat2cat/branch/main/graph/badge.svg" alt="codecov">
</a>
<a href="https://pypi.org/project/cat2cat/">
<img src="https://img.shields.io/pypi/v/cat2cat.svg" alt="pypi">
</a>
<div>

<br>

### Unifying an inconsistently coded categorical variable in a panel/longtitudal dataset

There is offered the cat2cat procedure to map a categorical variable according to a mapping (transition) table between two different time points. The mapping (transition) table should to have a candidate for each category from the targeted for an update period. The main rule is to replicate the observation if it could be assigned to a few categories, then using simple frequencies or statistical methods to approximate probabilities of being assigned to each of them.

This algorithm was invented and implemented in the paper by (Nasinski, Majchrowska and Broniatowska (2020) doi:10.24425/cejeme.2020.134747).

## Installation

```bash
$ pip install cat2cat
```

## Usage

For more examples and descriptions please vist [**the example notebook**](https://py-cat2cat.readthedocs.io/en/latest/example.html)

### load example data

```python
# cat2cat datasets
from cat2cat.datasets import load_trans, load_occup
trans = load_trans()
occup = load_occup()
```

### Low-level functions

```python
from cat2cat.mappings import get_mappings, get_freqs, cat_apply_freq

# convert the mapping table to two association lists
mappings = get_mappings(trans)
# get a variable levels freqencies
codes_new = occup.code[occup.year == 2010].values
freqs = get_freqs(codes_new)
# apply the frequencies to the (one) association list
mapp_new_p = cat_apply_freq(mappings["to_new"], freqs)

# mappings for a specific category
mappings["to_new"]['3481']
# probability mappings for a specific category
mapp_new_p['3481']
```

### cat2cat function

```python
from cat2cat import cat2cat
from cat2cat.dataclass import cat2cat_data, cat2cat_mappings, cat2cat_ml

from pandas import concat

# split the panel by the time variale
# here only two periods
o_old = occup.loc[occup.year == 2008, :].copy()
o_new = occup.loc[occup.year == 2010, :].copy()

# dataclasses, core arguments for the cat2cat function
data = cat2cat_data(
    old = o_old, 
    new = o_new,
    cat_var_old = "code", 
    cat_var_new = "code", 
    time_var = "year"
)
mappings = cat2cat_mappings(trans = trans, direction = "backward")

# apply the cat2cat procedure
c2c = cat2cat(data = data, mappings = mappings)
# pandas.concat used to bind per period datasets
data_final = concat([c2c["old"], c2c["new"]])
```

## Contributing

Interested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.

## License

`cat2cat` was created by Maciej Nasinski. It is licensed under the terms of the MIT license.

## Credits

`cat2cat` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE, message = FALSE, warning = FALSE,
  collapse = TRUE, comment = "#>"
)

## ----load-data----------------------------------------------------------------
library(cat2cat)
library(dplyr)
library(tidyr)
library(fixest)

data(occup, package = "cat2cat")
data(occup_panel, package = "cat2cat")
data(trans, package = "cat2cat")
data(verticals, package = "cat2cat")
data(verticals2, package = "cat2cat")

occup_2006 <- occup[occup$year == 2006, ]
occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]

## ----ml-minimal, eval=FALSE---------------------------------------------------
# ml_setup <- list(
#   data = bind_rows(occup_2010, occup_2012),
#   cat_var = "code",
#   method = c("knn", "rf", "lda"),
#   features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#   args = list(k = 10, ntree = 50),
#   on_fail = "freq",
#   fail_warn = TRUE
# )
# 
# result_ml <- cat2cat(
#   data = list(old = occup_2008, new = occup_2010,
#               cat_var = "code", time_var = "year"),
#   mappings = list(trans = trans, direction = "backward"),
#   ml = ml_setup
# )

## ----ml-validate, eval=FALSE--------------------------------------------------
# cv <- cat2cat_ml_run(
#   mappings = list(trans = trans, direction = "backward"),
#   ml = ml_setup
# )
# print(cv)

## ----baseline-only-validate, eval=FALSE---------------------------------------
# ml_baseline <- list(
#   data = bind_rows(occup_2010, occup_2012),
#   cat_var = "code",
#   method = character(0),
#   features = character(0)
# )
# 
# cv_baseline <- cat2cat_ml_run(
#   mappings = list(trans = trans, direction = "backward"),
#   ml = ml_baseline
# )
# print(cv_baseline)

## ----trans-stability----------------------------------------------------------
max_digits <- max(nchar(as.character(trans[[1]])), nchar(as.character(trans[[2]])))

stability <- sapply(1:max_digits, function(d) {
  old_trunc <- substr(as.character(trans[[1]]), 1, d)
  new_trunc <- substr(as.character(trans[[2]]), 1, d)
  mean(old_trunc != new_trunc) * 100
})

data.frame(
  digits = 1:max_digits,
  pct_changed = round(stability, 1)
)

## ----truncation-diagnostic----------------------------------------------------
occup_2008_trunc <- occup_2008
occup_2008_trunc$code <- substr(occup_2008_trunc$code, 1, 4)
occup_2010_trunc <- occup_2010
occup_2010_trunc$code <- substr(occup_2010_trunc$code, 1, 4)
trans_trunc <- unique(data.frame(
  old = substr(trans$old, 1, 4),
  new = substr(trans$new, 1, 4)
))

back_full <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)
back_trunc <- cat2cat(
  data = list(old = occup_2008_trunc, new = occup_2010_trunc,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans_trunc, direction = "backward")
)

fwd_full <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "forward")
)
fwd_trunc <- cat2cat(
  data = list(old = occup_2008_trunc, new = occup_2010_trunc,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans_trunc, direction = "forward")
)

data.frame(
  mapping = c("backward full (4->6)", "backward trunc (4->4)",
              "forward full (6->4)", "forward trunc (4->4)"),
  mean_rep = c(mean(back_full$old$rep_c2c), mean(back_trunc$old$rep_c2c),
               mean(fwd_full$new$rep_c2c), mean(fwd_trunc$new$rep_c2c))
)

## ----backward-chain-----------------------------------------------------------
step1 <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)

step2 <- cat2cat(
  data = list(old = occup_2006, new = step1$old,
              cat_var_old = "code", cat_var_new = "g_new_c2c",
              time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)

harmonised_back <- bind_rows(
  step2$old,
  step1$old,
  step1$new,
  dummy_c2c(occup_2012, "code")
)

## ----validate-backward-counts-------------------------------------------------
harmonised_back %>%
  group_by(year) %>%
  summarise(weighted_n = round(sum(wei_freq_c2c)), .groups = "drop") %>%
  left_join(count(occup, year), by = "year")

## ----forward-chain------------------------------------------------------------
trans_fwd <- rbind(
  trans,
  data.frame(old = "no_cat",
             new = setdiff(c(occup_2010$code, occup_2012$code), trans$new))
)

fwd1 <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans_fwd, direction = "forward")
)

fwd2 <- cat2cat(
  data = list(old = fwd1$new, new = occup_2012,
              cat_var_old = "g_new_c2c", cat_var_new = "code",
              time_var = "year"),
  mappings = list(trans = trans_fwd, direction = "forward")
)

harmonised_fwd <- bind_rows(
  dummy_c2c(occup_2006, "code"),
  fwd1$old,
  fwd1$new,
  fwd2$new
)

## ----ml-chain, eval=FALSE-----------------------------------------------------
# step1_ml <- cat2cat(
#   data = list(old = occup_2008, new = occup_2010,
#               cat_var = "code", time_var = "year"),
#   mappings = list(trans = trans, direction = "backward"),
#   ml = ml_setup
# )
# 
# step2_ml <- cat2cat(
#   data = list(old = occup_2006, new = step1_ml$old,
#               cat_var_old = "code", cat_var_new = "g_new_c2c",
#               time_var = "year"),
#   mappings = list(trans = trans, direction = "backward"),
#   ml = ml_setup
# )

## ----panel-load---------------------------------------------------------------
panel_old <- occup_panel[occup_panel$quarter == "2009Q4", ]
panel_new <- occup_panel[occup_panel$quarter == "2010Q1", ]
shared_ids <- intersect(panel_old$panel_id, panel_new$panel_id)
length(shared_ids)

## ----idvar-mapping------------------------------------------------------------
result_id <- cat2cat(
  data = list(
    old = panel_old,
    new = panel_new,
    id_var = "panel_id",
    cat_var = "code",
    time_var = "quarter"
  ),
  mappings = list(trans = trans, direction = "backward")
)

## ----idvar-inspect------------------------------------------------------------
table(result_id$old$rep_c2c)
sum(result_id$old$wei_freq_c2c)
nrow(panel_old)

## ----idvar-compare------------------------------------------------------------
result_no_id <- cat2cat(
  data = list(
    old = panel_old,
    new = panel_new,
    cat_var = "code",
    time_var = "quarter"
  ),
  mappings = list(trans = trans, direction = "backward")
)

cat("WITH id_var average replication:", round(mean(result_id$old$rep_c2c), 2), "\n")
cat("WITHOUT id_var average replication:", round(mean(result_no_id$old$rep_c2c), 2), "\n")

## ----agg-data-----------------------------------------------------------------
agg_old <- verticals[verticals$v_date == "2020-04-01", ]
agg_new <- verticals[verticals$v_date == "2020-05-01", ]

## ----agg-mapping--------------------------------------------------------------
agg <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  Automotive %<% c(Automotive1, Automotive2),
  c(Kids1, Kids2) %>% c(Kids),
  Home %>% c(Home, Supermarket)
)

## ----agg-inspect--------------------------------------------------------------
agg$old[agg$old$vertical %in% c("Automotive1", "Automotive2"), ]
agg$new[agg$new$vertical %in% c("Kids1", "Kids2"), ]

## ----hierarchical-codes-------------------------------------------------------
trans_2digit <- data.frame(
  old = substr(trans$old, 1, 2),
  new = substr(trans$new, 1, 2)
)
trans_2digit <- unique(trans_2digit)

cat("2-digit mapping rows:", nrow(trans_2digit),
    "vs full mapping rows:", nrow(trans))

## ----build-panel--------------------------------------------------------------
cat2cat_data_back <- bind_rows(
  step2$old,
  step1$old,
  step1$new,
  dummy_c2c(occup_2012, "code")
)

## ----regression-neutral-------------------------------------------------------
lms_orig <- lm(
  I(log(salary)) ~ age + sex + factor(edu) + parttime + exp,
  data = occup,
  weights = multiplier
)

lms_harmonised <- lm(
  I(log(salary)) ~ age + sex + factor(edu) + parttime + exp,
  data = cat2cat_data_back,
  weights = multiplier * wei_freq_c2c
)

summary_c2c(lms_harmonised, df_old = nrow(occup))

## ----regression-fe------------------------------------------------------------
harmonised_fe <- cat2cat_data_back %>%
  prune_c2c(method = "nonzero") %>%
  mutate(orig_obs_id = interaction(year, index_c2c, drop = TRUE, lex.order = TRUE)) %>%
  filter(!is.na(g_new_c2c), !is.na(salary), salary > 0)

fe_model_cluster <- feols(
  log(salary) ~ age + sex + factor(edu) + parttime + exp | g_new_c2c + year,
  data = harmonised_fe,
  weights = ~multiplier * wei_freq_c2c,
  cluster = ~orig_obs_id
)
summary(fe_model_cluster)

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE, message = FALSE, warning = FALSE,
  collapse = TRUE, comment = "#>"
)

## ----load-data, echo=FALSE----------------------------------------------------
library(cat2cat)
data(trans, package = "cat2cat")

## ----show-problem-------------------------------------------------------------
# Old occupation code "1111" became three different codes in the new system
trans[trans$old == "1111", ]

## ----use-case-agg-------------------------------------------------------------
library(dplyr)
data(verticals, package = "cat2cat")

agg_old <- verticals[verticals$v_date == "2020-04-01", ]
agg_new <- verticals[verticals$v_date == "2020-05-01", ]

agg_result <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  # Backward mapping: old Automotive split into Automotive1 + Automotive2
  Automotive %<% c(Automotive1, Automotive2),
  # Forward mapping: Kids1 + Kids2 merged into Kids  
  c(Kids1, Kids2) %>% c(Kids)
)

agg_result$old[c("vertical", "prop_c2c", "counts")]

## ----direction-comparison-----------------------------------------------------
# Setup for comparison
occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]

# Backward: old period gets replicated onto new codes
backward <- cat2cat(
  data = list(old = occup_2008, new = occup_2010, 
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)

# Forward: new period gets replicated onto old codes
forward <- cat2cat(
  data = list(old = occup_2008, new = occup_2010, 
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "forward")
)

# Which period gets replicated depends on direction
cat("Backward: old period replicated from", nrow(occup_2008), "to", nrow(backward$old), "rows\n")
cat("Forward: new period replicated from", nrow(occup_2010), "to", nrow(forward$new), "rows")

## ----quick-example------------------------------------------------------------
data(occup, package = "cat2cat")

occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]

result_back <- cat2cat(
  data = list(
    old = occup_2008,
    new = occup_2010,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward")
)

## ----show-replication---------------------------------------------------------
# A replicated observation (rep_c2c > 1 means replicated)
result_back$old[result_back$old$rep_c2c > 1, ][1:3,
                c("code", "g_new_c2c", "wei_freq_c2c", "rep_c2c")]

## ----forward-example----------------------------------------------------------
result_forward <- cat2cat(
  data = list(
    old = occup_2008,
    new = occup_2010,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "forward")
)

result_forward$new[result_forward$new$rep_c2c > 1, ][1:3,
                   c("code", "g_new_c2c", "wei_freq_c2c", "rep_c2c")]

## ----naive-vs-freq------------------------------------------------------------
# Compare weights for a replicated observation
result_back$old[result_back$old$rep_c2c > 1, ][1:3,
                c("g_new_c2c", "wei_naive_c2c", "wei_freq_c2c")]

# Same analysis with naive weights (robustness check)
c(freq_mean = weighted.mean(result_back$old$salary, result_back$old$wei_freq_c2c),
  naive_mean = weighted.mean(result_back$old$salary, result_back$old$wei_naive_c2c))

## ----pooled-regression--------------------------------------------------------
harmonised_two_period <- dplyr::bind_rows(result_back$old, result_back$new)

harmonised_two_period <- harmonised_two_period %>%
  dplyr::group_by(g_new_c2c) %>%
  dplyr::mutate(avg_age_g_new_c2c = mean(age, na.rm = TRUE)) %>%
  dplyr::ungroup()

pooled_model <- lm(
  log(salary) ~ factor(year) + age + exp + avg_age_g_new_c2c,
  data = harmonised_two_period,
  weights = multiplier * wei_freq_c2c
)

pooled_summary <- summary_c2c(
  pooled_model,
  df_old = nrow(occup_2008) + nrow(occup_2010) - length(coef(pooled_model))
)

pooled_summary[c("factor(year)2010", "age", "avg_age_g_new_c2c"),
               c("Estimate", "std.error_c", "p.value_c")]

## ----plot, fig.width=7, fig.height=3.5, fig.alt="Diagnostic plot displaying weight histograms and replication statistics for mapped observations"----
plot_c2c(result_back$old, type = "both")

## ----hierarchical-codes-------------------------------------------------------
head(trans, 5)

# Build a 3-digit mapping from the full codes
trans_3digit <- data.frame(
  old = substr(trans$old, 1, 3),
  new = substr(trans$new, 1, 3)
)
trans_3digit <- unique(trans_3digit)
cat("3-digit mapping rows:", nrow(trans_3digit),
    "vs full mapping rows:", nrow(trans))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE, message = FALSE, warning = FALSE,
  collapse = TRUE, comment = "#>"
)

## ----load-data----------------------------------------------------------------
library(cat2cat)
library(dplyr)
library(tidyr)
library(e1071)
library(randomForest)

data(occup, package = "cat2cat")
data(trans, package = "cat2cat")

occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]

## ----mixed-ml-weights---------------------------------------------------------
occup_2_mix <- cat2cat(
  data = list(
    old = occup_2008, new = occup_2010,
    cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_2010,
    cat_var = "code",
    method = c("knn", "rf", "lda", "nb"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50),
    on_fail = "na"
  )
)

## ----weight-correlations------------------------------------------------------
occup_2_mix$old %>%
  select(wei_knn_c2c, wei_rf_c2c, wei_lda_c2c, wei_nb_c2c, wei_freq_c2c, wei_naive_c2c) %>%
    cor(use = "pairwise.complete.obs")

## ----ml-failure-policy, eval=FALSE--------------------------------------------
# ml_setup <- list(
#   data = bind_rows(occup_2010, occup_2012),
#   cat_var = "code",
#   method = c("knn", "rf", "lda"),
#   features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#   args = list(k = 10, ntree = 50),
#   on_fail = "freq",   # default policy
#   fail_warn = TRUE     # default reporting
# )
# 
# # strict mode for QA pipelines
# ml_strict <- ml_setup
# ml_strict$on_fail <- "error"
# 
# # diagnostic mode to inspect failures directly
# ml_diag <- ml_setup
# ml_diag$on_fail <- "na"
# ml_diag$fail_warn <- FALSE

## ----ensemble-prune-----------------------------------------------------------
occup_old_mix <- occup_2_mix$old %>%
  cross_c2c(.) %>%
  prune_c2c(., column = "wei_cross_c2c", method = "nonzero")

## ----sensitivity-result-------------------------------------------------------
result_all <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_2010, cat_var = "code",
    method = c("knn", "rf", "lda", "nb"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50)
  )
)

## ----sensitivity-counts-------------------------------------------------------
weight_cols <- c("wei_naive_c2c", "wei_freq_c2c", "wei_knn_c2c", "wei_rf_c2c", "wei_lda_c2c", "wei_nb_c2c")

# Pick groups with high replication
top_groups <- result_all$old %>%
  filter(rep_c2c > 1) %>%
  count(g_new_c2c, sort = TRUE) %>%
  head(6) %>%
  pull(g_new_c2c)

# Weighted counts from OLD period (replicated)
old_counts <- lapply(weight_cols, function(wcol) {
  result_all$old %>%
    filter(g_new_c2c %in% top_groups) %>%
    group_by(g_new_c2c) %>%
    summarise(n = sum(.data[[wcol]]), .groups = "drop")
}) %>%
  setNames(gsub("wei_|_c2c", "", weight_cols)) %>%
  bind_rows(.id = "method") %>%
  tidyr::pivot_wider(names_from = method, values_from = n)

# Counts from NEW period (no replication, exact)
new_counts <- result_all$new %>%
  filter(code %in% top_groups) %>%
  count(code, name = "new_period") %>%
  rename(g_new_c2c = code)

# Combine for comparison
left_join(old_counts, new_counts, by = "g_new_c2c")

## ----sensitivity-weights------------------------------------------------------
# New-period counts per category (no replication, so plain tally)
new_counts_all <- result_all$new %>%
  count(code, name = "n_new") %>%
  rename(g_new_c2c = code)

# Old-period weighted counts, joined to new-period counts
group_sizes <- result_all$old %>%
  group_by(g_new_c2c) %>%
  summarise(n_old = sum(wei_freq_c2c), .groups = "drop") %>%
  left_join(new_counts_all, by = "g_new_c2c") %>%
  filter(n_old >= 10, n_new >= 10) %>%
  arrange(desc(n_old))

# Pick a group for regression analysis
target_group <- group_sizes$g_new_c2c[1]
cat("Analysing occupation group:", target_group, "\n")

## ----sensitivity-group-reg----------------------------------------------------
# Subset old period to target group (with weights)
old_subset <- result_all$old %>%
  filter(g_new_c2c == target_group)

# Subset new period to target group (no replication, weight = 1)
new_subset <- result_all$new %>%
  filter(code == target_group) %>%
  mutate(
    wei_naive_c2c = 1, wei_freq_c2c = 1, wei_knn_c2c = 1,
    wei_rf_c2c = 1, wei_lda_c2c = 1, wei_nb_c2c = 1
  )

# Combine both periods
d <- bind_rows(old_subset, new_subset)

# Compare all regression coefficients across weight methods
f <- I(log(salary)) ~ age + sex + factor(edu) + exp + parttime

coefs <- sapply(weight_cols, function(wcol) {
  d$w <- d$multiplier * d[[wcol]]
  coef(lm(f, data = d, weights = w))
})
colnames(coefs) <- gsub("wei_|_c2c", "", weight_cols)
round(coefs, 4)

## ----sensitivity-pruning------------------------------------------------------
# Compare regression coefficients under different pruning strategies
prune_methods <- c("nonzero", "highest", "highest1")

prune_coefs <- sapply(prune_methods, function(pm) {
  old_pruned <- result_all$old %>%
    prune_c2c(method = pm) %>%
    filter(g_new_c2c == target_group)
  
  d <- bind_rows(old_pruned, new_subset)
  d$w <- d$multiplier * d$wei_freq_c2c
  coef(lm(f, data = d, weights = w))
})
round(prune_coefs, 4)

## ----sensitivity-ensemble-----------------------------------------------------
configs <- list(
  equal      = c(1, 1) / 2,
  freq_heavy = c(3, 1) / 4,
  ml_heavy   = c(1, 3) / 4
)

ens_coefs <- sapply(names(configs), function(nm) {
  old_ens <- result_all$old %>%
    cross_c2c(c("wei_freq_c2c", "wei_knn_c2c"), configs[[nm]]) %>%
    filter(g_new_c2c == target_group)
  
  new_ens <- new_subset %>% mutate(wei_cross_c2c = 1)
  d <- bind_rows(old_ens, new_ens)
  d$w <- d$multiplier * d$wei_cross_c2c
  coef(lm(f, data = d, weights = w))
})
round(ens_coefs, 4)

## ----cv-basic-----------------------------------------------------------------
cv_knn <- cat2cat_ml_run(
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = bind_rows(occup_2010, occup_2012),
    cat_var = "code",
    method = "knn",
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)
print(cv_knn)

## ----cv-multiple--------------------------------------------------------------
cv_all <- cat2cat_ml_run(
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = bind_rows(occup_2010, occup_2012),
    cat_var = "code",
    method = c("knn", "lda", "rf", "nb"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50)
  )
)
print(cv_all)

## ----cv-inspect---------------------------------------------------------------
# Pick a group with multiple candidates
group_names <- names(cv_all)
example_group <- group_names[
  which(vapply(cv_all, function(g) !is.na(g$freq) && g$naive < 1, logical(1)))[1]
]
cv_all[[example_group]]

#' @keywords internal
read_eq <- function(...) {
  xx <- as.list(substitute(list(...))[-1])
  ll <- length(xx)
  res <- vector("list", ll)
  for (i in seq_along(xx)) {
    children <- as.list(xx[[i]])
    for (c in seq_along(children)) {
      if (is.call(children[[c]])) {
        children[[c]] <- vapply(as.list(children[[c]])[-1], function(x) {
          deparse(x, width.cutoff = 500L)
        }, character(1))
      } else if (is.symbol(children[[c]])) {
        children[[c]] <- deparse(children[[c]], width.cutoff = 500L)
      }
    }
    res[[i]] <- children
  }
  res
}

#' @keywords internal
format_trans <- function(x) {
  forw <- c("%>%", ">")
  back <- c("%<%", "<")

  direc_df <- data.frame(
    direction = c(forw, back),
    names = c(rep("forward", 2), rep("backward", 2)),
    stringsAsFactors = FALSE
  )

  direct <- direc_df$names[match(x[[1]], direc_df$direction)]
  key_old <- x[[2]]
  key_new <- x[[3]]

  list(
    direction = direct,
    old = key_old,
    new = key_new
  )
}

#' Manual mapping for an aggregated panel dataset
#'
#' @description
#' Applies user-defined mapping equations to redistribute aggregated counts
#' (and other numeric columns) when categorical encodings change between
#' time periods. Unlike \code{\link{cat2cat}}, which works on micro-data,
#' this function operates on pre-aggregated data where each row represents
#' a category with associated counts/totals.
#'
#' @param data Named list with 5-6 fields describing the datasets:
#' \describe{
#'   \item{\code{old}}{\code{data.frame} - older time period (one row per category).}
#'   \item{\code{new}}{\code{data.frame} - newer time period (one row per category).}
#'   \item{\code{cat_var}}{\code{character(1)} - (deprecated) category variable name
#'     if identical in both periods. Use \code{cat_var_old}/\code{cat_var_new} instead.}
#'   \item{\code{cat_var_old}}{\code{character(1)} - category variable name in old period.}
#'   \item{\code{cat_var_new}}{\code{character(1)} - category variable name in new period.}
#'   \item{\code{time_var}}{\code{character(1)} - time variable name.}
#'   \item{\code{freq_var}}{\code{character(1)} - frequency/count variable used to
#'     compute proportions when splitting one category into many.}
#' }
#'
#' @param ... Mapping equations specifying how categories relate across periods.
#'   See **Equation syntax** in Details.
#'
#' @return A named list with two elements:
#' \describe{
#'   \item{\code{$old}}{\code{data.frame} - old period with \code{prop_c2c} column added.
#'     Categories may be replicated if split by backward equations.}
#'   \item{\code{$new}}{\code{data.frame} - new period with \code{prop_c2c} column added.
#'     Categories may be replicated if split by forward equations.}
#' }
#' The \code{prop_c2c} column contains the proportion (0-1) to apply when
#' aggregating. For rows not affected by any equation, \code{prop_c2c = 1}.
#' For split categories, proportions sum to 1 within the original category.
#'
#' @details
#'
#' \strong{Equation syntax}
#'
#' Each equation has the form:
#' \preformatted{OLD_SIDE  DIRECTION  NEW_SIDE}
#'
#' where:
#' \itemize{
#'   \item \strong{OLD_SIDE} - one or more old-period category names
#'     (use \code{c(A, B)} for multiple)
#'   \item \strong{DIRECTION} - one of:
#'     \itemize{
#'       \item \code{\%>\%} or \code{>} - \strong{forward}: replicates the NEW period,
#'         renaming/splitting new categories to match old encoding
#'       \item \code{\%<\%} or \code{<} - \strong{backward}: replicates the OLD period,
#'         renaming/splitting old categories to match new encoding
#'     }
#'   \item \strong{NEW_SIDE} - one or more new-period category names
#' }
#'
#' \strong{How proportions are calculated}
#'
#' When one category maps to multiple:
#' \itemize{
#'   \item \strong{Backward} (\code{\%<\%}): proportions come from \code{freq_var}
#'     in the \strong{new} period (the target encoding)
#'   \item \strong{Forward} (\code{\%>\%}): proportions come from \code{freq_var}
#'     in the \strong{old} period (the target encoding)
#' }
#'
#' \strong{Examples of valid equations}
#'
#' \describe{
#'   \item{\code{Automotive \%<\% c(Automotive1, Automotive2)}}{
#'     Backward: the old "Automotive" row is split into two rows
#'     ("Automotive1", "Automotive2") with proportions from new-period counts.}
#'   \item{\code{c(Kids1, Kids2) \%>\% c(Kids)}}{
#'     Forward: the new "Kids" row is split into two rows
#'     ("Kids1", "Kids2") with proportions from old-period counts.}
#'   \item{\code{Home \%>\% c(Home, Supermarket)}}{
#'     Forward: the new "Home" and "Supermarket" rows are each renamed to
#'     "Home" (1-to-many from new perspective; after aggregation they merge).}
#' }
#'
#' \strong{Typical workflow}
#'
#' 1. Call \code{cat2cat_agg()} with all mapping equations.
#' 2. Bind \code{$old} and \code{$new} together.
#' 3. Group by time and the (now unified) category variable.
#' 4. Summarise numeric columns as \code{sum(value * prop_c2c)}.
#'
#' @note
#' \itemize{
#'   \item All equations must be valid - unknown category names cause an error.
#'   \item Each equation must have exactly one category on one side
#'     (the "one" in one-to-many). Many-to-many within a single equation
#'     is not allowed.
#'   \item Each category in \code{old} and \code{new} must appear exactly once
#'     (no duplicates allowed before mapping).
#' }
#' @seealso \code{vignette("cat2cat_advanced")} for a complete workflow example.
#' @export
#' @examples
#' data("verticals", package = "cat2cat")
#' agg_old <- verticals[verticals$v_date == "2020-04-01", ]
#' agg_new <- verticals[verticals$v_date == "2020-05-01", ]
#'
#' # cat2cat_agg - can map in both directions at once
#' # although usually we want to have the old or the new representation
#'
#' agg <- cat2cat_agg(
#'   data = list(
#'     old = agg_old,
#'     new = agg_new,
#'     cat_var_old = "vertical",
#'     cat_var_new = "vertical",
#'     time_var = "v_date",
#'     freq_var = "counts"
#'   ),
#'   Automotive %<% c(Automotive1, Automotive2),
#'   c(Kids1, Kids2) %>% c(Kids),
#'   Home %>% c(Home, Supermarket)
#' )
#'
#' ## possible processing
#' library("dplyr")
#' agg %>%
#'   bind_rows() %>%
#'   group_by(v_date, vertical) %>%
#'   summarise(
#'     sales = sum(sales * prop_c2c),
#'     counts = sum(counts * prop_c2c),
#'     v_date = first(v_date)
#'   )
cat2cat_agg <- function(data = list(
                          old = NULL,
                          new = NULL,
                          cat_var_old = NULL,
                          cat_var_new = NULL,
                          time_var = NULL,
                          freq_var = NULL
                        ), ...) {
  if (!is.null(data$cat_var)) {
    data$cat_var_old <- data$cat_var
    data$cat_var_new <- data$cat_var
  }
  stopifnot(
    "`data` must be a list with old, new, cat_var_old/new, time_var, freq_var; both old and new must be data.frames containing the specified columns" =
      is.list(data) &&
      (length(data) >= 5 || length(data) <= 6) &&
      all(vapply(data, Negate(is.null), logical(1))) &&
      inherits(data$old, "data.frame") &&
      inherits(data$new, "data.frame") &&
      all(
        c(
          "old", "new", "cat_var_old", "cat_var_new",
          "time_var", "freq_var"
        ) %in% names(data)
      ) &&
      all(c(data$cat_var, data$freq_var) %in% colnames(data$old)) &&
      all(c(data$cat_var, data$freq_var) %in% colnames(data$new))
  )

  d_old <- length(unique(data$old[[data$time_var]]))
  d_new <- length(unique(data$new[[data$time_var]]))

  stopifnot(
    "Each period must have exactly one unique time value in `time_var`" =
      (d_old == 1) && (d_new == 1)
  )

  stopifnot(
    "Each category must appear exactly once per period (no duplicates)" =
      all(table(data$old[[data$cat_var_old]]) == 1) &&
      all(table(data$new[[data$cat_var_new]]) == 1)
  )

  t <- as.list(substitute(list(...))[-1])

  trans <- do.call(read_eq, t)
  trans_map <- lapply(trans, format_trans)
  old_cats <- unlist(lapply(trans_map, function(x) x[["old"]]))
  new_cats <- unlist(lapply(trans_map, function(x) x[["new"]]))
  stopifnot(
    "All old categories in equations must exist in `data$old`" =
      all(old_cats %in% unique(data[["old"]][[data$cat_var_old]]))
  )
  stopifnot(
    "All new categories in equations must exist in `data$new`" =
      all(new_cats %in% unique(data[["new"]][[data$cat_var_new]]))
  )

  df_old <- data$old
  df_old$prop_c2c <- 1
  col_df_old <- colnames(df_old)
  df_new <- data$new
  df_new$prop_c2c <- 1

  col_df_new <- colnames(df_new)

  stopifnot(
    "Old and new data.frames must have identical columns (except cat_var)" =
      identical(
        setdiff(col_df_old, data$cat_var_old),
        setdiff(col_df_new, data$cat_var_new)
      )
  )

  for (i in trans_map) {
    stopifnot(
      "Each equation must have exactly one category on one side (one-to-many)" =
        ((length(i[[2]]) == 1) || (length(i[[3]]) == 1))
    )

    if (i$direction == "forward") {
      base <- df_new[!(df_new[, data$cat_var_new] %in% i[[3]]), ]
      base_rm <- df_new[df_new[, data$cat_var_new] %in% i[[3]], ]

      if (length(i[[2]]) > 1) {
        base_rm <- base_rm[rep(1, length(i[[2]])), ]
        base_rm[, data$cat_var_new] <- i[[2]]
        base_rm$prop_c2c <- prop.table(
          df_old[df_old[, data$cat_var_old] %in% i[[2]], data$freq_var]
        )
      } else {
        base_rm[, data$cat_var_new] <- i[[2]]
        base_rm$prop_c2c <- 1
      }

      df_new <- rbind(base, base_rm)
    } else if (i$direction == "backward") {
      base <- df_old[!(df_old[, data$cat_var_old] %in% i[[2]]), ]
      base_rm <- df_old[df_old[, data$cat_var_old] %in% i[[2]], ]

      if (length(i[[3]]) > 1) {
        base_rm <- base_rm[rep(1, length(i[[3]])), ]
        base_rm[, data$cat_var_old] <- i[[3]]
        base_rm$prop_c2c <- prop.table(
          df_new[df_new[, data$cat_var_new] %in% i[[3]], data$freq_var]
        )
      } else {
        base_rm[, data$cat_var_old] <- i[[3]]
        base_rm$prop_c2c <- 1
      }

      df_old <- rbind(base, base_rm)
    }
  }

  list(old = df_old, new = df_new)
}

set.seed(1234)

agg_old <- data.frame(
  vertical = c(
    "Electronics", "Kids1", "Kids2", "Automotive", "Books",
    "Clothes", "Home", "Fashion", "Health", "Sport"
  ),
  sales = rnorm(10, 100, 10),
  counts = rgeom(10, 0.0001),
  v_date = rep("2020-04-01", 10), stringsAsFactors = FALSE
)

agg_new <- data.frame(
  vertical = c(
    "Electronics", "Supermarket", "Kids", "Automotive1",
    "Automotive2", "Books", "Clothes", "Home", "Fashion", "Health", "Sport"
  ),
  sales = rnorm(11, 100, 10),
  counts = rgeom(11, 0.0001),
  v_date = rep("2020-05-01", 11), stringsAsFactors = FALSE
)

testthat::test_that("wrong input", {
  testthat::expect_error(cat2cat_agg(list()))
  testthat::expect_error(cat2cat_agg(list(
    old = agg_old,
    new = agg_new,
    cat_var = "WRONG",
    time_var = "v_date",
    freq_var = "counts"
  )))
  testthat::expect_error(cat2cat_agg(list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "WRONG",
    freq_var = "counts"
  )))
  testthat::expect_error(cat2cat_agg(list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "WRONG"
  )))
})

testthat::test_that("error when - wrong mapping", {
  testthat::expect_error(cat2cat_agg(list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ), ff %>% ee()))
})

testthat::test_that("no mapping is neutral for data", {
  agg_neu <- cat2cat_agg(list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ))
  testthat::expect_identical(agg_neu$old, cbind(agg_old, prop_c2c = 1))
  testthat::expect_identical(agg_neu$new, cbind(agg_new, prop_c2c = 1))
})

agg <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  Automotive %<% c(Automotive1, Automotive2),
  c(Kids1, Kids2) %>% c(Kids),
  Home %>% c(Home, Supermarket)
)

testthat::test_that("cat2cat_agg scenario one", {
  expect_true(sum(agg$old$prop_c2c) == nrow(agg_old))
  expect_true(sum(agg$new$prop_c2c) == nrow(agg_new))
  expect_true(all(agg$new$prop_c2c >= 0 & agg$new$prop_c2c <= 1))
  expect_true(all(agg$old$prop_c2c >= 0 & agg$old$prop_c2c <= 1))
})

agg2 <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  Automotive < c(Automotive1, Automotive2),
  c(Kids1, Kids2) > c(Kids),
  Home > c(Home, Supermarket)
)

testthat::test_that("cat2cat_agg scenario two", {
  expect_true(sum(agg2$old$prop_c2c) == nrow(agg_old))
  expect_true(sum(agg2$new$prop_c2c) == nrow(agg_new))
  expect_true(all(agg2$new$prop_c2c >= 0 & agg2$new$prop_c2c <= 1))
  expect_true(all(agg2$old$prop_c2c >= 0 & agg2$old$prop_c2c <= 1))
})

cols <- colnames(agg_new)
cols[1] <- "vertical2"
colnames(agg_new) <- cols
agg3 <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var_old = "vertical",
    cat_var_new = "vertical2",
    time_var = "v_date",
    freq_var = "counts"
  ),
  Automotive < c(Automotive1, Automotive2),
  c(Kids1, Kids2) > c(Kids),
  Home > c(Home, Supermarket)
)

testthat::test_that("cat2cat_agg scenario three", {
  cols <- colnames(agg3$new)
  cols[1] <- "vertical"
  colnames(agg3$new) <- cols
  expect_identical(agg2, agg3)
})

testthat::test_that("cat2cat_agg backward many-to-one sets prop_c2c to 1", {
  agg_back_single <- cat2cat_agg(
    data = list(
      old = agg_old,
      new = agg_new,
      cat_var_old = "vertical",
      cat_var_new = "vertical2",
      time_var = "v_date",
      freq_var = "counts"
    ),
    c(Kids1, Kids2) %<% c(Kids)
  )

  kids_rows <- agg_back_single$old$vertical %in% c("Kids")
  expect_true(any(kids_rows))
  expect_true(all(agg_back_single$old$prop_c2c[kids_rows] == 1))
})

#' Summary plots for cat2cat results
#' @description This function help to understand properties of cat2cat results.
#' It is recommended to run it before further processing, like next iterations.
#' @param data `data.frame` - one of the data.frames returned by
#' the `cat2cat` function.
#' @param weis `character(1)` - name of a certain wei_*_c2c column,
#' added by cat2cat function. Default `wei_freq_c2c`.
#' @param type `character(1)` - one of 3 types `"both"`, `"hist"`, `"bar"`.
#' @return base plot graphics
#' @importFrom graphics hist barplot par
#' @importFrom utils head
#' @note It will work only for data.frame produced by cat2cat function.
#' @export
#' @examples
#' data("occup_small", package = "cat2cat")
#' occup_old <- occup_small[occup_small$year == 2008, ]
#' occup_new <- occup_small[occup_small$year == 2010, ]
#'
#' occup_2 <- cat2cat(
#'   data = list(
#'     old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
#'   ),
#'   mappings = list(trans = trans, direction = "backward")
#' )
#'
#' plot_c2c(occup_2$old, type = c("both"))
#' plot_c2c(occup_2$old, type = c("hist"))
#' plot_c2c(occup_2$old, type = c("bar"))
plot_c2c <- function(data,
                     weis = "wei_freq_c2c",
                     type = c("both", "hist", "bar")) {
  stopifnot("`data` must be a data.frame" = is.data.frame(data))
  stopifnot(
    "`weis` must be a single character string" =
      is.character(weis) && (length(weis) == 1)
  )
  stopifnot(
    "There are no cat2cat additional wei_* columns/variables" =
      weis %in% colnames(data)
  )
  stopifnot(
    "There are no replications. Try to switch between old/new period." =
      nrow(data) != sum(data[[weis]])
  )
  type <- match.arg(type)

  if (type == "both") {
    default_mfrow <- par("mfrow")
    par(mfrow = c(1, 2))
    on.exit(par(mfrow = default_mfrow))
  }
  if (type %in% c("both", "bar")) {
    data_s <- unlist(
      tapply(data$rep_c2c, data$index_c2c, function(x) head(x, 1))
    )
    barplot(
      prop.table(table(data_s)),
      main = "Number of Replications",
      ylab = "Fraction",
      xlab = "rep_c2c"
    )
  }
  if (type %in% c("both", "hist")) {
    hist(
      data[[weis]],
      freq = FALSE,
      main = "Replications Probabilities",
      xlab = weis
    )
  }
}

data("occup_small", package = "cat2cat")

occup_old <- occup_small[occup_small$year == 2008, ]
occup_new <- occup_small[occup_small$year == 2010, ]

occup_2 <- cat2cat(
  data = list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward")
)

testthat::test_that("plot_c2c wrong input", {
  expect_error(plot_c2c(occup_2$new, type = c("both")))
  expect_error(plot_c2c(occup_2$old, type = c("wrong")))
  expect_error(plot_c2c(NULL, type = c("both")))
})

testthat::test_that("plot_c2c correct input", {
  expect_silent(plot_c2c(occup_2$old, type = c("both")))
  expect_silent(plot_c2c(occup_2$old, type = c("hist")))
  expect_silent(plot_c2c(occup_2$old, type = c("bar")))
})

#' Pruning which could be useful after the mapping process
#'
#' @description user could specify one of four methods to prune replications
#' created in the cat2cat procedure.
#'
#' @param df `data.frame` like result of the `cat2cat` function
#' for a specific period.
#' @param index `character(1)` a column name with the `cat2cat` identifier.
#' Should not be updated in most cases. Default `index_c2c`.
#' @param column `character(1)` a column name with weights,
#' default `wei_freq_c2c`.
#' @param method `character(1)` one of four available methods:
#' "nonzero" (default), "highest", "highest1" or "morethan".
#' @param percent `integer(1)` from 0 to 99
#' @return `data.frame` with the same structure and possibly reduced
#' number of rows
#' @details
#' method - specify a method to reduce number of replications
#' \describe{
#'  \item{"nonzero"}{ remove nonzero probabilities}
#'  \item{"highest"}{
#'  leave only highest probabilities for each subject- accepting ties
#'  }
#'  \item{"highest1"}{
#'  leave only highest probabilities for each subject -
#'  not accepting ties so always one is returned
#'  }
#'  \item{"morethan"}{
#'  leave rows where a probability is higher than value specify
#'  by percent argument
#'  }
#' }
#' @export
#' @examples
#' \dontrun{
#' data("occup_small", package = "cat2cat")
#' data("occup", package = "cat2cat")
#' data("trans", package = "cat2cat")
#'
#' occup_old <- occup_small[occup_small$year == 2008, ]
#' occup_new <- occup_small[occup_small$year == 2010, ]
#'
#' occup_ml <- cat2cat(
#'   data = list(
#'     old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
#'   ),
#'   mappings = list(trans = trans, direction = "backward"),
#'   ml = list(
#'     data = occup_new,
#'     cat_var = "code",
#'     method = "knn",
#'     features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#'     args = list(k = 10)
#'   )
#' )
#'
#' prune_c2c(occup_ml$old, method = "nonzero")
#' prune_c2c(occup_ml$old, method = "highest")
#' prune_c2c(occup_ml$old, method = "highest1")
#' prune_c2c(occup_ml$old, method = "morethan", percent = 90)
#'
#' prune_c2c(occup_ml$old, column = "wei_knn_c2c", method = "nonzero")
#' }
#'
prune_c2c <- function(df,
                      index = "index_c2c",
                      column = "wei_freq_c2c",
                      method = "nonzero",
                      percent = 50) {
  stopifnot("`df` must be a data.frame" = is.data.frame(df))
  stopifnot(
    "`index` and `column` must be columns in `df`" =
      all(c(index, column) %in% colnames(df))
  )
  stopifnot(
    "`method` must be one of: 'nonzero', 'highest', 'highest1', 'morethan'" =
      isTRUE(method %in% c("nonzero", "highest", "highest1", "morethan"))
  )
  stopifnot(
    "`percent` must be a single number between 0 and 100" =
      length(percent) == 1 && (percent >= 0 && percent < 100)
  )

  df <- df[order(df[[index]]), ]

  df <- switch(method,
    nonzero = df[df[[column]] > 0, ],
    highest1 = {
      highest1_fun <- function(x) seq_along(x) == which.max(x)
      df[unlist(tapply(df[[column]], df[[index]], highest1_fun)), ]
    },
    highest = {
      highest_fun <- function(x) x == max(x)
      df[unlist(tapply(df[[column]], df[[index]], highest_fun)), ]
    },
    morethan = df[df[[column]] > percent / 100, ]
  )
  # reweight to still sum to 1 per subject
  df[[column]] <- unlist(
    tapply(df[[column]], df[[index]], function(x) x / sum(x))
  )

  df
}

#' Make a combination of weights from different methods
#'
#' @description adding the additional column which is a mix of weights columns
#' by each row.
#' Ensemble of a few methods usually produces more accurate solutions
#' than a single model would.
#'
#' @param df `data.frame` like result of the `cat2cat` function
#' for a specific period.
#' @param cols `character` vector default all columns under
#' the regex "wei_.*_c2c".
#' @param weis `numeric` vector weighs for columns in the `cols` argument.
#' By default a vector of the same length as `cols` argument and with equally
#' spaced probability (summing to 1).
#' @param na.rm `logical(1)` if `NA` values should be omitted, default TRUE.
#' @return `data.frame` with the additional column `wei_cross_c2c`.
#' @export
#' @examples
#' \dontrun{
#' data("occup_small", package = "cat2cat")
#' data("occup", package = "cat2cat")
#' data("trans", package = "cat2cat")
#'
#' occup_old <- occup_small[occup_small$year == 2008, ]
#' occup_new <- occup_small[occup_small$year == 2010, ]
#'
#' # mix of methods - forward direction, try out backward too
#' occup_mix <- cat2cat(
#'   data = list(
#'     old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
#'   ),
#'   mappings = list(trans = trans, direction = "backward"),
#'   ml = list(
#'     data = occup_new,
#'     cat_var = "code",
#'     method = c("knn"),
#'     features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#'     args = list(k = 10, ntree = 20)
#'   )
#' )
#' # correlation between ml model
#' occup_mix_old <- occup_mix$old
#' cor(
#'   occup_mix_old[occup_mix_old$rep_c2c != 1, c("wei_knn_c2c", "wei_freq_c2c")]
#' )
#' # cross all methods and subset one highest probability category for each obs
#' occup_old_highest1_mix <- prune_c2c(cross_c2c(occup_mix$old),
#'   column = "wei_cross_c2c", method = "highest1"
#' )
#' }
#'
cross_c2c <- function(df,
                      cols = colnames(df)[grepl("^wei_.*_c2c$", colnames(df))],
                      weis = rep(1 / length(cols), length(cols)),
                      na.rm = TRUE) {
  stopifnot("`df` must be a data.frame" = is.data.frame(df))
  stopifnot(
    "All `cols` must be columns in `df`" =
      all(cols %in% colnames(df))
  )
  stopifnot(
    "`weis` must have same length as `cols`" =
      length(weis) == length(cols)
  )
  stopifnot("`na.rm` must be TRUE or FALSE" = is.logical(na.rm))

  weis <- weis / sum(weis)

  df[["wei_cross_c2c"]] <- as.vector(
    rowSums(t(t(as.matrix(df[, cols])) * weis), na.rm = na.rm)
  )

  df
}


#' Add default cat2cat columns to a `data.frame`
#' @description a utils function to add default cat2cat columns
#' to a `data.frame`.
#' It will be useful e.g. for a boarder periods which will not have additional
#'  `cat2cat` columns.
#' @param df `data.frame`.
#' @param cat_var `character(1)` a categorical variable name.
#' @param ml `character` vector of ml models applied,
#' any of `c("knn", "rf", "lda", "nb")`.
#' @return the provided `data.frame` with additional `cat2cat` like columns.
#' @export
#' @examples
#' \dontrun{
#' dummy_c2c(airquality, "Month")
#'
#' data("occup_small", package = "cat2cat")
#' occup_old <- occup_small[occup_small$year == 2008, ]
#' dummy_c2c(occup_old, "code")
#' dummy_c2c(occup_old, "code", "knn")
#' }
dummy_c2c <- function(df, cat_var, ml = NULL) {
  stopifnot("`df` must be a data.frame" = is.data.frame(df))
  stopifnot(
    "`cat_var` must be a single character string" =
      length(cat_var) == 1 && is.character(cat_var)
  )
  stopifnot(
    "`cat_var` must be a column in `df`" =
      isTRUE(cat_var %in% colnames(df))
  )
  stopifnot(
    "`ml` must be NULL or contain valid ML method names (knn, rf, lda, nb)" =
      is.null(ml) ||
      (all(ml %in% c("knn", "rf", "lda", "nb")) ||
       all(ml %in% paste0("wei_", c("knn", "rf", "lda", "nb"), "_c2c")))
  )

  base_cols <- c(
    "index_c2c", "g_new_c2c", "wei_freq_c2c",
    "rep_c2c", "wei_naive_c2c"
  )
  if (!all(base_cols %in% colnames(df))) {
    df$index_c2c <- seq_len(nrow(df))
    df$g_new_c2c <- df[[cat_var]]
    df$wei_freq_c2c <- 1
    df$rep_c2c <- 1
    df$wei_naive_c2c <- 1
  }

  if (!is.null(ml)) {
    ml_cols <- if (any(grepl("wei_.*_c2c", ml))) {
      ml
    } else {
      paste0("wei_", ml, "_c2c")
    }
    df[, setdiff(ml_cols, colnames(df))] <- 1
  }

  df
}

set.seed(1234)

data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup[occup$year == 2008, ]
occup_new <- occup[occup$year == 2010, ]

testthat::test_that("incorect input", {
  testthat::expect_error(cat2cat())
  testthat::expect_error(
    cat2cat(data = list(
      old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
    ))
  )
})

testthat::test_that("no mappings so empty dataset in one direction", {
  data <- list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  )
  mappings <- list(
    trans = data.frame(old = 1, new = 1)[NULL, ],
    direction = "backward"
  )

  testthat::expect_warning(
    cat2cat(data = data, mappings = mappings),
    "trans table does not cover some levels "
  )
  res <- suppressWarnings(cat2cat(data = data, mappings = mappings))
  testthat::expect_identical(nrow(res$old), 0L)
})

data_simple <- list(
  old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
)
mappings_simple_back <- list(trans = trans, direction = "backward")
mappings_simple_for <- list(trans = trans, direction = "forward")
ml <- list(
  data = occup_new,
  cat_var = "code",
  method = c("knn", "rf", "lda"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 30),
  fail_warn = FALSE
)

testthat::test_that("Simple backward cat2cat with 2 periods and no ml", {
  occup_1a <- cat2cat(data = data_simple, mappings = mappings_simple_back)
  expect_true(all(occup_1a$old$wei_freq_c2c <= 1 &
    occup_1a$old$wei_freq_c2c >= 0))
  expect_true(all(occup_1a$old$wei_naive_c2c <= 1 &
    occup_1a$old$wei_naive_c2c >= 0))
  expect_equal(sum(occup_1a$old$wei_naive_c2c), nrow(occup_old))
  expect_equal(sum(occup_1a$old$wei_freq_c2c), nrow(occup_old))
})


testthat::test_that("Simple backward cat2cat with 2 periods and ml", {
  occup_2 <- cat2cat(
    data = data_simple,
    mappings = mappings_simple_back,
    ml = ml
  )

  expect_true(!identical(occup_2$old$wei_freq_c2c, occup_2$old$wei_rf_c2c))
  expect_true(!identical(occup_2$old$wei_freq_c2c, occup_2$old$wei_knn_c2c))
  expect_true(!identical(occup_2$old$wei_freq_c2c, occup_2$old$wei_lda_c2c))
  expect_true(!identical(occup_2$old$wei_freq_c2c, occup_2$old$wei_naive_c2c))

  expect_equal(sum(occup_2$old$wei_freq_c2c), nrow(occup_old))
  expect_equal(sum(occup_2$old$wei_knn_c2c), nrow(occup_old))
  expect_equal(sum(occup_2$old$wei_rf_c2c), nrow(occup_old))
  expect_equal(sum(occup_2$old$wei_lda_c2c), nrow(occup_old))
  expect_equal(sum(occup_2$old$wei_naive_c2c), nrow(occup_old))

  expect_true(
    all(occup_2$old$wei_freq_c2c <= 1 & occup_2$old$wei_freq_c2c >= 0)
  )
  expect_true(
    all(occup_2$old$wei_freq_c2c <= 1 & occup_2$old$wei_freq_c2c >= 0)
  )
  expect_true(
    all(occup_2$old$wei_knn_c2c <= 1 & occup_2$old$wei_knn_c2c >= 0)
  )
  expect_true(
    all(occup_2$old$wei_rf_c2c <= 1 & occup_2$old$wei_rf_c2c >= 0)
  )
  expect_true(
    all(occup_2$old$wei_lda_c2c <= 1 & occup_2$old$wei_lda_c2c >= 0)
  )

  expect_equal(
    occup_2$old %>%
      cross_c2c(., c("wei_freq_c2c", "wei_knn_c2c"), c(1 / 2, 1 / 2)) %>%
      pull("wei_cross_c2c"),
    (occup_2$old$wei_knn_c2c + occup_2$old$wei_freq_c2c) / 2
  )
})

testthat::test_that(
  "Simple backward cat2cat with 2 periods and ml, not all mappings",
  {
    expect_warning(
      cat2cat(
        data = data_simple,
        mappings = list(trans = head(trans, -50), direction = "backward"),
        ml = ml
      ),
      "9321, 9312, 9311, 9331, 9313"
    )
  }
)

# handling NAs
occup_3a <- cat2cat(
  data = list(
    old = occup_old,
    new = occup_new,
    cat_var = "code",
    time_var = "year",
    multiplier_var = "multiplier"
  ),
  mappings = list(
    trans = trans,
    direction = "backward"
  ),
  ml = list(
    data = occup_new,
    cat_var = "code",
    method = c("knn"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)

na_row <- occup_old[1, ]
na_row$code <- NA
na_row2 <- occup_new[1, ]
na_row2$code <- NA
occup_3b <- cat2cat(
  data = list(
    old = rbind(occup_old, na_row),
    new = rbind(occup_new, na_row2),
    cat_var = "code",
    time_var = "year",
    multiplier_var = "multiplier"
  ),
  mappings = list(
    trans = do.call(rbind, list(trans, c(NA, NA), c(NA, "432190"))),
    direction = "backward"
  ),
  ml = list(
    data = rbind(occup_new, na_row2),
    cat_var = "code",
    method = c("knn"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)

expect_identical(nrow(occup_3b$old) - 2L, nrow(occup_3a$old))
expect_true(all(occup_3b$old$wei_freq_c2c <= 1 &
  occup_3b$old$wei_freq_c2c >= 0))
expect_true(all(occup_3b$old$wei_knn_c2c <= 1 & occup_3b$old$wei_knn_c2c >= 0))

na_row <- occup_old[1, ]
na_row$code <- "NA"
occup_3c <- cat2cat(
  data = list(
    old = rbind(occup_old, na_row),
    new = occup_new,
    cat_var = "code",
    time_var = "year",
    multiplier_var = "multiplier"
  ),
  mappings = list(trans = rbind(trans, c("NA", "NA")), direction = "backward"),
  ml = list(
    method = c("knn"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)

expect_identical(nrow(occup_3b$old), nrow(occup_3c$old) + 1L)
expect_true(all(occup_3c$old$wei_freq_c2c <= 1 &
  occup_3c$old$wei_freq_c2c >= 0))
expect_true(all(occup_3c$old$wei_knn_c2c <= 1 & occup_3c$old$wei_knn_c2c >= 0))

# forward and not all mappings

testthat::test_that(
  "Simple forward cat2cat with 2 periods and no ml, not all mappings",
  {
    expect_warning(
      cat2cat(
        data = data_simple,
        mappings = mappings_simple_for
      ),
      paste(
        "741103, 712604, 732201, 818116, 732301,",
        "816003, 741201, 713201, 818115, 815204"
      )
    )
  }
)

testthat::test_that(
  "Simple forward cat2cat with 2 periods and no ml - details",
  {
    occup_4 <- suppressWarnings(cat2cat(
      data = data_simple,
      mappings = mappings_simple_for
    ))

    expect_equal(
      sum(occup_4$new$wei_freq_c2c) +
        sum(occup_new$code %in% setdiff(occup_new$code, trans$new)),
      nrow(occup_new)
    )
    expect_true(all(occup_4$new$wei_freq_c2c <= 1 &
      occup_4$new$wei_freq_c2c >= 0))

    occup_4b <- cat2cat(
      data = list(
        old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
      ),
      mappings = list(
        trans = rbind(
          trans,
          data.frame(old = "no_cat", new = setdiff(occup_new$code, trans$new))
        ),
        direction = "forward"
      )
    )

    expect_identical(nrow(occup_4b$old), nrow(occup_4b$old))
    expect_true(all(occup_4b$old$wei_freq_c2c <= 1 &
      occup_4b$old$wei_freq_c2c >= 0))
  }
)

# automatic mapping table
# the ean variable is an unique identifier
data("verticals2", package = "cat2cat")

vert_old <- verticals2[verticals2$v_date == "2020-04-01", ]
vert_new <- verticals2[verticals2$v_date == "2020-05-01", ]

## get mapping (transition) table
trans_v <- vert_old %>%
  inner_join(vert_new, by = "ean") %>%
  select(vertical.x, vertical.y) %>%
  distinct()

# as then we merging categories 1 to 1 for this identifiers

testthat::test_that(
  "cat2cat - set id_var",
  {
    verts <- cat2cat(
      data = list(
        old = vert_old, new = vert_new, id_var = "ean",
        cat_var = "vertical", time_var = "v_date"
      ),
      mappings = list(trans = trans_v, direction = "backward")
    )

    expect_true(all(verts$old$wei_freq_c2c <= 1 & verts$old$wei_freq_c2c >= 0))
    expect_true(all(verts$new$wei_freq_c2c <= 1 & verts$new$wei_freq_c2c >= 0))
    expect_equal(sum(verts$old$wei_naive_c2c), nrow(vert_old))
    expect_equal(sum(verts$old$wei_freq_c2c), nrow(vert_old))
  }
)

testthat::test_that(
  "set id_var, backward and ml",
  {
    verts2 <- cat2cat(
      data = list(
        old = vert_old, new = vert_new, id_var = "ean",
        cat_var = "vertical", time_var = "v_date"
      ),
      mappings = list(trans = trans_v, direction = "backward"),
      ml = list(
        data = vert_new,
        cat_var = "vertical",
        method = c("knn", "rf", "lda"),
        features = c("sales"),
        args = list(k = 10, ntree = 30)
      )
    )

    expect_true(!identical(verts2$old$wei_freq_c2c, verts2$old$wei_naive_c2c))
    expect_true(!identical(verts2$old$wei_freq_c2c, verts2$old$wei_rf_c2c))
    expect_true(!identical(verts2$old$wei_freq_c2c, verts2$old$wei_knn_c2c))
    expect_true(!identical(verts2$old$wei_freq_c2c, verts2$old$wei_lda_c2c))

    expect_equal(sum(verts2$old$wei_freq_c2c), nrow(vert_old))
    expect_equal(sum(verts2$old$wei_knn_c2c), nrow(vert_old))
    expect_equal(sum(verts2$old$wei_rf_c2c), nrow(vert_old))
    expect_equal(sum(verts2$old$wei_lda_c2c), nrow(vert_old))

    expect_true(
      (all(verts2$old$wei_knn_c2c <= 1 & verts2$old$wei_knn_c2c >= 0))
    )
    expect_true(
      (all(verts2$old$wei_rf_c2c <= 1 & verts2$old$wei_rf_c2c >= 0))
    )
    expect_true(
      (all(verts2$old$wei_lda_c2c <= 1 & verts2$old$wei_lda_c2c >= 0))
    )

    expect_true(all(unique(verts2[["old"]][["g_new_c2c"]]) %in% trans_v[[2]]))
  }
)

testthat::test_that(
  "set id_var, forward and ml",
  {
    verts3 <- cat2cat(
      data = list(
        old = vert_old, new = vert_new, id_var = "ean",
        cat_var = "vertical", time_var = "v_date"
      ),
      mappings = list(trans = trans_v, direction = "forward"),
      ml = list(
        data = vert_old,
        cat_var = "vertical",
        method = c("knn", "rf", "lda"),
        features = c("sales"),
        args = list(k = 10, ntree = 30)
      )
    )

    expect_true(!identical(verts3$new$wei_freq_c2c, verts3$new$wei_rf_c2c))
    expect_true(!identical(verts3$new$wei_freq_c2c, verts3$new$wei_knn_c2c))
    expect_true(!identical(verts3$new$wei_freq_c2c, verts3$new$wei_lda_c2c))

    expect_equal(sum(verts3$new$wei_freq_c2c), nrow(vert_new))
    expect_equal(sum(verts3$new$wei_knn_c2c), nrow(vert_new))
    expect_equal(sum(verts3$new$wei_rf_c2c), nrow(vert_new))
    expect_equal(sum(verts3$new$wei_lda_c2c), nrow(vert_new))

    expect_true(
      (all(verts3$new$wei_knn_c2c <= 1 & verts3$new$wei_knn_c2c >= 0))
    )
    expect_true(
      (all(verts3$new$wei_rf_c2c <= 1 & verts3$new$wei_rf_c2c >= 0))
    )
    expect_true(
      (all(verts3$new$wei_lda_c2c <= 1 & verts3$new$wei_lda_c2c >= 0))
    )

    expect_true(all(unique(verts3[["new"]][["g_new_c2c"]]) %in% trans_v[[1]]))
  }
)

set.seed(1234)

data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_2006 <- occup[occup$year == 2006, ]
occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]

ml_setup <- list(
  data = occup_2010,
  cat_var = "code",
  method = c("knn"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 50)
)

occup_back_2008_2010 <- cat2cat(
  data = list(
    old = occup_2008, new = occup_2010, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)

# the counts could be any of wei_* or their combination
freqs_df <- occup_back_2008_2010$old[, c("g_new_c2c", "wei_freq_c2c")] %>%
  cross_c2c() %>%
  group_by(g_new_c2c) %>%
  summarise(counts = round(sum(wei_cross_c2c, na.rm = TRUE)))

occup_back_2006_2008_1 <- cat2cat(
  data = list(
    old = occup_2006,
    new = occup_back_2008_2010$old,
    cat_var_old = "code",
    cat_var_new = "g_new_c2c",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)

testthat::test_that("multi-period cat2cat neutral for base period", {
  expect_identical(occup_back_2008_2010$old, occup_back_2006_2008_1$new)
})

testthat::test_that("multi-period cat2cat probabilities", {
  expect_true(
    !identical(
      occup_back_2006_2008_1$old$wei_freq_c2c,
      occup_back_2006_2008_1$old$wei_naive_c2c
    )
  )
  expect_true(
    !identical(
      occup_back_2006_2008_1$old$wei_freq_c2c,
      occup_back_2006_2008_1$old$wei_knn_c2c
    )
  )
  expect_true(!identical(
    occup_back_2006_2008_1$old$wei_naive_c2c,
    occup_back_2006_2008_1$old$wei_knn_c2c
  ))

  expect_equal(
    sum(occup_back_2006_2008_1$old$wei_freq_c2c),
    nrow(occup_2006)
  )
  expect_equal(
    sum(occup_back_2006_2008_1$old$wei_knn_c2c),
    nrow(occup_2006)
  )

  expect_true((all(occup_back_2006_2008_1$old$wei_freq_c2c <= 1 &
    occup_back_2006_2008_1$old$wei_freq_c2c >= 0)))
  expect_true((all(occup_back_2006_2008_1$old$wei_knn_c2c <= 1 &
    occup_back_2006_2008_1$old$wei_knn_c2c >= 0)))
})

occup_2006_new <- occup_back_2006_2008_1$old
occup_2008_new <- occup_back_2008_2010$old # or occup_back_2006_2008$new
occup_2010_new <- occup_back_2008_2010$new
occup_2012_new <- dummy_c2c(occup_2012,
  cat_var = "code",
  ml = c("knn")
)

final_data <- do.call(rbind, list(
  occup_2006_new,
  occup_2008_new,
  occup_2010_new,
  occup_2012_new
))

counts_new <- final_data %>%
  cross_c2c() %>%
  group_by(year) %>%
  summarise(
    n = as.integer(round(sum(wei_freq_c2c))),
    n2 = as.integer(round(sum(wei_cross_c2c))),
    .groups = "drop"
  )

counts_old <- occup %>%
  group_by(year) %>%
  summarise(n = n(), n2 = n(), .groups = "drop")

testthat::test_that("multi-period cat2cat persist the number of observations", {
  expect_identical(counts_new, counts_old)
})

data(occup)
data(trans)

occup_2006 <- occup[occup$year == 2006, ]
occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]

trans2 <- rbind(
  trans,
  data.frame(
    old = "no_cat",
    new = setdiff(c(occup_2010$code, occup_2012$code), trans$new)
  )
)

# from 2008 to 2010
occup_for_2008_2010 <- cat2cat(
  data = list(
    old = occup_2008, new = occup_2010, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans2, direction = "forward"),
  ml = ml_setup
)

# optional, give more control
# the counts could be any of wei_* or their combination
freqs_df <- occup_for_2008_2010$new[, c("g_new_c2c", "wei_freq_c2c")] %>%
  group_by(g_new_c2c) %>%
  summarise(counts = round(sum(wei_freq_c2c)))

# from 2010 to 2012
occup_for_2010_2012 <- cat2cat(
  data = list(
    old = occup_for_2008_2010$new,
    new = occup_2012,
    cat_var_old = "g_new_c2c",
    cat_var_new = "code",
    time_var = "year",
    freqs_df = freqs_df
  ),
  mappings = list(trans = trans2, direction = "forward"),
  ml = ml_setup
)

# from 2010 to 2012
occup_for_2010_2012_2 <- cat2cat(
  data = list(
    old = occup_for_2008_2010$new,
    new = occup_2012,
    cat_var_old = "g_new_c2c",
    cat_var_new = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans2, direction = "forward", freqs_df = freqs_df),
  ml = ml_setup
)

expect_identical(occup_for_2010_2012_2, occup_for_2010_2012)

# use ml argument when applied ml models
occup_2006_new <- dummy_c2c(occup_2006, "code", ml = c("knn"))
occup_2008_new <- occup_for_2008_2010$old
occup_2010_new <- occup_for_2008_2010$new # or occup_for_2010_2012$old
occup_2012_new <- occup_for_2010_2012$new

final_data_for <- do.call(
  rbind,
  list(occup_2006_new, occup_2008_new, occup_2010_new, occup_2012_new)
)

# We persist the number of observations
counts_new <- final_data_for %>%
  cross_c2c() %>%
  group_by(year) %>%
  summarise(
    n = as.integer(round(sum(wei_freq_c2c))),
    n2 = as.integer(round(sum(wei_cross_c2c)))
  )

counts_old <- occup %>%
  group_by(year) %>%
  summarise(n = n(), n2 = n(), .groups = "drop")

testthat::test_that("", {
  testthat::expect_identical(counts_new, counts_old)
})

data("trans", package = "cat2cat")
data("occup", package = "cat2cat")
# Split the panel dataset to separate years
occup_2006 <- occup[occup$year == 2006, ]
occup_2008 <- occup_old <- occup[occup$year == 2008, ]
occup_2010 <- occup_new <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]

occup_simple_backward <- cat2cat::cat2cat(
  data = list(
    old = occup_old,
    new = occup_new,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(
    trans = trans,
    direction = "backward"
  )
)

testthat::test_that("Simple cat2cat example - two periods - backward", {
  testthat::expect_true(all(names(occup_simple_backward) %in% c("old", "new")))
  testthat::expect_true(is.data.frame(occup_simple_backward$old))
  testthat::expect_true(is.data.frame(occup_simple_backward$new))
})

# Adding the dummy level to the mapping table for levels without the candidate
# The best to fill them manually with proper candidates, if possible
trans2 <- rbind(trans, data.frame(
  old = "no_cat",
  new = setdiff(occup_new$code, trans$new)
))
# Forward mapping for the case with two periods
occup_simple_forward <- cat2cat::cat2cat(
  data = list(
    old = occup_old,
    new = occup_new,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(
    trans = trans2,
    direction = "forward"
  )
)

testthat::test_that("Simple cat2cat example - two periods - forward", {
  testthat::expect_true(all(names(occup_simple_forward) %in% c("old", "new")))
  testthat::expect_true(is.data.frame(occup_simple_forward$old))
  testthat::expect_true(is.data.frame(occup_simple_forward$new))
})

testthat::test_that("Table with number of replications for both mapping  directions", {
  # Build number of observations before and after unification table
  res <- data.frame(
    `before_mapping` = c(nrow(occup_old), nrow(occup_new)),
    `after_mapping` = c(
      paste0(
        nrow(occup_simple_backward$old),
        " (nonzero ",
        sum(occup_simple_backward$old$wei_freq_c2c > 0),
        ")"
      ),
      paste0(
        nrow(occup_simple_forward$new),
        " (nonzero ",
        sum(occup_simple_forward$new$wei_freq_c2c > 0),
        ")"
      )
    )
  )
  rownames(res) <- c("old (backward)", "new (forward)")
  res_tab <- knitr::kable(
    res,
    "latex",
    caption = "Number of observations before and after unification."
  )
  testthat::expect_s3_class(res_tab, "knitr_kable")
  testthat::expect_match(res_tab, "before\\\\_mapping")
  testthat::expect_match(res_tab, "after\\\\_mapping")
  testthat::expect_match(res_tab, "old \\(backward\\).*227662")
  testthat::expect_match(res_tab, "new \\(forward\\).*18680")
})

# Set the seed as e.g., randomForest is used
set.seed(1234)
# Statistical models setup
# It could be shared for different iterations for this scenario
ml_setup <- list(
  data = rbind(occup_2010, occup_2012),
  cat_var = "code",
  method = c("knn", "rf"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10)
)
# Use the cat2cat procedure to map 2010 to 2008
occup_back_2008_2010 <- cat2cat::cat2cat(
  data = list(
    old = occup_2008,
    new = occup_2010,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)
# Use the cat2cat procedure to map 2008 to 2006
occup_back_2006_2008 <- cat2cat::cat2cat(
  data = list(
    old = occup_2006,
    new = occup_back_2008_2010$old,
    cat_var_new = "g_new_c2c",
    cat_var_old = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)
# Select a proper dataset for each year
o_2006 <- occup_back_2006_2008$old
o_2008 <- occup_back_2008_2010$old # or occup_back_2006_2008$new
o_2010 <- occup_back_2008_2010$new
# Add default cat2cat procedure columns to not processed dataset
o_2012 <- cat2cat::dummy_c2c(occup_2012, cat_var = "code", ml = c("knn", "rf"))
# Combine datasets
final_data_back_ml <- do.call(rbind, list(o_2006, o_2008, o_2010, o_2012))

testthat::test_that("Backward mapping, with four periods, one mapping table, and ml models", {
  testthat::expect_true(is.data.frame(final_data_back_ml))
  testthat::expect_identical(
    dim(final_data_back_ml),
    c(475543L, 19L)
  )
  testthat::expect_identical(
    colnames(final_data_back_ml),
    c(
      "id", "age", "sex", "edu", "exp", "district", "parttime", "salary",
      "code", "multiplier", "year", "code4", "index_c2c", "g_new_c2c",
      "wei_freq_c2c", "rep_c2c", "wei_naive_c2c", "wei_knn_c2c", "wei_rf_c2c"
    )
  )
})

testthat::test_that("Correlations between ml methods", {
  corr_separate <-
    dplyr::filter(
      dplyr::do(
        dplyr::select(
          dplyr::select(
            dplyr::group_by(
              dplyr::filter(final_data_back_ml, rep_c2c >= 10),
              index_c2c
            ),
            matches("wei.*c2c")
          ),
          -wei_naive_c2c
        ),
        corr = tryCatch(
          cor(.[, -1]),
          error = function(e) NA, warning = function(w) NA
        )
      ), !any(is.na(corr))
    )
  corr_table <- Reduce("+", corr_separate$corr) / length(corr_separate$corr)

  testthat::expect_true(inherits(corr_table, "matrix"))
  testthat::expect_equal(dim(corr_table), c(3L, 3L))
})

testthat::test_that("Counts for a few random levels in the unified variable over time", {
  data_count_plot <- cat2cat::prune_c2c(df = final_data_back_ml)
  data_count_plot2 <-
    dplyr::summarise_all(
      dplyr::group_by(
        dplyr::select(
          dplyr::mutate(
            dplyr::filter(
              data_count_plot,
              g_new_c2c %in% c("261102", "325502", "352111")
            ),
            g_new_c2c_nams = forcats::fct_recode(
              g_new_c2c,
              `OHS Inspector` = "325502",
              `Sound Engineer` = "352111",
              `Prosecutor` = "261102"
            )
          ),
          "wei_freq_c2c", "wei_rf_c2c", "year", "g_new_c2c_nams", -"g_new_c2c"
        ),
        g_new_c2c_nams, year
      ),
      sum
    )
  # Build counts across a 3 random categories table
  counts_base <-
    tidyr::pivot_wider(
      data_count_plot2,
      names_from = "year",
      values_from = c("wei_freq_c2c", "wei_rf_c2c"),
      names_sep = " "
    )
  res_tab <- dplyr::select(
    counts_base,
    g_new_c2c_nams,
    `rf 2006` = `wei_rf_c2c 2006`,
    `freq 2006` = `wei_freq_c2c 2006`,
    `rf 2008` = `wei_rf_c2c 2008`,
    `freq 2008` = `wei_freq_c2c 2008`,
    `2010` = `wei_rf_c2c 2010`,
    `2012` = `wei_rf_c2c 2012`
  )

  testthat::expect_true(inherits(res_tab, "data.frame"))
  testthat::expect_equal(dim(res_tab), c(3L, 7L))
})

# Mincerian-like regression formula
formula_micer <- I(log(salary)) ~ sex + parttime + edu + exp + I(exp**2)

testthat::test_that("Regression - neutral impact of the unified variable", {
  # Fit a weighted linear regression on replicated
  lms_replicated <- lm(
    formula = formula_micer,
    data = final_data_back_ml,
    weights = multiplier * wei_freq_c2c
  )
  # Adjust size of stds as the replication process enlarges degrees of freedom
  lms_replicated$df.residual <- nrow(occup) - length(lms_replicated$assign)
  # Fit a linear regression on original dataset
  lms_original <- lm(
    formula = formula_micer,
    data = occup,
    weights = multiplier
  )

  summary_replicated <- suppressWarnings(summary(lms_replicated))
  summary_original <- summary(lms_original)

  testthat::expect_equal(
    summary_replicated$coefficients,
    summary_original$coefficients,
    tolerance = 1e-8
  )
  testthat::expect_equal(
    summary_replicated$r.squared,
    summary_original$r.squared,
    tolerance = 1e-8
  )
})

testthat::test_that("Regression for each level in the unified variable", {
  # Separate regression for each occupational group under newest classification
  regression_sep <-
    dplyr::filter(
      dplyr::do(
        dplyr::filter(
          dplyr::mutate(
            dplyr::group_by(
              cat2cat::prune_c2c(df = final_data_back_ml, method = "nonzero"),
              g_new_c2c
            ),
            n = dplyr::n()
          ),
          n >= 30
        ),
        lm = tryCatch(
          lm(formula_micer, .data, weights = multiplier * wei_freq_c2c),
          error = function(e) NULL
        )
      ),
      !is.null(lm)
    )
  # Regression results for the first group
  summary_group <- summary(regression_sep$lm[[1]])

  testthat::expect_equal(
    round(summary_group$coefficients, 2),
    structure(
      c(
        9.5, 0.52, 1.87, -0.25, 0, 0, 0.46, 0.1, 0.5, 0.07,
        0.02, 0, 20.63, 5.17, 3.77, -3.55, 0.08, 0.29, 0, 0, 0, 0, 0.94,
        0.77
      ),
      dim = c(6L, 4L),
      dimnames = list(
        c("(Intercept)", "sexTRUE", "parttime", "edu", "exp", "I(exp^2)"),
        c("Estimate", "Std. Error", "t value", "Pr(>|t|)")
      )
    )
  )
})

additional_cols <- c(
  "index_c2c", "g_new_c2c", "wei_freq_c2c",
  "rep_c2c", "wei_naive_c2c"
)

testthat::test_that("dummy_c2c", {
  expect_identical(
    c(colnames(airquality), additional_cols),
    colnames(dummy_c2c(airquality, "Month"))
  )
})

testthat::test_that("dummy_c2c", {
  expect_identical(
    c(colnames(airquality), additional_cols, "wei_knn_c2c"),
    colnames(dummy_c2c(airquality, "Month", ml = "knn"))
  )
})

testthat::test_that("dummy_c2c backward", {
  expect_identical(
    c(colnames(airquality), additional_cols, "wei_knn_c2c"),
    colnames(dummy_c2c(airquality, "Month", ml = "wei_knn_c2c"))
  )
})

testthat::test_that("dummy_c2c with nb method", {
  expect_identical(
    c(colnames(airquality), additional_cols, "wei_nb_c2c"),
    colnames(dummy_c2c(airquality, "Month", ml = "nb"))
  )
})

data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup[occup$year == 2008, ]
occup_new <- occup[occup$year == 2010, ]

testthat::test_that("prune_c2c on cat2cat output", {
  c2c <- cat2cat(
    data = list(
      old = occup_old,
      new = occup_new,
      cat_var = "code",
      time_var = "year",
      multiplier_var = "multiplier"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  expect_equal(sum(c2c$old$wei_freq_c2c), nrow(occup_old))
  expect_true(all(c2c$old$wei_freq_c2c <= 1 & c2c$old$wei_freq_c2c >= 0))
  expect_identical(
    nrow(occup_old),
    c2c$old %>%
      prune_c2c(method = "highest1") %>%
      nrow()
  )

  expect_identical(
    sum(c2c$old$wei_freq_c2c > 0),
    c2c$old %>%
      prune_c2c(method = "nonzero") %>%
      nrow()
  )

  expect_true(
    (c2c$old %>%
      prune_c2c(method = "highest") %>%
      nrow()) >=
      (c2c$old %>%
        prune_c2c(method = "highest1") %>%
        nrow())
  )

  expect_equal(
    nrow(occup_old),
    c2c$old %>%
      prune_c2c(method = "morethan", percent = 0.2) %>%
      pull("wei_freq_c2c") %>%
      sum()
  )
})

# Tests for id_var functionality with occup_panel dataset
# and other features demonstrated in vignettes

data("occup_panel", package = "cat2cat")
data("trans", package = "cat2cat")

# Split by encoding periods (2009 = old encoding, 2010 = new encoding)
panel_old <- occup_panel[occup_panel$year == 2009, ]
panel_new <- occup_panel[occup_panel$year == 2010, ]

# For quarterly granularity tests
panel_2009Q4 <- occup_panel[occup_panel$quarter == "2009Q4", ]
panel_2010Q1 <- occup_panel[occup_panel$quarter == "2010Q1", ]

# Subjects in both encoding periods
shared_ids <- intersect(panel_old$panel_id, panel_new$panel_id)

# -----------------------------------------------------------------------------
# id_var basic functionality tests
# -----------------------------------------------------------------------------

testthat::test_that("id_var direct matching reduces replication", {
  # Without id_var
  result_no_id <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  # With id_var
  result_with_id <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  # With id_var should have fewer rows due to direct matching
  expect_lt(nrow(result_with_id$old), nrow(result_no_id$old))

  # But weighted sums should equal original counts
  expect_equal(sum(result_no_id$old$wei_freq_c2c), nrow(panel_2009Q4))
  expect_equal(sum(result_with_id$old$wei_freq_c2c), nrow(panel_2009Q4))
})

testthat::test_that("id_var direct matches have rep_c2c = 1", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  # Get direct matched subjects
  crossing_ids <- intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id)
  direct_matches <- result$old[result$old$panel_id %in% crossing_ids, ]

  # Direct matches should have rep_c2c = 1 (no replication)
  expect_true(all(direct_matches$rep_c2c == 1))

  # Direct matches should have wei_freq_c2c = 1
  expect_true(all(direct_matches$wei_freq_c2c == 1))
})

testthat::test_that("id_var non-matched subjects still get replicated", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  crossing_ids <- intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id)
  non_matched <- result$old[!result$old$panel_id %in% crossing_ids, ]

  # Non-matched subjects should have some with rep_c2c > 1
  expect_true(any(non_matched$rep_c2c > 1))
})

testthat::test_that("id_var g_new_c2c comes from base period for direct matches", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  crossing_ids <- intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id)

  # For each direct match, g_new_c2c should equal their 2010Q1 code
  for (pid in head(crossing_ids, 10)) {
    matched_row <- result$old[result$old$panel_id == pid, ]
    new_code <- panel_2010Q1$code[panel_2010Q1$panel_id == pid]
    expect_equal(matched_row$g_new_c2c, new_code)
  }
})

testthat::test_that("id_var complete matched panel uses only direct matches", {
  complete_ids <- intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id)
  complete_old <- panel_2009Q4[panel_2009Q4$panel_id %in% complete_ids, ]
  complete_new <- panel_2010Q1[panel_2010Q1$panel_id %in% complete_ids, ]

  result <- cat2cat(
    data = list(
      old = complete_old,
      new = complete_new,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  # No observations should go through the replication path when every id matches.
  expect_equal(nrow(result$old), nrow(complete_old))
  expect_equal(nrow(result$new), nrow(complete_new))
  expect_true(all(result$old$rep_c2c == 1))
  expect_true(all(result$new$rep_c2c == 1))
  expect_true(all(result$old$wei_freq_c2c == 1))
  expect_true(all(result$old$wei_naive_c2c == 1))

  new_codes <- complete_new$code[match(result$old$panel_id, complete_new$panel_id)]
  expect_equal(result$old$g_new_c2c, new_codes)
})

# -----------------------------------------------------------------------------
# Weight sum preservation tests
# -----------------------------------------------------------------------------

testthat::test_that("weights sum to original n for all weight types", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  n_original <- nrow(panel_2009Q4)
  expect_equal(sum(result$old$wei_freq_c2c), n_original)
  expect_equal(sum(result$old$wei_naive_c2c), n_original)
})

testthat::test_that("per-subject weights sum to 1", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  # Group by index_c2c (original observation) and sum weights
  per_subject_freq <- tapply(result$old$wei_freq_c2c,
                              result$old$index_c2c, sum)
  per_subject_naive <- tapply(result$old$wei_naive_c2c,
                               result$old$index_c2c, sum)

  # All should sum to 1 (within floating point tolerance)
  expect_true(all(abs(per_subject_freq - 1) < 1e-6))
  expect_true(all(abs(per_subject_naive - 1) < 1e-6))
})

# -----------------------------------------------------------------------------
# prune_c2c tests
# -----------------------------------------------------------------------------

testthat::test_that("prune_c2c highest1 returns one row per original obs", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  pruned <- prune_c2c(result$old, method = "highest1")
  expect_equal(nrow(pruned), nrow(panel_2009Q4))
})

testthat::test_that("prune_c2c nonzero removes zero-weight rows", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  pruned <- prune_c2c(result$old, method = "nonzero")

  # Should have no zero weights
  expect_true(all(pruned$wei_freq_c2c > 0))

  # Fewer or equal rows

  expect_lte(nrow(pruned), nrow(result$old))
})

testthat::test_that("prune_c2c preserves weight sum after renormalization", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  pruned <- prune_c2c(result$old, method = "highest")

  # Weight sum should still equal original n
  expect_equal(sum(pruned$wei_freq_c2c), nrow(panel_2009Q4))
})

# -----------------------------------------------------------------------------
# cross_c2c ensemble tests
# -----------------------------------------------------------------------------

testthat::test_that("cross_c2c creates ensemble weight column", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  crossed <- cross_c2c(result$old,
                        c("wei_freq_c2c", "wei_naive_c2c"),
                        c(0.5, 0.5))

  expect_true("wei_cross_c2c" %in% names(crossed))
})

testthat::test_that("cross_c2c weights sum correctly", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  crossed <- cross_c2c(result$old,
                        c("wei_freq_c2c", "wei_naive_c2c"),
                        c(0.7, 0.3))

  # Ensemble weights should sum to original n
  expect_equal(sum(crossed$wei_cross_c2c), nrow(panel_2009Q4))
})

testthat::test_that("cross_c2c weights are between 0 and 1", {
  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward")
  )

  crossed <- cross_c2c(result$old,
                        c("wei_freq_c2c", "wei_naive_c2c"),
                        c(0.5, 0.5))

  expect_true(all(crossed$wei_cross_c2c >= 0))
  expect_true(all(crossed$wei_cross_c2c <= 1))
})

# -----------------------------------------------------------------------------
# Quarterly panel structure validation
# -----------------------------------------------------------------------------

testthat::test_that("occup_panel has expected BAEL-style structure", {
  # 8 quarters
  expect_equal(length(unique(occup_panel$quarter)), 8)

  # Subjects observed across encoding change
  crossing <- intersect(
    occup_panel$panel_id[occup_panel$quarter == "2009Q4"],
    occup_panel$panel_id[occup_panel$quarter == "2010Q1"]
  )
  expect_gt(length(crossing), 400)  # Should be ~450

  # Most subjects observed 4 quarters
  appearances <- table(table(occup_panel$panel_id))
  expect_true("4" %in% names(appearances))
  expect_gt(appearances["4"], 500)  # Should be ~750
})

testthat::test_that("occup_panel codes have correct length per encoding", {
  old_codes <- occup_panel$code[occup_panel$year == 2009]
  new_codes <- occup_panel$code[occup_panel$year == 2010]

  # Old encoding: 4-digit
  expect_true(all(nchar(old_codes) == 4))

  # New encoding: 6-digit
  expect_true(all(nchar(new_codes) == 6))
})

testthat::test_that("boundary transitions are valid per trans table", {
  p2009Q4 <- occup_panel[occup_panel$quarter == "2009Q4",
                          c("panel_id", "code")]
  p2010Q1 <- occup_panel[occup_panel$quarter == "2010Q1",
                          c("panel_id", "code")]

  # Merge to get transitions
  crossing <- merge(p2009Q4, p2010Q1, by = "panel_id",
                     suffixes = c("_old", "_new"))

  # Check each transition is valid per trans table
  valid <- mapply(function(o, n) n %in% trans$new[trans$old == o],
                  crossing$code_old, crossing$code_new)

  # All transitions should be valid
  expect_true(all(valid))
})

# -----------------------------------------------------------------------------
# Forward direction with id_var
# -----------------------------------------------------------------------------

testthat::test_that("id_var works with forward direction", {
  # Some 2010 codes don't have backward mapping (genuinely new occupations)
  # So we expect a warning about missing coverage
  result <- suppressWarnings(cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "forward")
  ))

  # Forward maps new to old, so $new is replicated
  # Some observations may be dropped due to missing mapping coverage
  expect_lte(sum(result$new$wei_freq_c2c), nrow(panel_2010Q1))
  expect_gt(sum(result$new$wei_freq_c2c), nrow(panel_2010Q1) - 10)

  # Direct matches should have rep_c2c = 1
  crossing_ids <- intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id)
  direct_matches <- result$new[result$new$panel_id %in% crossing_ids, ]
  expect_true(all(direct_matches$rep_c2c == 1))
})

# -----------------------------------------------------------------------------
# ML weights with id_var
# -----------------------------------------------------------------------------

testthat::test_that("id_var works with ML weights", {
  skip_if_not_installed("caret")
  skip_if_not_installed("randomForest")

  result <- cat2cat(
    data = list(
      old = panel_2009Q4,
      new = panel_2010Q1,
      id_var = "panel_id",
      cat_var = "code",
      time_var = "quarter"
    ),
    mappings = list(trans = trans, direction = "backward"),
    ml = list(
      data = panel_2010Q1,
      cat_var = "code",
      method = "knn",
      features = c("age", "sex", "edu", "exp", "salary"),
      args = list(k = 5)
    )
  )

  # ML weights should be added
  expect_true("wei_knn_c2c" %in% names(result$old))

  # ML weights sum to original n
  expect_equal(sum(result$old$wei_knn_c2c), nrow(panel_2009Q4))

  # Direct matches should have ML weight = 1 (no ML prediction needed)
  crossing_ids <- intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id)
  direct_matches <- result$old[result$old$panel_id %in% crossing_ids, ]
  expect_true(all(direct_matches$wei_knn_c2c == 1))
})

#' Occupational dataset
#'
#' @details occup dataset is an example of unbalance panel dataset.
#' This is a simulated data although there are applied a real world
#' characteristics from national statistical office survey.
#' The original survey is anonymous and take place every two years.
#' It is presenting a characteristics from randomly selected company and then
#' using k step procedure employees are chosen.
#'
#' @usage occup
#'
#' @format A data frame with around 70000 observations and 12 variables.
#' \describe{
#' \item{id}{	integer	id}
#' \item{age}{ numeric age of a subject}
#' \item{sex}{ numeric sex of a subject}
#' \item{edu}{
#'  integer edu level of education of a subject where lower means
#'  higher - 1 for at least master degree
#' }
#' \item{exp}{ numeric exp number of experience years for a subject}
#' \item{district}{ integer district}
#' \item{parttime}{
#'   numeric contract type regards time where 1 mean full-time
#'   (work a whole week)
#' }
#' \item{salary}{ numeric salary per year}
#' \item{code}{ character code - occupational code}
#' \item{multiplier}{
#'   numeric multiplier for the subject to reproduce a population - how many
#'   of such subjects in population
#' }
#' \item{year}{integer year}
#' \item{code4}{ character code - occupational code - first 4 digits}
#' }
#' @details occupational dataset
#'
"occup"

#' Occupational dataset - small one
#'
#' @details occup dataset is an example of unbalance panel dataset.
#' This is a simulated data although there are applied a real world
#' characteristics from national statistical office survey.
#' The original survey is anonymous and take place every two years.
#' It is presenting a characteristics from randomly selected company and
#' then using k step procedure employees are chosen.
#'
#' @usage occup_small
#'
#' @format A data frame with around 8000 observations and 12 variables.
#' \describe{
#' \item{id}{	integer	id}
#' \item{age}{ numeric age of a subject}
#' \item{sex}{ numeric sex of a subject}
#' \item{edu}{
#'   integer edu level of education of a subject where lower means
#'   higher - 1 for at least master degree
#' }
#' \item{exp}{ numeric exp number of experience years for a subject}
#' \item{district}{ integer district}
#' \item{parttime}{
#'   numeric contract type regards time where 1 mean full-time
#'   (work a whole week)
#' }
#' \item{salary}{ numeric salary per year}
#' \item{code}{ character code - occupational code}
#' \item{multiplier}{
#'   numeric multiplier for the subject to reproduce a population -
#'   how many of such subjects in population
#' }
#' \item{year}{integer year}
#' \item{code4}{ character code - occupational code - first 4 digits}
#' }
#' @details occupational dataset
#' @examples
#' set.seed(1234)
#' data("occup", package = "cat2cat")
#' occup_small <- occup[sort(sample(nrow(occup), 8000)), ]
"occup_small"

#' Occupational panel dataset with BAEL-style quarterly rotation
#'
#' @details A simulated rotational panel derived from \code{occup}, inspired by
#' the Polish BAEL (Badanie Aktywnosci Ekonomicznej Ludnosci - Labour Force
#' Survey). Unlike the main \code{occup} dataset (repeated cross-sections), this
#' dataset includes workers observed for 4 consecutive quarters, enabling
#' demonstration of the \code{id_var} feature in \code{cat2cat()}.
#'
#' Panel design:
#' \itemize{
#'   \item 8 quarters: 2009Q1 through 2010Q4
#'   \item Encoding change between 2009Q4 and 2010Q1
#'   \item Each cohort enters quarterly and stays for 4 consecutive quarters
#'   \item ~150 new subjects enter each quarter (1/4 rotation)
#'   \item ~450 subjects observed across the encoding change
#' }
#'
#' For subjects across quarters:
#' \itemize{
#'   \item \code{panel_id} is consistent across quarters
#'   \item Occupation code is preserved (or mapped via \code{trans} at encoding change)
#'   \item Age and experience increase annually (every 4 quarters)
#'   \item Salary varies slightly between quarters (-1\% to +2\%)
#' }
#'
#' @usage occup_panel
#'
#' @format A data frame with approximately 3900 observations and 15 variables.
#' \describe{
#' \item{id}{integer original row identifier from \code{occup}}
#' \item{panel_id}{integer consistent person identifier across quarters}
#' \item{cohort}{character entry quarter (e.g., "2009Q1")}
#' \item{age}{numeric age of the worker}
#' \item{sex}{numeric sex of the worker}
#' \item{edu}{integer education level (1 = highest)}
#' \item{exp}{numeric years of experience}
#' \item{district}{integer district code}
#' \item{parttime}{numeric contract type (1 = full-time)}
#' \item{salary}{numeric annual salary}
#' \item{code}{character occupational code (4-digit pre-2010, 6-digit post-2010)}
#' \item{multiplier}{numeric survey weight}
#' \item{quarter}{character survey quarter (e.g., "2009Q1", "2010Q2")}
#' \item{code4}{character first 4 digits of occupational code}
#' \item{year}{integer year extracted from quarter}
#' \item{quarter_num}{integer quarter number (1-4)}
#' }
#' @seealso \code{\link{occup}} for the full repeated cross-section dataset,
#'   \code{\link{trans}} for the transition table
#' @examples
#' data("occup_panel", package = "cat2cat")
#'
#' # Check panel structure
#' table(occup_panel$quarter)
#' table(table(occup_panel$panel_id))  # appearances per subject (target: 4)
#'
#' # Subjects observed across the encoding change (2009Q4 -> 2010Q1)
#' panel_2009Q4 <- occup_panel[occup_panel$quarter == "2009Q4", ]
#' panel_2010Q1 <- occup_panel[occup_panel$quarter == "2010Q1", ]
#' length(intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id))
"occup_panel"

#' trans dataset containing mappings (transitions) between
#' old (2008) and new (2010) occupational codes.
#' This table could be used to map encodings in both directions.
#'
#' @usage trans
#'
#' @format A data frame with 2693 observations and 2 variables.
#' \describe{
#' \item{old}{	character an old encoding of a certain occupation}
#' \item{new}{	character a new encoding of a certain occupation}
#' }
#' @details mapping (transition) table for occupations where first column
#' contains old encodings and second one a new encoding
#'
"trans"

#' verticals dataset
#'
#' @usage verticals
#'
#' @format A data frame with 21 observations and 4 variables.
#' \describe{
#' \item{vertical}{	character an certain sales vertical}
#' \item{sales}{	numeric a size of sale}
#' \item{counts}{	integer counts size}
#' \item{v_date}{	character Date}
#' }
#' @details random data - aggregate sales across e-commerce verticals
#' @examples
#' set.seed(1234)
#' agg_old <- data.frame(
#'   vertical = c(
#'     "Electronics", "Kids1", "Kids2", "Automotive", "Books",
#'     "Clothes", "Home", "Fashion", "Health", "Sport"
#'   ),
#'   sales = rnorm(10, 100, 10),
#'   counts = rgeom(10, 0.0001),
#'   v_date = rep("2020-04-01", 10), stringsAsFactors = FALSE
#' )
#'
#' agg_new <- data.frame(
#'   vertical = c(
#'     "Electronics", "Supermarket", "Kids", "Automotive1",
#'     "Automotive2", "Books", "Clothes", "Home", "Fashion", "Health", "Sport"
#'   ),
#'   sales = rnorm(11, 100, 10),
#'   counts = rgeom(11, 0.0001),
#'   v_date = rep("2020-05-01", 11), stringsAsFactors = FALSE
#' )
#' verticals <- rbind(agg_old, agg_new)
"verticals"

#' verticals2 dataset
#'
#' @usage verticals2
#'
#' @format A data frame with 202 observations and 4 variables.
#' \describe{
#' \item{ean}{ product ean}
#' \item{vertical}{	character an certain sales vertical}
#' \item{sales}{	numeric a size of sale}
#' \item{v_date}{	character Date}
#' }
#' @details random data - single products sales across e-commerce verticals
#' @examples
#' set.seed(1234)
#' vert_old <- data.frame(
#'   ean = 90000001:90000020,
#'   vertical = sample(c(
#'     "Electronics", "Kids1", "Kids2", "Automotive", "Books",
#'     "Clothes", "Home", "Fashion", "Health", "Sport"
#'   ), 20, replace = TRUE),
#'   sales = rnorm(20, 100, 10),
#'   v_date = rep("2020-04-01", 20), stringsAsFactors = FALSE
#' )
#'
#' vert_old2 <- data.frame(
#'   ean = 90000021:90000100,
#'   vertical = sample(c(
#'     "Electronics", "Kids1", "Kids2", "Automotive", "Books",
#'     "Clothes", "Home", "Fashion", "Health", "Sport"
#'   ), 80, replace = TRUE),
#'   sales = rnorm(80, 100, 10),
#'   v_date = rep("2020-04-01", 80), stringsAsFactors = FALSE
#' )
#'
#' vert_new <- vert_old2
#' vert_new$sales <- rnorm(nrow(vert_new), 80, 10)
#' vert_new$v_date <- "2020-05-01"
#' vert_new$vertical[vert_new$vertical %in% c("Kids1", "Kids2")] <- "Kids"
#' vert_new$vertical[vert_new$vertical %in% c("Automotive")] <-
#'   sample(
#'     c("Automotive1", "Automotive2"),
#'     sum(vert_new$vertical %in% c("Automotive")),
#'     replace = TRUE
#'   )
#' vert_new$vertical[vert_new$vertical %in% c("Home")] <-
#'   sample(
#'     c("Home", "Supermarket"),
#'     sum(vert_new$vertical %in% c("Home")),
#'     replace = TRUE
#'   )
#'
#' vert_new2 <- data.frame(
#'   ean = 90000101:90000120,
#'   vertical = sample(
#'     c(
#'       "Electronics", "Supermarket", "Kids", "Automotive1",
#'       "Automotive2", "Books", "Clothes", "Home",
#'       "Fashion", "Health", "Sport"
#'     ), 20,
#'     replace = TRUE
#'   ),
#'   sales = rnorm(20, 100, 10),
#'   v_date = rep("2020-05-01", 20), stringsAsFactors = FALSE
#' )
#'
#' verticals2 <- rbind(
#'   rbind(vert_old, vert_old2),
#'   rbind(vert_new, vert_new2)
#' )
#' verticals2$vertical <- as.character(verticals2$vertical)
"verticals2"

#' The internal function used in the cat2cat one
#' @description apply the ml models to the cat2cat data
#' @param ml `list` the same `ml` argument as provided to `cat2cat` function.
#' @param mapp `list` a mapping table
#' @param target_data `data.frame`
#' @param cat_var_target `character(1)` name of the categorical variable
#' in the target period.
#' @keywords internal
cat2cat_ml <- function(ml, mapp, target_data, cat_var_target) {
  encoded <- encode_factor_features(ml, target_data)
  ml <- encoded$ml
  target_data <- encoded$target_data

  ml <- validate_ml(ml)

  stopifnot(
    "All `ml$features` must be columns in target_data" =
      all(ml$features %in% colnames(target_data))
  )
  stopifnot(
    "`cat_var_target` must be a column in target_data" =
      cat_var_target %in% colnames(target_data)
  )

  stopifnot(
    "All `ml$features` columns must be numeric or logical" =
      all(vapply(
        target_data[, ml$features, drop = FALSE],
        function(x) is.numeric(x) || is.logical(x), logical(1)
      ))
  )

  features <- unique(ml$features)
  methods <- unique(ml$method)
  ml_names <- paste0("wei_", methods, "_c2c")

  target_data[ml_names] <- rep(list(NA_real_), length(ml_names))

  cat_ml_year_g <- split(
    ml$data[, c(features, ml$cat_var), drop = FALSE],
    factor(ml$data[[ml$cat_var]], exclude = NULL)
  )
  target_data_cats <- target_data[[cat_var_target]]
  target_data_cat_c2c <- split(
    target_data,
    factor(target_data_cats, exclude = NULL)
  )

  for (cat in unique(names(target_data_cat_c2c))) {
    matched_cat <- match(cat, names(target_data_cat_c2c))
    target_data_cat <- target_data_cat_c2c[[matched_cat]]

    dis <- try(do.call(rbind, cat_ml_year_g[mapp[[match(cat, names(mapp))]]]), silent = TRUE)
    if (inherits(dis, "try-error")) next

    udc <- unique(dis[[ml$cat_var]])

    if (length(udc) <= 1) {
      target_data_cat_c2c[[matched_cat]][ml_names] <-
        target_data_cat$wei_freq_c2c
      next
    }

    if (
      !(length(unique(target_data_cat$g_new_c2c)) > 1 &&
          length(udc) >= 1 &&
          nrow(target_data_cat) > 0 &&
          any(unique(target_data_cat$g_new_c2c) %in% names(cat_ml_year_g)))
    ) {
      next
    }

    base_ml <-
      target_data_cat[
        !duplicated(target_data_cat[["index_c2c"]]),
        c("index_c2c", features)
      ]
    cc <- complete.cases(base_ml[, features])

    for (m in methods) {
      ml_name <- paste0("wei_", m, "_c2c")

      try(
        {
          if (m == "knn") {
            group_prediction <- suppressWarnings(
              caret::knn3(
                x = dis[, features, drop = FALSE],
                y = factor(dis[[ml$cat_var]]),
                k = min(ml$args$k, ceiling(nrow(dis) / 4))
              )
            )
            pp <- as.data.frame(
              stats::predict(
                group_prediction,
                base_ml[cc, features, drop = FALSE],
                type = "prob"
              )
            )
          } else if (m == "rf") {
            group_prediction <- suppressWarnings(
              randomForest::randomForest(
                y = factor(dis[[ml$cat_var]]),
                x = dis[, features, drop = FALSE],
                ntree = min(ml$args$ntree, 100)
              )
            )
            pp <- as.data.frame(
              stats::predict(
                group_prediction,
                base_ml[cc, features, drop = FALSE],
                type = "prob"
              )
            )
          } else if (m == "lda") {
            group_prediction <- suppressWarnings(
              MASS::lda(
                grouping = factor(dis[[ml$cat_var]]),
                x = as.matrix(dis[, features, drop = FALSE])
              )
            )
            pp <- as.data.frame(
              stats::predict(
                group_prediction,
                as.matrix(base_ml[cc, features, drop = FALSE])
              )$posterior
            )
          } else if (m == "nb") {
            group_prediction <- suppressWarnings(
              e1071::naiveBayes(
                x = dis[, features, drop = FALSE],
                y = factor(dis[[ml$cat_var]])
              )
            )
            pp <- as.data.frame(
              stats::predict(
                group_prediction,
                base_ml[cc, features, drop = FALSE],
                type = "raw"
              )
            )
          }
          ll <- setdiff(unique(target_data_cat$g_new_c2c), colnames(pp))
          # imputing rest of the class to zero prob
          if (length(ll)) {
            pp[ll] <- 0
          }
          pp_stack <- utils::stack(pp)
          pp[["index_c2c"]] <- base_ml[["index_c2c"]][cc]
          res <- cbind(pp_stack, index_c2c = rep(pp$index_c2c, ncol(pp) - 1))
          colnames(res) <- c("val", "g_new_c2c", "index_c2c")
          ress <- merge(
            target_data_cat[, c("index_c2c", "g_new_c2c")],
            res,
            by = c("index_c2c", "g_new_c2c"),
            all.x = TRUE,
            sort = FALSE
          )
          resso <- ress[order(ress$index_c2c), ]
          target_data_cat_c2c[[matched_cat]][[ml_name]] <- resso$val
        },
        silent = TRUE
      )
    }
  }

  target_data <- do.call(rbind, target_data_cat_c2c)
  target_data <- target_data[order(target_data[["index_c2c"]]), ]
  grp_id <- match(target_data[["index_c2c"]], unique(target_data[["index_c2c"]]))

  if (ml$on_fail == "freq") {
    fallback <- target_data[["wei_freq_c2c"]]
  } else if (ml$on_fail == "naive") {
    fallback <- target_data[["wei_naive_c2c"]]
  } else {
    fallback <- rep(NA_real_, nrow(target_data))
  }

  for (ml_name in ml_names) {
    col <- target_data[[ml_name]]
    failed <- is.na(col) | !is.finite(col)

    if (any(failed)) {
      n_rows <- sum(failed)
      n_obs <- length(unique(target_data[["index_c2c"]][failed]))
      total_rows <- length(col)
      total_obs <- length(unique(target_data[["index_c2c"]]))
      pct_rows <- 100 * n_rows / total_rows
      pct_obs <- 100 * n_obs / total_obs

      if (ml$on_fail == "error") {
        stop(sprintf(
          paste0(
            "ML weights failed for method '%s': ",
            "%.1f%% rows (%d/%d) and %.1f%% observations (%d/%d)."
          ),
          sub("^wei_(.*)_c2c$", "\\1", ml_name),
          pct_rows, n_rows, total_rows,
          pct_obs, n_obs, total_obs
        ))
      }

      if (isTRUE(ml$fail_warn)) {
        warning(sprintf(
          paste0(
            "ML weights failed for method '%s': ",
            "%.1f%% rows (%d/%d) and %.1f%% observations (%d/%d); ",
            "on_fail = '%s' was applied."
          ),
          sub("^wei_(.*)_c2c$", "\\1", ml_name),
          pct_rows, n_rows, total_rows,
          pct_obs, n_obs, total_obs,
          ml$on_fail
        ), call. = FALSE)
      }

      col[failed] <- fallback[failed]
    }

    sum_by_group <- as.numeric(rowsum(col, grp_id, reorder = FALSE)[, 1])
    scale_factor <- sum_by_group[grp_id]
    can_scale <- is.finite(scale_factor) & scale_factor > 0
    col[can_scale] <- col[can_scale] / scale_factor[can_scale]
    target_data[[ml_name]] <- col
  }

  list(target_data = target_data)
}

# " Validate cat2cat ml
#' @keywords internal
validate_ml <- function(ml) {
  if (is.null(ml$on_fail)) ml$on_fail <- "freq"
  ml$on_fail <- tolower(ml$on_fail)
  if (is.null(ml$fail_warn)) ml$fail_warn <- TRUE

  stopifnot(
    "`ml` must contain 'method', 'features', and 'data'" =
      all(c("method", "features", "data") %in% names(ml))
  )
  stopifnot(
    "`ml$method` must be one or more of: 'knn', 'rf', 'lda', 'nb'" =
      all(ml$method %in% c("knn", "rf", "lda", "nb"))
  )

  stopifnot(
    "`ml$on_fail` must be one of: 'freq', 'naive', 'na', 'error'" =
      length(ml$on_fail) == 1 && ml$on_fail %in% c("freq", "naive", "na", "error")
  )
  stopifnot(
    "`ml$fail_warn` must be TRUE or FALSE" =
      is.logical(ml$fail_warn) && length(ml$fail_warn) == 1 && !is.na(ml$fail_warn)
  )

  if ("rf" %in% ml$method) {
    delayed_package_load(
      "randomForest",
      sprintf("Please install %s package to use the %s model in the cat2cat function.", "randomForest", "rf")
    )
  }

  if ("knn" %in% ml$method) {
    delayed_package_load(
      "caret",
      sprintf("Please install %s package to use the %s model in the cat2cat function.", "caret", "knn")
    )
  }

  if ("nb" %in% ml$method) {
    delayed_package_load(
      "e1071",
      sprintf("Please install %s package to use the %s model in the cat2cat function.", "e1071", "nb")
    )
  }

  stopifnot(
    "`ml$cat_var` must be a column in `ml$data`" =
      ml$cat_var %in% colnames(ml$data)
  )
  stopifnot(
    "All `ml$features` must be columns in `ml$data`" =
      all(ml$features %in% colnames(ml$data))
  )
  stopifnot(
    "All `ml$features` columns in `ml$data` must be numeric or logical" =
      all(vapply(
        ml$data[, ml$features, drop = FALSE],
        function(x) is.numeric(x) || is.logical(x), logical(1)
      ))
  )

  ml
}

# " Delayed load of a package
#' @keywords internal
delayed_package_load <- function(package, msg = sprintf("Please install %s package.", package)) {
  if (isFALSE(suppressPackageStartupMessages(requireNamespace(package, quietly = TRUE)))) {
    stop(msg)
  }
}

# " Normalized multiclass Brier score
#' @keywords internal
brier_score <- function(prob_matrix, true_cats, classes = colnames(prob_matrix)) {
  prob_matrix <- as.data.frame(prob_matrix)
  classes <- as.character(classes)
  true_cats <- as.character(true_cats)

  missing_classes <- setdiff(classes, colnames(prob_matrix))
  if (length(missing_classes)) {
    prob_matrix[missing_classes] <- 0
  }
  prob_matrix <- as.matrix(prob_matrix[, classes, drop = FALSE])

  truth <- matrix(0, nrow = length(true_cats), ncol = length(classes))
  colnames(truth) <- classes
  matched <- match(true_cats, classes)
  truth[cbind(seq_along(true_cats), matched)] <- 1

  mean(rowSums((prob_matrix - truth)^2) / 2)
}

# " One-hot encode factor features in `ml$data` and `target_data`
#' @description Replaces any `factor` columns listed in `ml$features` with
#' 0/1 indicator columns built from the union of levels observed in either
#' dataset. Numeric/logical features are left unchanged. Character columns
#' are not auto-encoded; convert them to `factor` explicitly.
#' @keywords internal
#' @noRd
encode_factor_features <- function(ml, target_data) {
  feats <- ml$features
  is_factorish <- function(x) is.factor(x)

  to_encode <- feats[vapply(feats, function(f) {
    (f %in% colnames(ml$data) && is_factorish(ml$data[[f]])) ||
      (f %in% colnames(target_data) && is_factorish(target_data[[f]]))
  }, logical(1))]

  if (length(to_encode) == 0) {
    return(list(ml = ml, target_data = target_data))
  }

  new_features <- setdiff(feats, to_encode)
  for (f in to_encode) {
    train_vals <- if (f %in% colnames(ml$data)) as.character(ml$data[[f]]) else character(0)
    target_vals <- if (f %in% colnames(target_data)) as.character(target_data[[f]]) else character(0)
    lv <- unique(c(train_vals, target_vals))
    lv <- lv[!is.na(lv)]
    for (l in lv) {
      col_name <- paste0(f, "_", l)
      ml$data[[col_name]] <- as.integer(
        !is.na(ml$data[[f]]) & as.character(ml$data[[f]]) == l
      )
      target_data[[col_name]] <- as.integer(
        !is.na(target_data[[f]]) & as.character(target_data[[f]]) == l
      )
      new_features <- c(new_features, col_name)
    }
  }
  ml$features <- new_features
  list(ml = ml, target_data = target_data)
}

#' Cross-validation diagnostics for cat2cat ML models
#'
#' @description
#' Evaluates whether machine-learning models used in the \code{ml} argument of
#' \code{\link{cat2cat}} actually improve category assignment over simpler
#' baselines. The function runs a per-group train/test split across every
#' mapping group defined by the transition table.
#'
#' @details
#' For each mapping group (set of candidate categories linked by the transition
#' table) the function:
#' \enumerate{
#'   \item Collects all observations from \code{ml$data} whose category belongs
#'         to the group.
#'   \item Randomly splits them into training (\code{1 - test_prop}) and test
#'         (\code{test_prop}) sets.
#'   \item Computes two baselines on the test set:
#'         \itemize{
#'           \item \strong{naive}: accuracy of a random guess (\eqn{1 / k}
#'                 where \eqn{k} is the number of candidate categories).
#'           \item \strong{freq}: accuracy of always predicting the most
#'                 frequent category in the training set.
#'         }
#'   \item Trains each ML model specified in \code{ml$method} on the training
#'         set and records its classification accuracy on the test set.
#' }
#'
#' Groups with fewer than 5 observations or only one candidate category are
#' skipped (their accuracy is recorded as \code{NA}).
#'
#' \subsection{Baseline-Only Diagnostics}{
#' To inspect only baseline diagnostics (\code{naive}, \code{freq}, and their
#' Brier/mean-probability variants), pass empty model and feature vectors:
#' \code{ml$method = character(0)} and \code{ml$features = character(0)}.
#' In this mode, no ML models are trained, but baseline diagnostics are still
#' computed for each mapping group.
#' }
#'
#' \subsection{Understanding the Metrics}{
#' Three complementary metrics evaluate model quality:
#'
#' \strong{Accuracy} measures how often the model's top prediction matches
#' the true category. Use this when you only care about the single most likely
#' assignment. Higher is better; theoretical maximum is 1.0.
#'
#' \strong{Mean P(true class)} measures the average probability the model
#' assigns to the correct category. This evaluates the full probability
#' distribution, not just the top prediction. For cat2cat, where weights ARE
#' probabilities distributed across candidates, this metric directly measures
#' weight quality. Higher is better; range is \eqn{[0, 1]}.
#'
#' \strong{Brier score} measures the squared error between predicted
#' probabilities and the one-hot encoded true outcome, normalized to
#' \eqn{[0, 1]}. Unlike log-loss, Brier score is bounded and does not explode
#' when P(true) is near zero. Lower is better; 0 means perfect prediction. For k
#' categories, the naive baseline (uniform 1/k) gives Brier = \eqn{(1 - 1/k) / 2}.
#' }
#'
#' \subsection{Choosing a Method}{
#' \itemize{
#'   \item If accuracy and mean P(true) are similar across methods, prefer
#'         simpler methods (freq, lda) over complex ones (rf, knn).
#'   \item If ML methods rarely beat the frequency baseline, use
#'         \code{wei_freq_c2c} --- ML adds complexity without benefit.
#'   \item If Brier score for ML is similar to or worse than naive, the model
#'         is not well-calibrated. Consider \code{wei_freq_c2c} or a different ML method.
#'   \item Use \code{\link{cross_c2c}} to ensemble multiple methods if no
#'         single method dominates.
#' }
#' }
#'
#' Because the split is random, results will vary between runs. For more
#' stable estimates, call the function several times or use a larger
#' \code{ml$data} set (e.g. pool multiple survey waves).
#'
#' @param ... additional options:
#' \describe{
#'   \item{\code{test_prop}}{\code{numeric(1)} proportion of observations
#'     held out for testing in each mapping group. Must be in \eqn{(0, 1)}.
#'     Default: \code{0.2}.}
#'   \item{\code{min_match}}{\code{numeric(1)} minimum fraction of
#'     \code{ml$data[[ml$cat_var]]} values that must appear in the transition
#'     table. If the match rate falls below this threshold, the function
#'     stops with an error --- usually indicating the wrong \code{direction}.
#'     Must be in \eqn{[0, 1)}. Default: \code{0.5}.}
#' }
#' @inheritParams cat2cat
#' @return An object of class \code{"cat2cat_ml_run"} (a named list). Each
#'   element corresponds to one mapping group and contains:
#'   \describe{
#'     \item{\code{naive}}{\code{numeric(1)} --- random-guess baseline accuracy
#'       (1/k where k is number of categories). Theoretical lower bound.}
#'     \item{\code{freq}}{\code{numeric(1)} --- most-frequent-category baseline
#'       accuracy. A simple but often strong baseline.}
#'     \item{\code{acc}}{Named \code{numeric} vector --- test-set accuracy for
#'       each ML method. Higher is better; compare to \code{freq}.}
#'     \item{\code{brier}}{Named \code{numeric} vector --- Brier score
#'       for each ML method, computed from the full probability vector.
#'       Lower is better; range is [0, 1].}
#'     \item{\code{mean_prob}}{Named \code{numeric} vector --- average probability
#'       assigned to the true class. Higher is better. This directly measures
#'       the quality of probability weights used by cat2cat.}
#'     \item{\code{naive_brier}}{\code{numeric(1)} --- Brier score for uniform
#'       baseline (= (1 - 1/k) / 2). Serves as a calibration reference.}
#'     \item{\code{naive_mean_prob}}{\code{numeric(1)} --- mean P(true) for
#'       uniform baseline (= 1/k). Equals \code{naive} by definition.}
#'     \item{\code{freq_brier}}{\code{numeric(1)} --- Brier score using training
#'       set category frequencies as probabilities.}
#'     \item{\code{freq_mean_prob}}{\code{numeric(1)} --- mean P(true) using
#'       training set category frequencies.}
#'   }
#'   The object also carries an \code{"ml_models"} attribute listing the
#'   methods evaluated. Use \code{print()} for a human-readable summary.
#' @seealso \code{\link{cat2cat}} for the main mapping function,
#'   \code{\link{cross_c2c}} for ensembling weights from multiple methods.
#' @export
#' @rdname cat2cat_ml_run
#' @examples
#' \donttest{
#' library("cat2cat")
#' data("occup", package = "cat2cat")
#' data("trans", package = "cat2cat")
#'
#' occup_2006 <- occup[occup$year == 2006, ]
#' occup_2008 <- occup[occup$year == 2008, ]
#'
#' # Forward direction: old encoding -> new encoding
#' # Use data from OLD encoding periods (2006, 2008)
#' ml_setup <- list(
#'   data = rbind(occup_2006, occup_2008),
#'   cat_var = "code",
#'   method = c("knn", "rf", "lda", "nb"),
#'   features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#'   args = list(k = 10, ntree = 50)
#' )
#' mappings <- list(trans = trans, direction = "forward")
#'
#' set.seed(1234)
#' res <- cat2cat_ml_run(mappings, ml_setup, test_prop = 0.2)
#' print(res)
#'
#' # Typical good results show:
#' # - ML accuracy > freq baseline (ML adds value)
#' # - ML Brier < naive (well-calibrated probabilities)
#' # - ML mean P(true) > freq (better probability weights)
#' #
#' # If Brier(ML) >= Brier(naive), the model is poorly calibrated
#' # and wei_freq_c2c may be safer. Use cross_c2c() to ensemble.
#'
#' # High failure rate is normal - most groups have <5 observations
#'
#' # Baseline-only diagnostics (no ML models):
#' ml_baseline <- list(
#'   data = rbind(occup_2006, occup_2008),
#'   cat_var = "code",
#'   method = character(0),
#'   features = character(0)
#' )
#' baseline_cv <- cat2cat_ml_run(mappings, ml_baseline)
#' print(baseline_cv)
#' }
#'
cat2cat_ml_run <- function(mappings, ml, ...) {
  stopifnot("`ml` must be a list" = is.list(ml))
  stopifnot("`mappings` must be a list" = is.list(mappings))

  ml <- encode_factor_features(ml, ml$data)$ml

  elargs <- list(...)
  if (is.null(elargs$test_prop)) elargs$test_prop <- 0.2
  stopifnot(
    "`test_prop` must be a single number between 0 and 1" =
      length(elargs$test_prop) == 1 && elargs$test_prop > 0 && elargs$test_prop < 1
  )
  if (is.null(elargs$min_match)) elargs$min_match <- 0.5
  stopifnot(
    "`min_match` must be a single number between 0 and 1" =
      length(elargs$min_match) == 1 && elargs$min_match >= 0 && elargs$min_match < 1
  )

  validate_mappings(mappings)
  validate_ml(ml)

  if (mappings$direction == "forward") {
    base_name <- "old"
    target_name <- "new"
  } else if (mappings$direction == "backward") {
    base_name <- "new"
    target_name <- "old"
  }

  mapps <- get_mappings(mappings$trans)
  mapp <- mapps[[paste0("to_", base_name)]]

  cat_var <- ml$data[[ml$cat_var]]
  cat_var_vals <- unlist(mappings$trans[, base_name])
  if (sum(cat_var %in% cat_var_vals) / length(cat_var) < elargs$min_match) {
    stop(
      paste0(
        "There is no mappings to group the cat_var variable.\n",
        "Probably you should change the direction in the mappings argument.\n"
      )
    )
  }

  features <- unique(ml$features)
  methods <- unique(ml$method)

  train_g <- split(
    ml$data[, c(features, ml$cat_var), drop = FALSE],
    factor(ml$data[[ml$cat_var]], exclude = NULL)
  )

  res <- list()
  for (cat in names(mapp)) {
    try(
      {
        matched_cat <- mapp[[match(cat, names(mapp))]]
        cat_nam <- if (cat == "") " " else cat
        res[[cat_nam]] <- list(
          naive = NA_real_,
          acc = stats::setNames(rep(NA_real_, length(methods)), methods),
          freq = NA_real_,
          brier = stats::setNames(rep(NA_real_, length(methods)), methods),
          mean_prob = stats::setNames(rep(NA_real_, length(methods)), methods),
          naive_brier = NA_real_,
          naive_mean_prob = NA_real_,
          freq_brier = NA_real_,
          freq_mean_prob = NA_real_
        )

        data_small_g <- do.call(rbind, train_g[matched_cat])

        if (isTRUE(is.null(data_small_g) || nrow(data_small_g) < 5 ||
                   length(matched_cat) < 2 || sum(matched_cat %in% data_small_g[[ml$cat_var]]) == 1)) {
          next
        }

        n_categories <- length(matched_cat)
        res[[cat_nam]][["naive"]] <- 1 / n_categories
        # Naive baseline: uniform probability = 1/k for each category
        res[[cat_nam]][["naive_mean_prob"]] <- 1 / n_categories
        res[[cat_nam]][["naive_brier"]] <- (1 - 1 / n_categories) / 2

        index_tt <- sample(c(0, 1),
                           nrow(data_small_g),
                           prob = c(1 - elargs$test_prop, elargs$test_prop), replace = TRUE)
        data_test_small <- data_small_g[index_tt == 1, , drop = FALSE]
        data_train_small <- data_small_g[index_tt == 0, , drop = FALSE]
        if (length(features) == 0) {
          cc <- rep(TRUE, nrow(data_test_small))
        } else {
          cc <- complete.cases(data_test_small[, features])
        }

        if (isTRUE(nrow(data_test_small[cc, ]) == 0 || nrow(data_train_small) < 5)) {
          next
        }

        gcounts <- table(data_train_small[[ml$cat_var]])
        gfreq <- names(gcounts)[which.max(gcounts)]
        res[[cat_nam]][["freq"]] <- mean(gfreq == data_test_small[[ml$cat_var]])

        # Frequency baseline probabilities (train frequencies applied to test)
        train_freqs <- gcounts / sum(gcounts)
        true_cats_test <- data_test_small[[ml$cat_var]][cc]
        freq_probs <- vapply(true_cats_test, function(tc) {
          if (tc %in% names(train_freqs)) train_freqs[[tc]] else 0
        }, numeric(1))
        res[[cat_nam]][["freq_mean_prob"]] <- mean(freq_probs)
        freq_prob_matrix <- matrix(
          rep(0, length(true_cats_test) * length(matched_cat)),
          nrow = length(true_cats_test),
          dimnames = list(NULL, matched_cat)
        )
        freq_prob_matrix[, names(train_freqs)] <- rep(
          train_freqs,
          each = length(true_cats_test)
        )
        res[[cat_nam]][["freq_brier"]] <- brier_score(
          freq_prob_matrix,
          true_cats_test,
          matched_cat
        )

        for (m in methods) {
          try(
            {
              if (m == "knn") {
                group_prediction <- suppressWarnings(
                  caret::knn3(
                    x = data_train_small[, features, drop = FALSE],
                    y = factor(data_train_small[[ml$cat_var]]),
                    k = min(ml$args$k, ceiling(nrow(data_train_small) / 4))
                  )
                )
                pred <- stats::predict(
                  group_prediction,
                  data_test_small[cc, features, drop = FALSE],
                  type = "class"
                )
                prob_matrix <- stats::predict(
                  group_prediction,
                  data_test_small[cc, features, drop = FALSE],
                  type = "prob"
                )
              } else if (m == "rf") {
                group_prediction <- suppressWarnings(
                  randomForest::randomForest(
                    y = factor(data_train_small[[ml$cat_var]]),
                    x = data_train_small[, features, drop = FALSE],
                    ntree = min(ml$args$ntree, 100)
                  )
                )
                pred <- stats::predict(
                  group_prediction,
                  data_test_small[cc, features, drop = FALSE]
                )
                prob_matrix <- stats::predict(
                  group_prediction,
                  data_test_small[cc, features, drop = FALSE],
                  type = "prob"
                )
              } else if (m == "lda") {
                group_prediction <- suppressWarnings(
                  MASS::lda(
                    grouping = factor(data_train_small[[ml$cat_var]]),
                    x = as.matrix(data_train_small[, features, drop = FALSE])
                  )
                )
                lda_pred <- stats::predict(
                  group_prediction,
                  as.matrix(data_test_small[cc, features, drop = FALSE])
                )
                pred <- lda_pred$class
                prob_matrix <- lda_pred$posterior
              } else if (m == "nb") {
                group_prediction <- suppressWarnings(
                  e1071::naiveBayes(
                    x = data_train_small[, features, drop = FALSE],
                    y = factor(data_train_small[[ml$cat_var]])
                  )
                )
                pred <- stats::predict(
                  group_prediction,
                  data_test_small[cc, features, drop = FALSE],
                  type = "class"
                )
                prob_matrix <- stats::predict(
                  group_prediction,
                  data_test_small[cc, features, drop = FALSE],
                  type = "raw"
                )
              }
              res[[cat_nam]][["acc"]][m] <- mean(pred == data_test_small[[ml$cat_var]])

              # Extract probability assigned to true class for each test observation
              true_cats <- as.character(data_test_small[[ml$cat_var]][cc])
              prob_true <- vapply(seq_along(true_cats), function(i) {
                tc <- true_cats[i]
                if (tc %in% colnames(prob_matrix)) prob_matrix[i, tc] else 0
              }, numeric(1))
              res[[cat_nam]][["mean_prob"]][m] <- mean(prob_true)
              res[[cat_nam]][["brier"]][m] <- brier_score(
                prob_matrix,
                true_cats,
                matched_cat
              )
            },
            silent = TRUE
          )
        }
      },
      silent = TRUE
    )
  }

  invisible(structure(res, ml_models = methods, class = c("cat2cat_ml_run", "list")))
}


#' @rdname cat2cat_ml_run
#' @param x \code{cat2cat_ml_run} instance created with
#'   \code{\link{cat2cat_ml_run}}.
#' @param ... other arguments (currently unused).
#' @return \code{print} returns \code{x} invisibly, after printing a summary
#'   with average accuracy per method, baseline comparisons, and failure rates.
#' @method print cat2cat_ml_run
#' @export
print.cat2cat_ml_run <- function(x, ...) {
  # Average accuracy - please take into account it is multi-level classification
  ml_models <- attr(x, "ml_models")

  ml_acc_message <- NULL
  ml_brier_message <- NULL
  ml_meanprob_message <- NULL
  how_ml_message_n <- NULL
  how_ml_message_f <- NULL
  na_message <- NULL

  for (m in ml_models) {
    acc <- mean(vapply(x, function(i) i$acc[m], numeric(1)), na.rm = TRUE)
    ml_acc_message <- c(
      ml_acc_message,
      sprintf("  %s: accuracy = %.4f", m, acc)
    )

    brier <- mean(vapply(x, function(i) i$brier[m], numeric(1)), na.rm = TRUE)
    ml_brier_message <- c(
      ml_brier_message,
      sprintf("  %s: brier = %.4f", m, brier)
    )

    meanprob <- mean(vapply(x, function(i) i$mean_prob[m], numeric(1)), na.rm = TRUE)
    ml_meanprob_message <- c(
      ml_meanprob_message,
      sprintf("  %s: mean P(true) = %.4f", m, meanprob)
    )

    howaccn <- mean(vapply(x, function(i) i$naive < i$acc[m], numeric(1)), na.rm = TRUE)
    how_ml_message_n <- c(
      how_ml_message_n,
      sprintf("  %s > naive: %.1f%%", m, howaccn * 100)
    )

    howaccf <- mean(vapply(x, function(i) i$freq < i$acc[m], numeric(1)), na.rm = TRUE)
    how_ml_message_f <- c(
      how_ml_message_f,
      sprintf("  %s > freq: %.1f%%", m, howaccf * 100)
    )

    nna <- vapply(x, function(i) is.na(i$acc[m]), logical(1))
    pna <- sum(nna) / length(nna) * 100
    na_message <- c(
      na_message,
      sprintf("  %s: %.1f%%", m, pna)
    )
  }

  # Baseline stats
  acc_naive <- mean(vapply(x, function(i) i$naive, numeric(1)), na.rm = TRUE)
  acc_freq <- mean(vapply(x, function(i) i$freq, numeric(1)), na.rm = TRUE)
  brier_naive <- mean(vapply(x, function(i) i$naive_brier, numeric(1)), na.rm = TRUE)
  brier_freq <- mean(vapply(x, function(i) i$freq_brier, numeric(1)), na.rm = TRUE)
  meanprob_naive <- mean(vapply(x, function(i) i$naive_mean_prob, numeric(1)), na.rm = TRUE)
  meanprob_freq <- mean(vapply(x, function(i) i$freq_mean_prob, numeric(1)), na.rm = TRUE)

  cat(
    paste(
      c(
        "=== cat2cat ML Cross-Validation Results ===",
        "",
        "ACCURACY (higher is better):",
        sprintf("  naive (1/k): %.4f", acc_naive),
        sprintf("  freq (most common): %.4f", acc_freq),
        ml_acc_message,
        "",
        "BRIER SCORE (lower is better, range 0-1):",
        sprintf("  naive: %.4f", brier_naive),
        sprintf("  freq: %.4f", brier_freq),
        ml_brier_message,
        "",
        "MEAN P(TRUE CLASS) (higher is better):",
        sprintf("  naive: %.4f", meanprob_naive),
        sprintf("  freq: %.4f", meanprob_freq),
        ml_meanprob_message,
        "",
        "ACCURACY: ML vs BASELINES (percent of groups where ML wins):",
        how_ml_message_n,
        how_ml_message_f,
        "",
        "SKIPPED GROUPS (single category or <5 observations):",
        na_message
      ),
      collapse = "\n"
    ),
    "\n"
  )
  invisible(x)
}

set.seed(1234)
library("cat2cat")
data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]

test_data <- list(
  old = occup_2008,
  new = occup_2010,
  cat_var = "code",
  time_var = "year"
)

ml_nb_base <- list(
  data = rbind(occup_2010, occup_2012),
  cat_var = "code",
  method = "nb",
  features = c("age", "sex", "edu", "exp", "parttime", "salary")
)

testthat::test_that("cat2cat with ml adds expected nb weights", {
  library("e1071")

  ml_nb <- ml_nb_base
  ml_nb$fail_warn <- FALSE

  set.seed(1234)
  result <- cat2cat(
    data = test_data,
    mappings = list(trans = trans, direction = "backward"),
    ml = ml_nb
  )

  testthat::expect_s3_class(result$old, "data.frame")
  testthat::expect_s3_class(result$new, "data.frame")
  testthat::expect_true("wei_nb_c2c" %in% colnames(result$old))
  testthat::expect_true("wei_nb_c2c" %in% colnames(result$new))

  non_na_wei <- result$old$wei_nb_c2c[!is.na(result$old$wei_nb_c2c)]
  testthat::expect_gt(length(non_na_wei), 0)
  testthat::expect_true(all(non_na_wei >= 0 & non_na_wei <= 1))

  wei_sums <- tapply(result$old$wei_nb_c2c, result$old$index_c2c, sum, na.rm = TRUE)
  valid_sums <- wei_sums[wei_sums > 0]
  prop_valid <- mean(abs(valid_sums - 1) < 0.01)
  testthat::expect_gt(prop_valid, 0.5)
})

testthat::test_that("cat2cat ml on_fail controls fallback behavior", {
  library("e1071")

  ml_freq <- ml_nb_base
  ml_freq$on_fail <- "freq"
  freq_result <- NULL
  testthat::expect_warning(
    freq_result <- cat2cat(
      data = test_data,
      mappings = list(trans = trans, direction = "backward"),
      ml = ml_freq
    ),
    "[0-9]+\\.?[0-9]*% rows.*[0-9]+\\.?[0-9]*% observations.*on_fail = 'freq'"
  )

  ml_naive <- ml_nb_base
  ml_naive$on_fail <- "naive"
  naive_result <- NULL
  testthat::expect_warning(
    naive_result <- cat2cat(
      data = test_data,
      mappings = list(trans = trans, direction = "backward"),
      ml = ml_naive
    ),
    "on_fail = 'naive'"
  )

  ml_na <- ml_nb_base
  ml_na$on_fail <- "na"
  na_result <- NULL
  testthat::expect_warning(
    na_result <- cat2cat(
      data = test_data,
      mappings = list(trans = trans, direction = "backward"),
      ml = ml_na
    ),
    "on_fail = 'na'"
  )

  failed_rows <- is.na(na_result$old$wei_nb_c2c)
  testthat::expect_true(any(failed_rows))
  testthat::expect_true(all(!is.na(freq_result$old$wei_nb_c2c[failed_rows])))
  testthat::expect_true(all(!is.na(naive_result$old$wei_nb_c2c[failed_rows])))

  on_fail_diff <- abs(
    freq_result$old$wei_nb_c2c[failed_rows] -
      naive_result$old$wei_nb_c2c[failed_rows]
  )
  on_fail_diff <- on_fail_diff[is.finite(on_fail_diff)]
  testthat::expect_true(any(on_fail_diff > 1e-12))

  ml_error <- ml_nb_base
  ml_error$on_fail <- "error"
  testthat::expect_error(
    cat2cat(
      data = test_data,
      mappings = list(trans = trans, direction = "backward"),
      ml = ml_error
    ),
    "[0-9]+\\.?[0-9]*% rows.*[0-9]+\\.?[0-9]*% observations"
  )
})

testthat::test_that("cat2cat ml fail_warn suppresses fallback warnings", {
  library("e1071")

  ml_quiet <- ml_nb_base
  ml_quiet$on_fail <- "freq"
  ml_quiet$fail_warn <- FALSE

  testthat::expect_warning(
    cat2cat(
      data = test_data,
      mappings = list(trans = trans, direction = "backward"),
      ml = ml_quiet
    ),
    NA
  )
})

testthat::test_that("validate_ml rejects unsupported on_fail", {
  ml_bad <- list(
    data = rbind(occup_2010, occup_2012),
    cat_var = "code",
    method = c("knn", "rf", "lda"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50),
    on_fail = "unsupported"
  )

  testthat::expect_error(
    validate_ml(ml_bad),
    "`ml\\$on_fail` must be one of"
  )
})

testthat::test_that("encode_factor_features one-hot encodes factor features", {
  train <- data.frame(
    age = c(20, 30, 40),
    grp = factor(c("a", "b", "a")),
    code = c("1", "2", "3"),
    stringsAsFactors = FALSE
  )
  target <- data.frame(
    age = c(25, 35),
    grp = factor(c("a", "c"), levels = c("a", "c")),
    stringsAsFactors = FALSE
  )

  ml <- list(
    data = train,
    cat_var = "code",
    method = "nb",
    features = c("age", "grp")
  )

  out <- encode_factor_features(ml, target)

  testthat::expect_setequal(
    out$ml$features,
    c("age", "grp_a", "grp_b", "grp_c")
  )
  testthat::expect_true(all(out$ml$features %in% colnames(out$ml$data)))
  testthat::expect_true(all(out$ml$features %in% colnames(out$target_data)))

  testthat::expect_equal(out$ml$data$grp_a, c(1L, 0L, 1L))
  testthat::expect_equal(out$ml$data$grp_b, c(0L, 1L, 0L))
  testthat::expect_equal(out$ml$data$grp_c, c(0L, 0L, 0L))
  testthat::expect_equal(out$target_data$grp_a, c(1L, 0L))
  testthat::expect_equal(out$target_data$grp_c, c(0L, 1L))

  # original column preserved
  testthat::expect_true(is.factor(out$ml$data$grp))
})

testthat::test_that("encode_factor_features is a no-op when no factor features", {
  train <- data.frame(age = c(1, 2), x = c(TRUE, FALSE), code = c("a", "b"))
  target <- data.frame(age = c(3, 4), x = c(FALSE, TRUE))
  ml <- list(
    data = train, cat_var = "code", method = "nb",
    features = c("age", "x")
  )
  out <- encode_factor_features(ml, target)
  testthat::expect_identical(out$ml$features, c("age", "x"))
  testthat::expect_identical(colnames(out$ml$data), colnames(train))
  testthat::expect_identical(colnames(out$target_data), colnames(target))
})

testthat::test_that("encode_factor_features does not auto-encode character columns", {
  train <- data.frame(
    age = c(1, 2),
    region = c("n", "s"),
    code = c("a", "b"),
    stringsAsFactors = FALSE
  )
  target <- data.frame(age = c(3, 4), region = c("n", "s"), stringsAsFactors = FALSE)
  ml <- list(
    data = train, cat_var = "code", method = "nb",
    features = c("age", "region")
  )
  out <- encode_factor_features(ml, target)
  testthat::expect_identical(out$ml$features, c("age", "region"))
})

testthat::test_that("brier_score uses the full multiclass probability vector", {
  probs <- data.frame(
    a = c(0.8, 0.1),
    b = c(0.1, 0.6),
    c = c(0.1, 0.3)
  )

  expected <- mean(c(
    ((0.8 - 1)^2 + 0.1^2 + 0.1^2) / 2,
    (0.1^2 + (0.6 - 1)^2 + 0.3^2) / 2
  ))

  testthat::expect_equal(brier_score(probs, c("a", "b"), c("a", "b", "c")), expected)

  same_true_prob_1 <- data.frame(a = 0.6, b = 0.2, c = 0.2)
  same_true_prob_2 <- data.frame(a = 0.6, b = 0.4, c = 0.0)
  testthat::expect_gt(
    brier_score(same_true_prob_2, "a", c("a", "b", "c")),
    brier_score(same_true_prob_1, "a", c("a", "b", "c"))
  )
})

testthat::test_that("cat2cat with ml automatically one-hot encodes factor features", {
  library("e1071")

  occup_2008_f <- occup_2008
  occup_2010_f <- occup_2010
  occup_2012_f <- occup_2012

  # Convert a numeric column into a factor
  occup_2008_f$edu <- factor(occup_2008_f$edu)
  occup_2010_f$edu <- factor(occup_2010_f$edu)
  occup_2012_f$edu <- factor(occup_2012_f$edu)

  test_data_f <- list(
    old = occup_2008_f,
    new = occup_2010_f,
    cat_var = "code",
    time_var = "year"
  )

  ml_nb <- list(
    data = rbind(occup_2010_f, occup_2012_f),
    cat_var = "code",
    method = "nb",
    features = c("age", "sex", "edu", "salary"),
    fail_warn = FALSE
  )

  set.seed(1234)
  result <- cat2cat(
    data = test_data_f,
    mappings = list(trans = trans, direction = "backward"),
    ml = ml_nb
  )

  testthat::expect_s3_class(result$old, "data.frame")
  testthat::expect_true("wei_nb_c2c" %in% colnames(result$old))
  # original factor column preserved
  testthat::expect_true(is.factor(result$old$edu))

  non_na_wei <- result$old$wei_nb_c2c[!is.na(result$old$wei_nb_c2c)]
  testthat::expect_gt(length(non_na_wei), 0)
  testthat::expect_true(all(non_na_wei >= 0 & non_na_wei <= 1))
})

set.seed(1234)
library("cat2cat")
data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_2010 <- occup[occup$year == 2010,]
occup_2012 <- occup[occup$year == 2012,]

library("caret")
library("randomForest")
ml_setup_run <- list(
  data = rbind(occup_2010, occup_2012),
  cat_var = "code",
  method = c("knn", "rf", "lda"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 50)
)

testthat::test_that("cat2cat_ml_run returns stable diagnostics", {
  mappings <- list(trans = trans, direction = "backward")

  set.seed(1234)
  res <- cat2cat_ml_run(mappings, ml_setup_run)
  set.seed(1234)
  res2 <- cat2cat_ml_run(mappings, ml_setup_run, test_prop = 0.2)
  testthat::expect_equal(res, res2)
  testthat::expect_s3_class(res, c("cat2cat_ml_run", "list"))

  printed <- paste(capture.output(print(res)), collapse = "\n")
  testthat::expect_match(
    printed,
    "cat2cat ML Cross-Validation Results.*BRIER SCORE.*MEAN P\\(TRUE CLASS\\).*SKIPPED GROUPS"
  )

  # Check new metric fields exist in results
  non_na_idx <- which(!is.na(vapply(res, function(g) g$acc["knn"], numeric(1))))
  testthat::expect_true(length(non_na_idx) > 0)
  non_na_group <- res[[non_na_idx[1]]]
  testthat::expect_true(all(c("brier", "mean_prob", "naive_brier", "freq_brier") %in% names(non_na_group)))
  testthat::expect_true(all(c("knn", "rf", "lda") %in% names(non_na_group$brier)))
  # Mean stats across all groups should be positive
  avg_brier <- mean(vapply(res, function(g) g$brier["knn"], numeric(1)), na.rm = TRUE)
  avg_mean_prob <- mean(vapply(res, function(g) g$mean_prob["knn"], numeric(1)), na.rm = TRUE)
  testthat::expect_true(avg_brier > 0 && avg_brier <= 1)
  testthat::expect_true(avg_mean_prob > 0 && avg_mean_prob <= 1)
})

testthat::test_that("cat2cat_ml_run errors for unsupported direction", {
  mappings <- list(trans = trans, direction = "forward")

  testthat::expect_error(
    cat2cat_ml_run(mappings, ml_setup_run, test_prop = 0.2),
    "There is no mappings to group the cat_var variable"
  )
})

testthat::test_that("cat2cat_ml_run supports naive bayes metrics", {
  library("e1071")
  mappings <- list(trans = trans, direction = "backward")
  ml_nb <- list(
    data = rbind(occup_2010, occup_2012),
    cat_var = "code",
    method = c("nb", "lda"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary")
  )

  set.seed(1234)
  res <- cat2cat_ml_run(mappings, ml_nb)
  testthat::expect_s3_class(res, c("cat2cat_ml_run", "list"))

  printed_nb <- paste(capture.output(print(res)), collapse = "\n")
  testthat::expect_match(
    printed_nb,
    "nb: accuracy.*nb: brier.*nb: mean P\\(true\\)"
  )

  # Check nb metrics exist
  non_na_idx <- which(!is.na(vapply(res, function(g) g$acc["nb"], numeric(1))))
  testthat::expect_true(length(non_na_idx) > 0)
  avg_acc_nb <- mean(vapply(res, function(g) g$acc["nb"], numeric(1)), na.rm = TRUE)
  testthat::expect_true(avg_acc_nb > 0 && avg_acc_nb <= 1)
})

testthat::test_that("cat2cat_ml_run method skips are independent across methods", {
  mappings <- list(trans = trans, direction = "backward")
  ml_features <- c("age", "sex", "edu", "exp", "parttime", "salary")
  methods <- c("knn", "lda", "rf", "nb")

  skipped_rate <- function(res, method) {
    mean(vapply(res, function(g) is.na(g$acc[[method]]), logical(1)))
  }

  set.seed(1234)
  together <- cat2cat_ml_run(
    mappings,
    list(
      data = occup_2010,
      cat_var = "code",
      method = methods,
      features = ml_features,
      args = list(k = 10, ntree = 50)
    ),
    test_prop = 0.2
  )

  alone_rates <- setNames(numeric(length(methods)), methods)
  for (m in methods) {
    set.seed(1234)
    res_single <- cat2cat_ml_run(
      mappings,
      list(
        data = occup_2010,
        cat_var = "code",
        method = m,
        features = ml_features,
        args = list(k = 10, ntree = 50)
      ),
      test_prop = 0.2
    )
    alone_rates[[m]] <- skipped_rate(res_single, m)
  }

  together_rates <- vapply(methods, function(m) skipped_rate(together, m), numeric(1))
  testthat::expect_lt(max(abs(together_rates - alone_rates)), 0.05)
})

#' Transforming a mapping (transition) table to two associative lists
#' @description to rearrange the one classification encoding into another,
#'  an associative list that maps keys to values is used.
#' More precisely, an association list is used which is a linked list in which
#' each list element consists of a key and value or values.
#' An association list where unique categories codes are keys and matching
#' categories from next or previous time point are values.
#' A mapping (transition) table is used to build such associative lists.
#' @param x `data.frame` or `matrix` - mapping (transition) table with 2 columns
#'  where first column is assumed to be the older encoding.
#' @return a list with 2 named lists `to_old` and `to_new`.
#' @examples
#' data("trans", package = "cat2cat")
#'
#' mappings <- get_mappings(trans)
#' mappings$to_old[1:4]
#' mappings$to_new[1:4]
#' @export
get_mappings <- function(x = data.frame()) {
  stopifnot(
    "`x` must be a data.frame or matrix" =
      is.data.frame(x) || is.matrix(x)
  )
  stopifnot(
    "`x` must have exactly 2 columns (old encoding, new encoding)" =
      ncol(x) == 2
  )

  x <- as.matrix(x)

  ff <- x[, 1]
  ss <- x[, 2]

  from_old <- unique(ff)
  from_new <- unique(ss)

  to_old <- lapply(from_new, function(i) {
    tryCatch(
      unique(ff[ss %in% i]),
      error = function(e) {
        NA
      }
    )
  })
  names(to_old) <- from_new

  to_new <- lapply(from_old, function(i) {
    tryCatch(
      unique(ss[ff %in% i]),
      error = function(e) {
        NA
      }
    )
  })
  names(to_new) <- from_old

  list(to_old = to_old, to_new = to_new)
}

#' Getting frequencies from a vector with an optional multiplier
#' @description getting frequencies for a vector with an optional multiplier.
#' @param x `vector` categorical variable to summarize.
#' @param multiplier `numeric` vector how many times to repeat certain value,
#' additional weights.
#' @return `data.frame` with two columns `input` `Freq`
#' @note without multiplier variable it is a basic `table` function wrapped with
#'  the `as.data.frame` function.
#' The `table` function is used with the `useNA = "ifany"` argument.
#' @export
#' @examples
#' data("occup", package = "cat2cat")
#'
#' head(get_freqs(occup$code[occup$year == "2008"]))
#' head(get_freqs(occup$code[occup$year == "2010"]))
#'
#' head(
#'   get_freqs(
#'     occup$code[occup$year == "2008"],
#'     occup$multiplier[occup$year == "2008"]
#'   )
#' )
#' head(
#'   get_freqs(
#'     occup$code[occup$year == "2010"],
#'     occup$multiplier[occup$year == "2010"]
#'   )
#' )
get_freqs <- function(x, multiplier = NULL) {
  stopifnot("`x` must be a vector" = is.vector(x))
  stopifnot(
    "`multiplier` must be NULL or same length as `x`" =
      is.null(multiplier) || length(x) == length(multiplier)
  )

  input <- if (!is.null(multiplier)) {
    rep(x, times = as.numeric(multiplier))
  } else {
    x
  }
  # Build data.frame directly to avoid an r-devel regression in
  # as.data.frame.table() when NA appears in table names.
  tab <- table(input, useNA = "ifany")
  res <- data.frame(
    input = names(tab),
    Freq = as.integer(tab),
    stringsAsFactors = FALSE
  )
  res
}


#' Applying frequencies to the object returned by the `get_mappings` function
#' @description applying frequencies to the object returned by
#' the `get_mappings` function.
#' We will get a symmetric object to the one returned by
#' the `get_mappings` function, nevertheless categories are replaced
#' with frequencies.
#' Frequencies for each category/key are sum to 1, so could be interpreted
#' as probabilities.
#' @param to_x `list` object returned by `get_mappings`.
#' @param freqs `data.frame` object like the one returned by
#' the `get_freqs` function.
#' @return a `list` with a structure like `to_x` object but with probabilities
#' for each category.
#' @note
#' `freqs` arg first column (keys) and the to_x arg values have to be of
#' the same type.
#' The uniform distribution (outcomes are equally likely) is assumed
#' for no match for all possible categories.
#' @examples
#' data("trans", package = "cat2cat")
#' data("occup", package = "cat2cat")
#'
#' mappings <- get_mappings(trans)
#'
#' mappings$to_old[1:4]
#' mappings$to_new[1:4]
#'
#' mapp_p <- cat_apply_freq(
#'   mappings$to_old,
#'   get_freqs(
#'     occup$code[occup$year == "2008"],
#'     occup$multiplier[occup$year == "2008"]
#'   )
#' )
#' head(data.frame(I(mappings$to_old), I(mapp_p)))
#' mapp_p <- cat_apply_freq(
#'   mappings$to_new,
#'   get_freqs(
#'     occup$code[occup$year == "2010"],
#'     occup$multiplier[occup$year == "2010"]
#'   )
#' )
#' head(data.frame(I(mappings$to_new), I(mapp_p)))
#' @export
cat_apply_freq <- function(to_x, freqs) {
  stopifnot("`to_x` must be a list" = is.list(to_x))
  stopifnot(
    "`freqs` must have exactly 2 columns (category, frequency)" =
      ncol(freqs) == 2
  )
  res <- lapply(
    to_x,
    function(x) {
      alls <- freqs[, 2, drop = TRUE][match(x, freqs[, 1, drop = TRUE])]
      ff <- alls / sum(alls, na.rm = TRUE)
      # NA to 0
      ff[is.na(ff) | is.nan(ff)] <- 0
      # all equal to zero so proportional probability
      if (all(ff == 0)) {
        ff_len <- length(ff)
        ff <- rep(1 / ff_len, ff_len)
      }
      ff
    }
  )
  names(res) <- names(to_x)
  res
}

testthat::test_that("get_freqs", {
  expected_freqs <- function(x) {
    tab <- table(x, useNA = "ifany")
    data.frame(
      input = names(tab),
      Freq = as.integer(tab),
      stringsAsFactors = FALSE
    )
  }

  input <- c("a", "a", "a", "c", "c", "b", "d", NA)
  expect_identical(
    get_freqs(input),
    expected_freqs(input)
  )
  set.seed(1234)
  input_multiplier <- sample(1:10, length(input))
  elem1 <- get_freqs(input, multiplier = input_multiplier)
  input <- rep(input, times = input_multiplier)
  elem2 <- expected_freqs(input)
  expect_identical(elem1, elem2)
})

mapping_table <- data.frame(
  old = c(1, 1, 1, 2, 2, 3, NA, NA),
  new = c(NA, 1, 2, 2, 3, 2, NA, 3),
  stringsAsFactors = FALSE
)

testthat::test_that("get_mappings", {
  mappings <- get_mappings(mapping_table)
  expect_identical(length(mappings$to_new), 4L)
  expect_identical(
    lengths(mappings$to_new),
    structure(c(3L, 2L, 1L, 2L), names = c("1", "2", "3", NA))
  )
  expect_identical(length(mappings$to_old), 4L)
  expect_identical(
    lengths(mappings$to_old),
    structure(c(2L, 1L, 3L, 2L), names = c(NA, "1", "2", "3"))
  )
})

testthat::test_that("cat_apply_freq", {
  mappings <- get_mappings(mapping_table)
  mappings_freq <- cat_apply_freq(
    mappings$to_new,
    get_freqs(c(1, 1, 1, 1, 2, 3, 3, NA, NA))
  )
  expect_identical(mappings_freq$`1`, c(2 / 7, 4 / 7, 1 / 7))
  expect_identical(length(mappings_freq), 4L)
  expect_equal(
    lengths(mappings_freq),
    structure(c(3L, 2L, 1L, 2L), names = c("1", "2", "3", NA))
  )
  mappings_freq2 <- cat_apply_freq(mappings$to_new, get_freqs(c(1, 1, 1, 1)))
  expect_identical(mappings_freq2$`1`, c(0, 1L, 0))
  expect_identical(mappings_freq2$`2`, c(0.5, 0.5))
})

testthat::test_that("resolve_frequencies falls back to get_freqs when weights are absent", {
  cat_base_year <- data.frame(
    code = c("A", "A", "B", "C"),
    mult = c(2, 1, 3, 1),
    stringsAsFactors = FALSE
  )

  out <- resolve_frequencies(
    cat_base_year = cat_base_year,
    cat_var_base = "code",
    freqs_df = NULL,
    multiplier_var = "mult"
  )

  expect_identical(out, get_freqs(cat_base_year$code, multiplier = cat_base_year$mult))
})

testthat::test_that("delayed_package_load throws custom error for missing namespace", {
  expect_error(
    delayed_package_load("definitely_not_installed_pkg_for_cat2cat", "custom missing package message"),
    "custom missing package message"
  )
})

#' Automatic mapping in a panel dataset
#' @description
#' Unifies an inconsistently coded categorical variable in a panel dataset
#' according to a mapping (transition) table.
#' Apply iteratively for each pair of neighboring periods.
#' Use \code{\link{prune_c2c}} to limit growing replications across steps.
#' @param data `named list` with fields `old`, `new`,
#' `cat_var` (shorthand for `cat_var_old = cat_var_new = cat_var`),
#' `time_var` and optional `id_var`, `multiplier_var`.
#' @param mappings `named list` with 3 fields `trans`, `direction` and
#' optional `freqs_df`.
#' @param ml `named list` (optional) with fields
#' `data`, `cat_var`, `method`, `features` and optional `args`,
#' `on_fail`, `fail_warn`.
#' @details
#' data args
#' \describe{
#'  \item{"old"}{ data.frame older time point in a panel}
#'  \item{"new"}{ data.frame more recent time point in a panel}
#'  \item{"time_var"}{ character(1) name of the time variable.}
#'  \item{"cat_var"}{ character(1) name of the categorical variable.
#'   Shorthand: sets both \code{cat_var_old} and \code{cat_var_new} to the same value.
#'  }
#'  \item{"cat_var_old"}{
#'  Optional character(1) name of the categorical variable
#'  in the older time point. Default `cat_var`.
#'  }
#'  \item{"cat_var_new"}{
#'  Optional character(1) name of the categorical variable
#'  in the newer time point. Default `cat_var`.
#'  }
#'  \item{"id_var"}{Optional character(1) name of the unique identifier variable.
#'   When specified, subjects observed in both periods are mapped 1-to-1
#'   directly (no replication, \code{wei_freq_c2c = 1, rep_c2c = 1}).
#'   Only subjects absent from the base period enter the replication path.
#'   This assumes a subject's true category does not change between adjacent
#'   waves. See \code{vignette("cat2cat_advanced")} for details on when
#'   this assumption is and is not satisfied.
#'  }
#'  \item{"multiplier_var"}{
#'  Optional character(1) name of the multiplier variable -
#'  number of replication needed to reproduce the population
#'  }
#'  \item{"freqs_df"}{
#'  Deprecated - use \code{mappings$freqs_df} instead.
#'  }
#' }
#' mappings args
#' \describe{
#'  \item{"trans"}{ data.frame with 2 columns - mapping (transition) table -
#'   all categories for cat_var in old and new datasets have to be included.
#'   First column contains an old encoding and second a new one.
#'   The mapping (transition) table should include a candidate for each category
#'   in the period being harmonised.
#' }
#'  \item{"direction"}{ character(1) direction - "backward" or "forward"}
#'  \item{"freqs_df"}{
#'  Optional - data.frame with 2 columns where first one
#'  is category name (base period) and second counts.
#'  If It is not provided then is assessed automatically.
#'  Artificial counts for each variable level in the base period.
#'  It is optional nevertheless will be often needed, as gives more control.
#'  It will be used to assess the probabilities.
#'  The multiplier variable is omitted so sb has to apply it in this table.
#'  }
#' }
#' Optional ml args
#' \describe{
#'  \item{"data"}{ data.frame - dataset with features and the `cat_var`.}
#'  \item{"cat_var"}{ character(1) - the dependent variable name.}
#'  \item{"method"}{
#'  character vector - one or a few from
#'  "knn", "rf", "lda" and "nb" methods - "knn" k-NearestNeighbors,
#'  "lda" Linear Discriminant Analysis, "rf" Random Forest,
#'  "nb" Naive Bayes
#'  }
#'  \item{"features"}{
#'  character vector of features names where all
#'  have to be numeric, logical or factor.
#'  Factor features are automatically one-hot encoded using the union of
#'  levels observed in \code{ml$data} and the target period.
#'  }
#'  \item{"args"}{ optional - list parameters: knn: k ; rf: ntree  }
#'  \item{"on_fail"}{ optional character(1) controlling failed ML weights:
#'  \code{"freq"} (default) uses \code{wei_freq_c2c}, \code{"naive"} uses
#'  \code{wei_naive_c2c}, \code{"na"} leaves failed weights as \code{NA},
#'  and \code{"error"} stops when failed weights are detected.}
#'  \item{"fail_warn"}{ optional logical(1), default \code{TRUE}; warn when
#'  failed ML weights are replaced or retained as \code{NA}.}
#' }
#'
#' Without the \code{ml} argument, only frequency-based weights are computed.
#' If an ML model fails, \code{ml$on_fail} controls whether frequency weights,
#' naive weights, \code{NA}, or an error are used.
#' The \code{knn} method is recommended for smaller datasets.
#' @return `named list` with 2 fields old and new - 2 data.frames.
#' There will be added additional columns like
#' index_c2c, g_new_c2c, wei_freq_c2c, rep_c2c, wei_(ml method name)_c2c.
#' Additional columns will be informative only for a one data.frame
#' as we always make the changes to one direction.
#' The new columns are added instead of the additional metadata as
#' we are working with new datasets
#' where observations could be replicated.
#' For the transparency the probability and number of replications are part
#'  of each observation in the `data.frame`.
#' @importFrom stats predict complete.cases setNames
#' @importFrom MASS lda
#' @note
#' \code{trans} columns and \code{cat_var} must be of the same type.
#' The mapping table must include a candidate for every category in the target
#' period. Observations without a matched candidate are dropped; add a
#' \code{c(NA, NA)} row to \code{trans} to retain them as \code{NA}.
#' @seealso
#' \itemize{
#'   \item \code{\link{cat2cat_ml_run}} - validate ML performance before use
#'   \item \code{\link{prune_c2c}}, \code{\link{cross_c2c}} - post-processing
#'   \item \code{\link{summary_c2c}}, \code{\link{dummy_c2c}} - helpers
#'   \item \code{\link{cat2cat_agg}} - for pre-aggregated data
#'   \item \code{vignette("cat2cat")} - identification assumptions and worked examples
#' }
#' @export
#' @examples
#' \dontrun{
#' data("occup_small", package = "cat2cat")
#' data("occup", package = "cat2cat")
#' data("trans", package = "cat2cat")
#'
#' occup_old <- occup_small[occup_small$year == 2008, ]
#' occup_new <- occup_small[occup_small$year == 2010, ]
#'
#' # Adding the dummy level to the mapping table for levels without a candidate
#' # The best to fill them manually with proper candidates, if possible
#' # In this case it is only needed for forward mapping, to suppress warnings
#' trans2 <- rbind(
#'   trans,
#'   data.frame(
#'     old = "no_cat",
#'     new = setdiff(c(occup_new$code), trans$new)
#'   )
#' )
#'
#' # default only simple frequencies
#' occup_simple <- cat2cat(
#'   data = list(
#'     old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
#'   ),
#'   mappings = list(trans = trans2, direction = "forward")
#' )
#'
#' mappings <- list(trans = trans, direction = "backward")
#'
#' ml_setup <- list(
#'     data = occup_small[occup_small$year >= 2010, ],
#'     cat_var = "code",
#'     method = "knn",
#'     features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#'     args = list(k = 10),
#'     # defaults for failed ML weights
#'     on_fail = "freq",
#'     fail_warn = TRUE
#' )
#'
#' # ml model performance check
#' print(cat2cat_ml_run(mappings, ml_setup))
#'
#' # additional probabilities from knn
#' occup_ml <- cat2cat(
#'   data = list(
#'     old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
#'   ),
#'   mappings = mappings,
#'   ml = ml_setup
#' )
#'
#' # strict mode: stop when any ML weights cannot be produced
#' ml_setup_strict <- ml_setup
#' ml_setup_strict$on_fail <- "error"
#'
#' # diagnostic mode: keep failed ML weights as NA and silence warnings
#' ml_setup_diag <- ml_setup
#' ml_setup_diag$on_fail <- "na"
#' ml_setup_diag$fail_warn <- FALSE
#' }
#'
cat2cat <- function(data = list(
                      old = NULL,
                      new = NULL,
                      time_var = NULL,
                      cat_var = NULL,
                      cat_var_old = NULL,
                      cat_var_new = NULL,
                      id_var = NULL,
                      multiplier_var = NULL
                    ),
                    mappings = list(
                      trans = NULL,
                      direction = NULL,
                      freqs_df = NULL
                    ),
                    ml = list(
                      data = NULL,
                      cat_var = NULL,
                      method = NULL,
                      features = NULL,
                      args = NULL
                    )) {
  stopifnot("`data` must be a list" = is.list(data))
  stopifnot("`mappings` must be a list" = is.list(mappings))
  stopifnot("`ml` must be a list (or NULL)" = is.list(ml))

  # Backward compatibility
  if (is.null(data$freqs_df)) data$freqs_df <- mappings$freqs_df
  if (is.null(data$cat_var_old)) data$cat_var_old <- data$cat_var
  if (is.null(data$cat_var_new)) data$cat_var_new <- data$cat_var

  validate_data(data)
  validate_mappings(mappings)

  mapps <- get_mappings(mappings$trans)

  if (mappings$direction == "forward") {
    base_name <- "old"
    target_name <- "new"
  } else if (mappings$direction == "backward") {
    base_name <- "new"
    target_name <- "old"
  }

  cat_base_year <- data[[base_name]]
  cat_target_year <- data[[target_name]]
  cat_var_base <- data[[paste0("cat_var_", base_name)]]
  cat_var_target <- data[[paste0("cat_var_", target_name)]]
  cat_mid <- NULL # direct match

  is_direct_match <- !is.null(data$id_var)
  if (is_direct_match) {
    id_inner <- intersect(data$old[[data$id_var]], data$new[[data$id_var]])
    id_outer <- setdiff(
      data[[target_name]][[data$id_var]],
      data[[base_name]][[data$id_var]]
    )

    cat_target_year <-
      data[[target_name]][data[[target_name]][[data$id_var]] %in% id_outer, ]
    cat_mid <- dummy_c2c(
      data[[target_name]][data[[target_name]][[data$id_var]] %in% id_inner, ],
      cat_var_base
    )

    tos <- merge(data$old[, c(data$id_var, data$cat_var_old)],
      data$new[, c(data$id_var, data$cat_var_new)],
      by = data$id_var
    )
    colnames(tos) <- c("id", "cat_old", "cat_new")
    cat_mid$g_new_c2c <-
      tos[[paste0("cat_", base_name)]][match(cat_mid[[data$id_var]], tos$id)]
  }

  cats_target <- cat_target_year[[cat_var_target]]

  mapp <- mapps[[paste0("to_", base_name)]]

  validate_cover_cats(unique(cats_target), mapp)

  # frequencies
  fre <- resolve_frequencies(
    cat_base_year = cat_base_year, cat_var_base = cat_var_base,
    freqs_df = data$freqs_df, multiplier_var = data$multiplier_var
  )
  # frequencies per category
  freqs_list <- cat_apply_freq(mapp, fre)

  # Replicate and apply freq probabilities, target period
  g_vec <- mapp[match(cats_target, names(mapp))]
  rep_vec <- unname(lengths(g_vec))
  wei_vec <- freqs_list[match(cats_target, names(freqs_list))]
  wei_vec <- unlist(wei_vec, use.names = FALSE)
  cat_target_year$index_c2c <- seq_len(nrow(cat_target_year))
  cat_target_rep <-
    cat_target_year[rep(seq_len(nrow(cat_target_year)), times = rep_vec), ]

  # Target
  cat_target_rep$g_new_c2c <- unlist(g_vec, use.names = FALSE)
  cat_target_rep$wei_freq_c2c <- wei_vec
  cat_target_rep$rep_c2c <- rep(rep_vec, times = rep_vec)
  cat_target_rep$wei_naive_c2c <- 1 / cat_target_rep$rep_c2c
  # Base
  cat_base_year <- dummy_c2c(cat_base_year, cat_var_base)

  # ML
  if (sum(vapply(ml, Negate(is.null), logical(1))) >= 1) {
    # Backward compatibility
    if (is.null(ml$data)) ml$data <- cat_base_year
    if (is.null(ml$cat_var)) ml$cat_var <- cat_var_base

    ml_names <- paste0("wei_", unique(ml$method), "_c2c")

    cat_base_year[, setdiff(ml_names, colnames(cat_base_year))] <- 1
    if (is_direct_match) cat_mid[, setdiff(ml_names, colnames(cat_mid))] <- 1

    ml_results <- cat2cat_ml(
      ml = ml,
      mapp = mapp,
      target_data = cat_target_rep,
      cat_var_target = cat_var_target
    )

    cat_target_rep <- ml_results$target_data
  }
  # bind direct match if applied
  cat_target_rep_f <- rbind(cat_target_rep, cat_mid)

  res <- list()
  res[[base_name]] <- cat_base_year
  res[[target_name]] <- cat_target_rep_f
  res
}

#' Validate if the trans table contains all proper mappings
#' @param u_cats_target vector of unique target period categories
#' @param mapp transition (mapping) table process with `get_mappings`,
#'  the "to_base" direction is taken.
#' @keywords internal
validate_cover_cats <- function(u_cats_target, mapp) {
  if (
    cats_target_diff_len <-
      length(cats_target_diff <- setdiff(u_cats_target, names(mapp)))
  ) {
    warning(
      paste(
        sprintf(
          paste0(
            "trans table does not cover some levels for a target period, ",
            "so the result will not contain all original observations.",
            " Lacking %s levels: "
          ),
          cats_target_diff_len
        ),
        paste(head(cats_target_diff, 10), collapse = ", "),
        if (cats_target_diff_len >= 10) "..." else NULL
      )
    )
  }
}

#' Resolve the frequencies
#' @param cat_base_year `data.frame` a base period dataset.
#' @param cat_var_base `character(1)` name of the base period
#' categorical variable.
#' @param freqs_df `data.frame` with 2 columns, the first one with category
#' and second with counts.
#' @param multiplier_var `character(1)` name of the multiplier variable.
#' @keywords internal
resolve_frequencies <- function(cat_base_year,
                                cat_var_base,
                                freqs_df,
                                multiplier_var) {
  cats_base <- cat_base_year[[cat_var_base]]
  freqs <- if (is.data.frame(freqs_df)) {
    freqs_df
  } else if ("wei_freq_c2c" %in% colnames(cat_base_year)) {
    stats::aggregate(
      if (!is.null(multiplier_var)) {
        cat_base_year[["wei_freq_c2c"]] * cat_base_year[[multiplier_var]]
      } else {
        cat_base_year[["wei_freq_c2c"]]
      },
      list(g = cats_base),
      function(x) round(sum(x, na.rm = TRUE))
    )
  } else {
    get_freqs(
      cats_base,
      if (!is.null(multiplier_var)) cat_base_year[[multiplier_var]]
    )
  }
  freqs
}

#' Validate cat2cat data
#' @keywords internal
validate_data <- function(data) {
  stopifnot("`data$old` must be a data.frame" = inherits(data$old, "data.frame"))
  stopifnot("`data$new` must be a data.frame" = inherits(data$new, "data.frame"))
  stopifnot(
    "`data$cat_var_old` and `data$cat_var_new` must be specified" =
      !is.null(data$cat_var_old) && !is.null(data$cat_var_new)
  )

  stopifnot(
    "`cat_var_old` and `time_var` must be columns in `data$old`" =
      all(c(data$cat_var_old, data$time_var) %in% colnames(data$old))
  )
  stopifnot(
    "`cat_var_new` and `time_var` must be columns in `data$new`" =
      all(c(data$cat_var_new, data$time_var) %in% colnames(data$new))
  )
  stopifnot(
    "`multiplier_var` must be a column in both `data$old` and `data$new`" =
      is.null(data$multiplier_var) ||
      (data$multiplier_var %in% colnames(data$new) &&
       data$multiplier_var %in% colnames(data$old))
  )

  stopifnot(
    "`freqs_df` must be NULL or a 2-column data.frame" =
      is.null(data$freqs_df) ||
      (is.data.frame(data$freqs_df) && ncol(data$freqs_df) == 2)
  )
  stopifnot(
    "Each period must have exactly one unique time value in `time_var`" =
      (length(unique(data$old[[data$time_var]])) == 1) &&
      (length(unique(data$new[[data$time_var]])) == 1)
  )
  stopifnot(
    "`id_var` must be a column in both periods with no duplicates within each period" =
      is.null(data$id_var) ||
      ((data$id_var %in% colnames(data$old)) &&
       (data$id_var %in% colnames(data$new)) &&
       !anyDuplicated(data$old[[data$id_var]]) &&
       !anyDuplicated(data$new[[data$id_var]]))
  )
}

#' Validate cat2cat mappings
#' @keywords internal
validate_mappings <- function(mappings) {
  stopifnot(
    "All `mappings` elements must be non-NULL" =
      all(vapply(mappings, Negate(is.null), logical(1)))
  )
  stopifnot(
    "`mappings` must contain 'trans' and 'direction'" =
      all(c("trans", "direction") %in% names(mappings))
  )
  stopifnot(
    "`mappings$direction` must be 'forward' or 'backward'" =
      isTRUE(mappings$direction %in% c("forward", "backward"))
  )
  stopifnot(
    "`mappings$trans` must be a 2-column data.frame" =
      is.data.frame(mappings$trans) && ncol(mappings$trans) == 2
  )
}

set.seed(1234)

aa <- airquality
aa2 <- rbind(aa, aa)

testthat::test_that("summary_c2c is properly adjust the std error", {
  ll <- lm(Ozone ~ ., aa)
  ss <- summary(ll)

  ll2 <- lm(Ozone ~ ., aa2)
  ss2 <- summary_c2c(ll2, ll$df.residual, ll2$df.residual)

  expect_equal(unname(ss$coefficients[, 3]), ss2[, 7], tolerance = 1e-8)
})

data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup[occup$year == 2008, ]
occup_new <- occup[occup$year == 2010, ]

data_simple <- list(
  old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
)
mappings_simple_back <- list(trans = trans, direction = "backward")
ml <- list(
  data = occup_new,
  cat_var = "code",
  method = c("knn", "rf", "lda"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 30),
  fail_warn = FALSE
)

testthat::test_that(
  "summary_c2c is properly adjust the std error - cat2cat case",
  {
    occup <- cat2cat(
      data = data_simple,
      mappings = mappings_simple_back,
      ml = ml
    )

    lms <- lm(
      I(log(salary)) ~ age + sex + factor(edu) + parttime + exp, occup$old,
      weights = multiplier * wei_freq_c2c
    )

    ss_c2c <- summary_c2c(lms, df_old = nrow(occup_old) - length(lms$assign))
    lms$df.residual <- nrow(occup_old) - length(lms$assign)
    ss1 <- suppressWarnings(summary(lms))

    lms2 <- lm(
      I(log(salary)) ~ age + sex + factor(edu) + parttime + exp, occup_old,
      weights = multiplier
    )
    ss2 <- summary(lms2)

    expect_true(sum((ss2$coefficients[, 2] - ss1$coefficients[, 2])**2) < 0.01)
    expect_true(sum((ss2$coefficients[, 2] - ss_c2c$std.error_c)**2) < 0.01)
  }
)

# -----------------------------------------------------------------------------
# summary_c2c with occup_panel (id_var case)
# -----------------------------------------------------------------------------

data("occup_panel", package = "cat2cat")

panel_2009Q4 <- occup_panel[occup_panel$quarter == "2009Q4", ]
panel_2010Q1 <- occup_panel[occup_panel$quarter == "2010Q1", ]

testthat::test_that(
  "summary_c2c corrects SE with occup_panel id_var data",
  {
    result <- cat2cat(
      data = list(
        old = panel_2009Q4,
        new = panel_2010Q1,
        id_var = "panel_id",
        cat_var = "code",
        time_var = "quarter"
      ),
      mappings = list(trans = trans, direction = "backward")
    )

    # Fit model on replicated data
    lm_rep <- lm(
      I(log(salary)) ~ age + sex + factor(edu),
      data = result$old,
      weights = wei_freq_c2c
    )

    # Get corrected SE
    df_orig <- nrow(panel_2009Q4) - length(coef(lm_rep))
    ss_c2c <- summary_c2c(lm_rep, df_old = df_orig)

    # Corrected SE should be larger than naive SE (replication inflates df)
    naive_se <- summary(lm_rep)$coefficients[, "Std. Error"]
    expect_true(all(ss_c2c$std.error_c >= naive_se * 0.99))

    # Corrected stats should use original df
    expect_equal(nrow(ss_c2c), length(coef(lm_rep)))
    expect_true(all(c("statistic_c", "p.value_c", "std.error_c") %in% names(ss_c2c)))
  }
)

testthat::test_that(

  "summary_c2c SE correction ratio matches replication factor",
  {
    result <- cat2cat(
      data = list(
        old = panel_2009Q4,
        new = panel_2010Q1,
        cat_var = "code",
        time_var = "quarter"
      ),
      mappings = list(trans = trans, direction = "backward")
    )

    lm_rep <- lm(
      I(log(salary)) ~ age + sex,
      data = result$old,
      weights = wei_freq_c2c
    )

    df_orig <- nrow(panel_2009Q4) - length(coef(lm_rep))
    df_rep <- nrow(result$old) - length(coef(lm_rep))
    ss_c2c <- summary_c2c(lm_rep, df_old = df_orig, df_new = df_rep)

    # SE correction factor should be sqrt(df_rep / df_orig)
    expected_factor <- sqrt(df_rep / df_orig)
    naive_se <- summary(lm_rep)$coefficients[, "Std. Error"]
    actual_factor <- unname(ss_c2c$std.error_c[1] / naive_se[1])

    expect_equal(actual_factor, expected_factor, tolerance = 0.01)
  }
)

testthat::test_that(
  "summary_c2c handles glm gaussian path",
  {
    aa_glm <- na.omit(airquality)
    aa_glm2 <- rbind(aa_glm, aa_glm)

    g1 <- glm(Ozone ~ Solar.R + Wind + Temp,
              data = aa_glm, family = gaussian())
    g2 <- glm(Ozone ~ Solar.R + Wind + Temp,
              data = aa_glm2, family = gaussian())

    ss1 <- summary(g1)$coefficients[, 3]
    ss2 <- summary_c2c(g2, g1$df.residual, g2$df.residual)

    expect_true(all(c("std.error_c", "statistic_c", "p.value_c", "reference_dist") %in% names(ss2)))
    expect_equal(unname(ss1), ss2$statistic_c, tolerance = 0.06)
    expect_true(all(ss2$reference_dist == "t"))
  }
)

testthat::test_that(
  "summary_c2c handles glm binomial path",
  {
    d1 <- mtcars
    d2 <- rbind(d1, d1)

    g1 <- glm(vs ~ mpg + wt,
              data = d1, family = binomial())
    g2 <- glm(vs ~ mpg + wt,
              data = d2, family = binomial())

    ss1 <- summary(g1)$coefficients[, "z value"]
    ss2 <- summary_c2c(g2, g1$df.residual, g2$df.residual)

    expect_true(all(c("std.error_c", "statistic_c", "p.value_c", "reference_dist") %in% names(ss2)))
    expect_equal(unname(ss1), ss2$statistic_c, tolerance = 0.06)
    expect_true(all(ss2$reference_dist == "normal"))
  }
)

testthat::test_that(
  "summary_c2c rejects invalid df inputs",
  {
    ll <- lm(Ozone ~ Solar.R + Wind + Temp, data = na.omit(airquality))

    expect_error(summary_c2c(ll, df_old = 0), "`df_old` must be a single positive finite numeric value.")
    expect_error(summary_c2c(ll, df_old = NA_real_), "`df_old` must be a single positive finite numeric value.")
    expect_error(summary_c2c(ll, df_old = ll$df.residual, df_new = -1), "`df_new` must be a single positive finite numeric value.")
  }
)

testthat::test_that(
  "summary_c2c rejects unsupported coefficient layout",
  {
    summary.badglm <- function(object, ...) {
      list(coefficients = cbind("Estimate" = c(1, 2), "Std. Error" = c(0.1, 0.2)))
    }
    assign("summary.badglm", summary.badglm, envir = .GlobalEnv)
    on.exit(rm("summary.badglm", envir = .GlobalEnv), add = TRUE)

    bad <- structure(list(df.residual = 10), class = c("badglm", "glm", "lm"))

    expect_error(
      summary_c2c(bad, df_old = 10, df_new = 20),
      "must contain either 't value' or 'z value'"
    )
  }
)

testthat::test_that(
  "summary_c2c rejects non-tabular coefficients object",
  {
    summary.badcoef <- function(object, ...) {
      list(coefficients = c(1, 2, 3))
    }
    assign("summary.badcoef", summary.badcoef, envir = .GlobalEnv)
    on.exit(rm("summary.badcoef", envir = .GlobalEnv), add = TRUE)

    bad <- structure(list(df.residual = 10), class = c("badcoef", "lm"))

    expect_error(
      summary_c2c(bad, df_old = 10, df_new = 20),
      "must be a matrix or data.frame"
    )
  }
)

testthat::test_that(
  "summary_c2c requires Std. Error column",
  {
    summary.nose <- function(object, ...) {
      list(coefficients = cbind("Estimate" = c(1, 2), "t value" = c(3, 4)))
    }
    assign("summary.nose", summary.nose, envir = .GlobalEnv)
    on.exit(rm("summary.nose", envir = .GlobalEnv), add = TRUE)

    bad <- structure(list(df.residual = 10), class = c("nose", "lm"))

    expect_error(
      summary_c2c(bad, df_old = 10, df_new = 20),
      "must contain a 'Std. Error' column"
    )
  }
)

data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup[occup$year == 2008, ]
occup_new <- occup[occup$year == 2010, ]

testthat::test_that("validate_mappings incorrect", {
  testthat::expect_error(
    validate_mappings(list()), "`mappings` must contain 'trans' and 'direction'"
  )
})

testthat::test_that("validate_mappings correct", {
  mappings_simple_back <- list(trans = trans, direction = "backward")
  mappings_simple_for <- list(trans = trans, direction = "forward")
  testthat::expect_silent(validate_mappings(mappings_simple_back))
  testthat::expect_silent(validate_mappings(mappings_simple_for))
})

testthat::test_that("validate_data incorrect", {
  testthat::expect_error(validate_data(list()), "`data\\$old` must be a data.frame")

  data_simple <- list(old = occup_old, new = occup_new, cat_var_old = "code", cat_var_new = "code", time_var = "WRONG")

  testthat::expect_error(
    validate_data(data_simple),
    "`cat_var_old` and `time_var` must be columns in `data\\$old`"
  )

  data_simple <- list(old = occup_old, new = occup_new, cat_var_old = "WRONG", cat_var_new = "code", time_var = "year")

  testthat::expect_error(
    validate_data(data_simple),
    "`cat_var_old` and `time_var` must be columns in `data\\$old`"
  )

  data_simple <- list(old = occup_old, new = occup_new, cat_var_old = "code", cat_var_new = "WRONG", time_var = "year")

  testthat::expect_error(
    validate_data(data_simple),
    "`cat_var_new` and `time_var` must be columns in `data\\$new`"
  )

})

testthat::test_that("validate_data correct", {
  data_simple <- list(
    old = occup_old, new = occup_new, cat_var_old = "code", cat_var_new = "code", time_var = "year"
  )
  testthat::expect_silent(validate_data(data_simple))
})

testthat::test_that("validate_ml incorrect", {
  testthat::expect_error(
    validate_ml(list()), "`ml` must contain 'method', 'features', and 'data'"
  )
})

testthat::test_that("validate_ml correct", {
  ml <- list(
    data = occup_new,
    cat_var = "code",
    method = c("knn", "rf", "lda"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 30)
  )
  testthat::expect_silent(validate_ml(ml))
})

#' Adjusted summary for regressions on replicated datasets
#' @description adjusting lm/glm object results according to original number of
#' degree of freedom.
#' The standard errors, test statistics and p values have to be adjusted because of
#' replicated observations.
#' @param x lm or glm object
#' @param df_old integer number of d.f in original dataset. For bigger datasets
#' `nrow` should be sufficient.
#' @param df_new integer number of d.f in dataset with replicated rows,
#' Default: x$df.residual
#' @return data.frame with additional columns over a regular summary output,
#' like correct and statistics adjusted by it.
#' @importFrom stats pt
#' @details The replication step in \code{cat2cat} inflates the nominal sample
#' size: the model sees \code{n_rep} rows but only \code{n_orig} are independent
#' observations. Naive OLS therefore under-estimates standard errors.
#'
#' The correction factor is \code{sqrt(df_new / df_old)}, where \code{df_new} is
#' the residual d.f. from the replicated model and \code{df_old} the residual
#' d.f. from the original dataset. Standard errors are multiplied by this
#' factor, test statistics divided by it, and p-values recomputed from the
#' appropriate reference distribution: \code{t(df_old)} for t-based summaries
#' and standard normal for z-based summaries.
#'
#' This is a pragmatic d.f. adjustment. It works well when per-subject
#' weights sum to one and no extreme weights dominate.
#'
#' Note: Ordinary R-squared can be interpreted in neutral replication cases
#' where the response and covariates do not vary across replicated copies and
#' per-observation weights sum to the original observation weight. Adjusted
#' R-squared, AIC, and BIC depend on sample-size and degrees-of-freedom
#' conventions, so the values from the replicated model should not be reported
#' without recomputing them on the intended original-observation scale. If the
#' harmonised category itself enters the model, all fit statistics are
#' conditional on the chosen harmonisation weights.
#' @importFrom stats pnorm predict
#' @examples
#' data("occup_small", package = "cat2cat")
#' data("trans", package = "cat2cat")
#'
#' occup_old <- occup_small[occup_small$year == 2008, ]
#' occup_new <- occup_small[occup_small$year == 2010, ]
#'
#' occup_2 <- cat2cat(
#'   data = list(
#'     old = occup_old,
#'     new = occup_new,
#'     cat_var = "code",
#'     time_var = "year"
#'   ),
#'   mappings = list(trans = trans, direction = "backward"),
#'   ml = list(
#'     data = occup_new,
#'     cat_var = "code",
#'     method = "knn",
#'     features = c("age", "sex", "edu", "exp", "parttime", "salary"),
#'     args = list(k = 10)
#'   )
#' )
#'
#' # Regression
#' # we have to adjust size of std as we artificialy enlarge degrees of freedom
#' lms <- lm(
#'   formula = I(log(salary)) ~ age + sex + factor(edu) + parttime + exp,
#'   data = occup_2$old,
#'   weights = multiplier * wei_freq_c2c
#' )
#'
#' summary_c2c(lms, df_old = nrow(occup_old))
#' @export
#'
summary_c2c <- function(x, df_old, df_new = x$df.residual) {
  stopifnot("`x` must be an lm or glm object" = inherits(x, c("lm", "glm")))

  if (!is.numeric(df_old) || length(df_old) != 1 || !is.finite(df_old) || df_old <= 0) {
    stop("`df_old` must be a single positive finite numeric value.")
  }
  if (!is.numeric(df_new) || length(df_new) != 1 || !is.finite(df_new) || df_new <= 0) {
    stop("`df_new` must be a single positive finite numeric value.")
  }

  ss <- summary(x)
  cc <- ss$coefficients

  if (is.null(cc) || !(is.matrix(cc) || is.data.frame(cc))) {
    stop("`summary(x)$coefficients` must be a matrix or data.frame.")
  }

  correct <- sqrt(df_new / df_old)
  dd <- as.data.frame(cc)

  if (!"Std. Error" %in% names(dd)) {
    stop("`summary(x)$coefficients` must contain a 'Std. Error' column.")
  }

  dd$correct <- correct
  dd$std.error_c <- dd$`Std. Error` * correct

  if ("t value" %in% names(dd)) {
    dd$statistic_c <- dd$`t value` / correct
    dd$p.value_c <- 2 * pt(abs(dd$statistic_c), df_old, lower.tail = FALSE)
    dd$reference_dist <- "t"
  } else if ("z value" %in% names(dd)) {
    dd$statistic_c <- dd$`z value` / correct
    dd$p.value_c <- 2 * pnorm(abs(dd$statistic_c), lower.tail = FALSE)
    dd$reference_dist <- "normal"
  } else {
    stop("`summary(x)$coefficients` must contain either 't value' or 'z value'.")
  }

  dd
}

---
title: "Get Started with cat2cat"
author: "Maciej Nasinski"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Get Started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE, message = FALSE, warning = FALSE,
  collapse = TRUE, comment = "#>"
)
```

## What is cat2cat?

**cat2cat** harmonizes categorical variables whose encoding changes over time - occupation codes (ISCO revisions), disease classifications (ICD-9 -> ICD-10), industry codes (NACE), product taxonomies.

| Data type | Example | Function |
|-----------|---------|----------|
| Longitudinal data | Survey responses, administrative records | `cat2cat()` |
| (rotational) Panel data | Subjects observed in multiple periods | `cat2cat()` with `id_var` |
| Aggregated counts | Published statistics, summary tables | `cat2cat_agg()` |

## How to read the documentation

Use the package documentation in this order:

1. **This vignette** for the core problem, the main assumptions, and a minimal two-period workflow.
2. [Choosing Weights and Validating ML](cat2cat_validation.html) when you need to decide between naive, frequency, and ML weights.
3. [Advanced Workflows](cat2cat_advanced.html) for multi-period chaining, (rotational) panels with `id_var`, aggregated data, and regression workflows.

If you only need a standard two-period harmonisation with default frequency weights, this vignette is usually enough.

## The Problem

Classifications evolve. 
When they do, **one old category maps to multiple new ones** (and vice versa):
         
```{r load-data, echo=FALSE}
library(cat2cat)
data(trans, package = "cat2cat")
```

```{r show-problem}
# Old occupation code "1111" became three different codes in the new system
trans[trans$old == "1111", ]
```

A worker coded `1111` in 2008 could be `111101`, `111102`, or `111103` in 2010. 
Which one? The original data doesn't say.

Naive approaches fail:

- Ignoring the problem leads to biased comparisons
- Manual assignment is arbitrary and not reproducible

## The Solution

**cat2cat** uses an approach:

1. The mapping table specifies how categories from one classification (e.g., the old one) correspond to categories from a second classification (e.g., the new one).
2. When one category corresponds to multiple categories in another system, the observation is replicated an equal number of times.
3. The replication probabilities (weights) sum to one and can be constructed using different rules.
4. The weights can be based on frequencies from the previous period, a uniform split, or ML algorithms.

Formally, for each original observation $i$ mapped to $k$ candidate categories, the weights satisfy:

$$w_{ij} \geq 0 \quad \text{and} \quad \sum_{j=1}^{k} w_{ij} = 1$$

This constraint ensures that replication is *neutral* for weighted statistics that do not use the unresolved mapped category.

For estimands that do not directly depend on the unresolved mapped category, the weighted point estimates are unchanged by replication. Inference still needs correction because replication increases the row count. See [Nasinski et al. (2020)](https://doi.org/10.24425/cejeme.2020.134747) for the theoretical foundation and [Nasinski & Gajowniczek (2023)](https://doi.org/10.1016/j.softx.2023.101525) for R implementation details.

### Value Added of cat2cat

cat2cat separates true structural change from coding-system change. This is the
main value for longitudinal analysis.

After harmonisation, you can:

- Track trends within specific groups (for example occupations, industries, diagnoses) across waves
- Compare subgroup dynamics on one consistent coding scheme
- Estimate models with group-level effects and interactions without period-specific recoding artifacts
- Run sensitivity checks across weighting assumptions and report uncertainty transparently

Without harmonisation, measured differences between periods can be driven by
reclassification rules rather than real changes in the population.

## Harmonisation Assumptions and Sensitivity

Suppose a worker is coded "1111" in 2008 and the mapping table says this code can correspond to "111101", "111102", or "111103" in 2010. 
The mapping table tells us the feasible target categories, but it does not tell us which target code this particular observation would have received under the 2010 classification.

This is the key assumption in probabilistic harmonisation. 
`cat2cat` does not discover the true latent split from the mapping table alone. 
Instead, it applies a transparent rule for distributing each ambiguous observation across its feasible target categories. 
Different rules can imply different weights while remaining consistent with the same observed data.

The workflow is therefore sequential. 
First, choose a harmonisation rule: naive, frequency, or ML. 
Conditional on that rule, `cat2cat` constructs the weights. 
Second, estimate the substantive quantity of interest - for example a mean, trend, group comparison, or OLS coefficient - on the harmonised data.

Identification is therefore a property of the estimate or model fitted after harmonisation, not of `cat2cat` alone. 
When the harmonised category enters that model, the chosen `cat2cat` rule becomes part of the model's maintained assumptions.

The weighting rule matters only for observations with `rep_c2c > 1`. 
If a mapping is deterministic and no observation is replicated, the choice between naive, frequency, and ML weights is irrelevant. 
A deterministic recode still relies on the mapping table being correct, but it does not require a distributional assumption over candidate categories.

In anonymous repeated cross-sections, harmonisation is a counterfactual recoding exercise: it asks how an old code should be expressed in the target classification, not whether the same anonymous person kept the same occupation over time. 
In panel data with `id_var`, returning subjects are observed in both periods, so their categories can be linked directly rather than assigned probabilistically.

Sensitivity checks are most important when the final analysis uses the harmonised category itself, for example as a grouping variable, fixed effect, or interaction term. 
If results are similar under naive, frequency, and ML weights, conclusions are less dependent on the harmonisation assumption. If results differ, report the range and explain which assumption drives the difference.
See [Advanced Workflows](cat2cat_advanced.html#regression-on-replicated-data) for the model-level picture.

### Three weighting schemes

Each weighting method adds one harmonisation assumption:

**Naive weights** (`wei_naive_c2c`): Assume a uniform distribution - each candidate is equally likely.
$$w_j = \frac{1}{k}$$
This is the maximum entropy (least informative) prior. It requires only the mapping table - no data from either period. Use when you have no information favoring any candidate, or as a robustness bound.

**Frequency weights** (`wei_freq_c2c`): Assume that ambiguous observations are distributed as in the base-period population.
$$w_j = \frac{n_j^{\text{base}}}{\sum_{j'} n_{j'}^{\text{base}}}$$
This requires observed counts in the base period. It works well when ambiguous cases resemble the overall population; it breaks down when they differ systematically (e.g., workers in transitional occupations may not match the stable population).

**ML weights** (`wei_*_c2c` for knn, rf, lda): Assume that individual features predict category membership.
$$w_j = P(j \mid X_i, g)$$
This requires training data with both the category and predictive features. ML can correct for heterogeneity that frequency weights miss - if a 25-year-old programmer is more likely in "111101" than a 55-year-old manager, ML captures this. Use `cat2cat_ml_run()` to verify ML improves over baselines.

### When each assumption matters

| Method | Assumption fails when... | Practical check |
|--------|--------------------------|-----------------|
| Naive | Distribution is highly skewed | Compare `wei_naive_c2c` vs `wei_freq_c2c` - large differences suggest uniform is wrong |
| Frequency | Ambiguous cases differ from population | Compare ML vs freq accuracy in `cat2cat_ml_run()` |
| ML | Features don't predict category | ML accuracy ≈ frequency accuracy in `cat2cat_ml_run()` |

### Reducing assumption dependence

The weighting assumption only affects observations that are replicated (`rep_c2c > 1`). 

Strategies to reduce replication:

- **Panel `id_var`**: returning subjects get direct matches - no replication. See [Advanced Workflows](cat2cat_advanced.html#panel-data-with-subject-identifiers).
- **Forward mapping**: replicates the newer (more granular) period, which often maps 1-to-1 backward.
- **Code truncation**: collapsing 6-digit to 4-digit codes often creates more 1-to-1 mappings in hierarchical systems (ISCO, ICD, NACE), especially for backward mapping. For forward mapping, truncation can also increase replication in some mappings.

If most observations have `rep_c2c = 1`, results depend minimally on the assumption.

**Sensitivity check**: run analyses with `wei_naive_c2c`, `wei_freq_c2c`, and ML weights. Similar conclusions -> robust. Divergent -> report the range.

## When cat2cat won't help

**Blockers** (cat2cat cannot be applied):

- **No mapping table**: cat2cat requires a crosswalk. Alternative: expert coding or separate analyses per period.
- **Unobserved category**: cat2cat harmonizes *observed* codes. Missing category is a missing-data problem - impute first.

**Method-specific limits** (workarounds exist):

- **Empty base categories**: frequency weights fall back to naive (1/k). Use `wei_naive_c2c` or ML with external training data.
- **Few observations per ML class**: models fail or overfit. Use frequency weights or coarser categories.

Naive weights (`wei_naive_c2c`) always work - they require only the mapping table.

**Diagnostic**: 

check `mean(rep_c2c)` and `cor(wei_freq_c2c, wei_naive_c2c)`. 
Very high replication or correlation near 1 suggests the mapping is too diffuse to improve over uniform weights. 
See [Advanced Workflows](cat2cat_advanced.html#mapping-table-stability-and-truncation) for truncation strategies.

## If you need something more advanced

This vignette focuses on the standard two-period micro-data workflow. Move to the other vignettes when:

- you need to decide whether ML improves on simpler baselines,
- you need to chain mappings across 3 or more waves,
- you have a rotational panel with stable identifiers,
- you only have aggregated counts rather than individual observations,
- or your main task is regression and inference after harmonisation.

## Use Cases

**Repeated cross-sections** (most common): 
harmonize independent surveys using base-period frequencies. 
The Quick Example below demonstrates this.

**Multi-period analysis** (3+ waves): 
chain `cat2cat()` calls sequentially - each step uses the previous step's mapped frequencies. 
See [Advanced Workflows](cat2cat_advanced.html#multi-period-chaining).

**Panel data** (subjects observed in multiple periods): 
provide `id_var` to directly match returning subjects - no replication for them, only new entrants go through the probabilistic path. See [Advanced Workflows](cat2cat_advanced.html#panel-data-with-subject-identifiers).

If every subject is observed in both periods and the target-period category is known, `cat2cat()` may not be necessary: you can often join the target category back to the earlier record by `id_var`. `cat2cat()` is useful when some records cannot be linked directly and still need the mapping-table replication path.

**Aggregated data** (published counts only, no micro-data): 
use `cat2cat_agg()` with explicit equations:

```{r use-case-agg}
library(dplyr)
data(verticals, package = "cat2cat")

agg_old <- verticals[verticals$v_date == "2020-04-01", ]
agg_new <- verticals[verticals$v_date == "2020-05-01", ]

agg_result <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  # Backward mapping: old Automotive split into Automotive1 + Automotive2
  Automotive %<% c(Automotive1, Automotive2),
  # Forward mapping: Kids1 + Kids2 merged into Kids  
  c(Kids1, Kids2) %>% c(Kids)
)

agg_result$old[c("vertical", "prop_c2c", "counts")]
```

See [Advanced Workflows](cat2cat_advanced.html#aggregated-data-and-special-cases) for details.

## Key Concepts

### Direction: Backward vs Forward

| Direction | Base period | Replicated period | Result encoding |
|-----------|-------------|-------------------|-----------------|
| `"backward"` | NEW | OLD | New (modern) codes |
| `"forward"` | OLD | NEW | Old (legacy) codes |

```{r direction-comparison}
# Setup for comparison
occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]

# Backward: old period gets replicated onto new codes
backward <- cat2cat(
  data = list(old = occup_2008, new = occup_2010, 
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)

# Forward: new period gets replicated onto old codes
forward <- cat2cat(
  data = list(old = occup_2008, new = occup_2010, 
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "forward")
)

# Which period gets replicated depends on direction
cat("Backward: old period replicated from", nrow(occup_2008), "to", nrow(backward$old), "rows\n")
cat("Forward: new period replicated from", nrow(occup_2010), "to", nrow(forward$new), "rows")
```

### ML weights in practice

Model-based weights are optional in `cat2cat()`. 
They are useful when individual features (age, education, experience, etc.) carry information about category assignment beyond base-period frequencies.

For a compact ML workflow (`ml` setup, method comparison, and failure handling with `on_fail` / `fail_warn`), see [Advanced Workflows](cat2cat_advanced.html#ml-weights).

### Weights

| Weight | Source | Use when... |
|--------|--------|-------------|
| `wei_freq_c2c` | Base period frequencies | Default choice |
| `wei_naive_c2c` | Uniform (1/k) | Robustness checks |
| `wei_knn_c2c` | k-Nearest Neighbours on features | Non-linear boundaries, no distributional assumption |
| `wei_rf_c2c` | Random Forest on features | Feature interactions, larger training sets |
| `wei_lda_c2c` | Linear Discriminant Analysis | Fast, assumes normality & equal covariance |
| `wei_nb_c2c` | Naive Bayes on features | Strong independence assumption |

ML weights (`wei_knn_c2c`, `wei_rf_c2c`, `wei_lda_c2c`, `wei_nb_c2c`) are added only when you pass an `ml` argument to `cat2cat()`. 
For method selection, holdout diagnostics (`cat2cat_ml_run()`), and failed-ML handling (`on_fail`, `fail_warn`), see [Choosing Weights and Validating ML](cat2cat_validation.html).

## Quick Example

First, map 2008 observations backward onto the 2010 coding scheme:

```{r quick-example}
data(occup, package = "cat2cat")

occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]

result_back <- cat2cat(
  data = list(
    old = occup_2008,
    new = occup_2010,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward")
)
```

**What happened?** 2008 observations were replicated onto 2010 category codes. 
One worker may appear multiple times with different `g_new_c2c` values:

```{r show-replication}
# A replicated observation (rep_c2c > 1 means replicated)
result_back$old[result_back$old$rep_c2c > 1, ][1:3,
                c("code", "g_new_c2c", "wei_freq_c2c", "rep_c2c")]
```

### Forward mapping example

If instead you want one common 2008-style coding scheme, map 2010 observations forward onto the older codes:

```{r forward-example}
result_forward <- cat2cat(
  data = list(
    old = occup_2008,
    new = occup_2010,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "forward")
)

result_forward$new[result_forward$new$rep_c2c > 1, ][1:3,
                   c("code", "g_new_c2c", "wei_freq_c2c", "rep_c2c")]
```

Now the 2010 observations are replicated onto 2008-style codes. Forward mapping is often attractive when the newer classification is more detailed and you prefer the older, coarser coding scheme, but it does not guarantee fewer replications in every dataset.

### Naive vs Frequency Weights

Both types of weights are always available. 
Naive weights (`wei_naive_c2c`) assign equal probability to each candidate and serve as a useful robustness baseline:

```{r naive-vs-freq}
# Compare weights for a replicated observation
result_back$old[result_back$old$rep_c2c > 1, ][1:3,
                c("g_new_c2c", "wei_naive_c2c", "wei_freq_c2c")]

# Same analysis with naive weights (robustness check)
c(freq_mean = weighted.mean(result_back$old$salary, result_back$old$wei_freq_c2c),
  naive_mean = weighted.mean(result_back$old$salary, result_back$old$wei_naive_c2c))
```

If results are similar, conclusions don't depend on the distributional assumption. 
Large differences warrant investigation.

### Value added: pooled regression across both periods

The practical payoff is that you can now combine both waves in one regression while keeping a common occupation classification. 
Without harmonisation, a pooled model with one set of occupation effects would not make sense because the 2008 and 2010 codes are not directly comparable.

Replication inflates the row count, so standard errors must still be corrected:

$$\text{SE}_{\text{corrected}} = \text{SE}_{\text{naive}} \times \sqrt{\frac{n_{\text{rep}}}{n_{\text{orig}}}}$$

`summary_c2c()` applies this correction to the pooled model.
To keep this example fast, we use a lightweight group-level control
(`avg_age_g_new_c2c`) instead of full `factor(g_new_c2c)` fixed effects:

```{r pooled-regression}
harmonised_two_period <- dplyr::bind_rows(result_back$old, result_back$new)

harmonised_two_period <- harmonised_two_period %>%
  dplyr::group_by(g_new_c2c) %>%
  dplyr::mutate(avg_age_g_new_c2c = mean(age, na.rm = TRUE)) %>%
  dplyr::ungroup()

pooled_model <- lm(
  log(salary) ~ factor(year) + age + exp + avg_age_g_new_c2c,
  data = harmonised_two_period,
  weights = multiplier * wei_freq_c2c
)

pooled_summary <- summary_c2c(
  pooled_model,
  df_old = nrow(occup_2008) + nrow(occup_2010) - length(coef(pooled_model))
)

pooled_summary[c("factor(year)2010", "age", "avg_age_g_new_c2c"),
               c("Estimate", "std.error_c", "p.value_c")]
```

This is the value added of `cat2cat`: you can estimate a period effect after controlling for one harmonised occupation structure instead of running separate regressions under incompatible code systems.
Use the corrected coefficient table for inference. Ordinary $R^2$ is preserved
in neutral replication cases like this, where the response and covariates do not
vary across replicated copies and the weights for each source observation sum to
the original weight. Adjusted $R^2$, AIC, and BIC depend on sample-size and
degrees-of-freedom conventions, so do not report their replicated-model values
unless you recompute them on the intended original-observation scale.

## Diagnostic Plot

```{r plot, fig.width=7, fig.height=3.5, fig.alt="Diagnostic plot displaying weight histograms and replication statistics for mapped observations"}
plot_c2c(result_back$old, type = "both")
```

## Mapping table

Mapping table is a data frame with columns `old` and `new` defining the mapping between old and new categories.
The mapping table is usually provided by the classification authority (e.g., statistical office).
Often the classification is evolutionary, so the new codes are more detailed (e.g. more digits) than the old ones.
In hierarchical classifications each digit adds a level of detail, so truncating to fewer digits creates coarser groupings with fewer one-to-many relationships.

### Common hierarchical classifications

| Classification | Domain | Hierarchy |
|---------------|--------|-----------|
| ISCO | Occupations | 1-digit -> 2-digit -> 3-digit -> 4-digit |
| ICD | Diseases | Chapter -> Block -> Category -> Subcategory |
| NACE | Industries | Section -> Division -> Group -> Class |
| CPC | Products | Section -> Division -> Group -> Class -> Subclass |

### Truncating mapping tables from hierarchical codes

For example, in the `trans` table, old codes are 4 digits and new codes are 6 digits.
You can construct a coarser mapping by truncating to the first N digits:

```{r hierarchical-codes}
head(trans, 5)

# Build a 3-digit mapping from the full codes
trans_3digit <- data.frame(
  old = substr(trans$old, 1, 3),
  new = substr(trans$new, 1, 3)
)
trans_3digit <- unique(trans_3digit)
cat("3-digit mapping rows:", nrow(trans_3digit),
    "vs full mapping rows:", nrow(trans))
```

Truncating codes to fewer digits creates coarser mappings and often reduces many-to-many relationships (especially for backward mapping), but the unified categories are broader. 
Under forward mapping, truncation can also increase replication in some cases.
This approach works for any classification where codes share a hierarchical prefix structure (ISCO, ICD, NACE, etc.).

## Learn More

| Topic | Vignette |
|-------|----------|
| Choosing between naive, frequency, and ML weights | [Choosing Weights and Validating ML](cat2cat_validation.html) |
| Multi-period chaining, panels, aggregates, and regression | [Advanced Workflows](cat2cat_advanced.html) |

---
title: "Advanced Workflows"
author: "Maciej Nasinski"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Advanced Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE, message = FALSE, warning = FALSE,
  collapse = TRUE, comment = "#>"
)
```

This vignette collects the advanced `cat2cat` workflows in one place: 
ML weights, multi-period chaining, panel data with identifiers, aggregated data, and regression on replicated data. 
It assumes you've read [Get Started](cat2cat.html).

Use this vignette by module:

1. **ML weights** if you want feature-informed probabilities rather than only naive or frequency weights.
2. **Multi-period chaining** if your harmonisation spans 3 or more waves.
3. **Panel data with subject identifiers** if you have a rotational panel or other stable subject IDs.
4. **Aggregated data and special cases** if you only observe category totals or need hierarchical-code mappings.
5. **Regression on replicated data** if your main task is estimation and inference after harmonisation.

```{r load-data}
library(cat2cat)
library(dplyr)
library(tidyr)
library(fixest)

data(occup, package = "cat2cat")
data(occup_panel, package = "cat2cat")
data(trans, package = "cat2cat")
data(verticals, package = "cat2cat")
data(verticals2, package = "cat2cat")

occup_2006 <- occup[occup$year == 2006, ]
occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]
```

## ML weights

Machine-learning weights are useful when the mapping table alone is too coarse and observed features help distinguish which target category is most plausible for a replicated observation.

This section shows:

- when ML weights are worth trying,
- how to specify the `ml` argument in `cat2cat()`,
- how to validate whether ML improves on simpler baselines,
- and how to handle rows where ML probabilities cannot be produced.

In practice, ML is most helpful when replication is substantial and the ambiguous categories differ systematically by observed characteristics such as age, education, experience, or salary. If ML does not improve on the frequency baseline in `cat2cat_ml_run()`, the simpler `wei_freq_c2c` weights are usually the better choice.

ML features must be numeric, logical, or factor columns. Factor columns are
one-hot encoded automatically; character columns are not, so convert character
categories to factors before listing them in `features`.

```{r ml-minimal, eval=FALSE}
ml_setup <- list(
  data = bind_rows(occup_2010, occup_2012),
  cat_var = "code",
  method = c("knn", "rf", "lda"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 50),
  on_fail = "freq",
  fail_warn = TRUE
)

result_ml <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)
```

Validate whether ML adds value before using it in production:

```{r ml-validate, eval=FALSE}
cv <- cat2cat_ml_run(
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)
print(cv)
```

Baseline-only diagnostics are also available:

```{r baseline-only-validate, eval=FALSE}
ml_baseline <- list(
  data = bind_rows(occup_2010, occup_2012),
  cat_var = "code",
  method = character(0),
  features = character(0)
)

cv_baseline <- cat2cat_ml_run(
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_baseline
)
print(cv_baseline)
```

Use the same `direction` in diagnostics as in the mapping workflow you want to
evaluate, because the mapping groups and base-period frequencies differ by
direction.

If ML probabilities cannot be produced for some replicated rows, use `on_fail` and `fail_warn`:

- `on_fail = "freq"` (default): replace failed ML rows with `wei_freq_c2c`
- `on_fail = "naive"`: replace failed ML rows with `wei_naive_c2c`
- `on_fail = "na"`: keep failed ML rows as `NA`
- `on_fail = "error"`: stop immediately
- `fail_warn = TRUE` (default): warn with affected rows/observations per method
- `fail_warn = FALSE`: silence the warning

## Multi-period chaining

Repeated cross-sections or longitudinal datasets often contain more than two waves. To bring all waves onto the same encoding, apply `cat2cat()` iteratively, feeding the output of one step into the next.

Use this module when you need one harmonised categorical variable across 3 or more periods. The main design choice is whether to chain backward or forward, and whether truncating hierarchical codes can reduce replication before chaining.

In the examples below, `occup` is a repeated cross-section dataset, not a panel. The chained outputs therefore combine harmonised cross-sections from multiple years.

### Optional: passing `mappings$freqs_df`

In most cases you do **not** need to pass `mappings$freqs_df`.
If it is omitted, `cat2cat()` computes base-period frequencies internally from the provided data.

Pass `mappings$freqs_df` only when you need explicit control over base frequencies.

The object should be a two-column data frame: category name in the first column and counts in the second.

### Mapping table stability and truncation

Before chaining, inspect how disruptive the transition table is:

```{r trans-stability}
max_digits <- max(nchar(as.character(trans[[1]])), nchar(as.character(trans[[2]])))

stability <- sapply(1:max_digits, function(d) {
  old_trunc <- substr(as.character(trans[[1]]), 1, d)
  new_trunc <- substr(as.character(trans[[2]]), 1, d)
  mean(old_trunc != new_trunc) * 100
})

data.frame(
  digits = 1:max_digits,
  pct_changed = round(stability, 1)
)
```

Truncating codes to fewer digits can reduce replication:

For backward mapping this is often true, but for forward mapping truncation can move in the opposite direction. Collapsing detailed codes into broader prefixes can create additional many-to-one and one-to-many ties on the old-code side, which may increase replication in the mapped period.

```{r truncation-diagnostic}
occup_2008_trunc <- occup_2008
occup_2008_trunc$code <- substr(occup_2008_trunc$code, 1, 4)
occup_2010_trunc <- occup_2010
occup_2010_trunc$code <- substr(occup_2010_trunc$code, 1, 4)
trans_trunc <- unique(data.frame(
  old = substr(trans$old, 1, 4),
  new = substr(trans$new, 1, 4)
))

back_full <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)
back_trunc <- cat2cat(
  data = list(old = occup_2008_trunc, new = occup_2010_trunc,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans_trunc, direction = "backward")
)

fwd_full <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "forward")
)
fwd_trunc <- cat2cat(
  data = list(old = occup_2008_trunc, new = occup_2010_trunc,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans_trunc, direction = "forward")
)

data.frame(
  mapping = c("backward full (4->6)", "backward trunc (4->4)",
              "forward full (6->4)", "forward trunc (4->4)"),
  mean_rep = c(mean(back_full$old$rep_c2c), mean(back_trunc$old$rep_c2c),
               mean(fwd_full$new$rep_c2c), mean(fwd_trunc$new$rep_c2c))
)
```

### Backward chaining

```{r backward-chain}
step1 <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)

step2 <- cat2cat(
  data = list(old = occup_2006, new = step1$old,
              cat_var_old = "code", cat_var_new = "g_new_c2c",
              time_var = "year"),
  mappings = list(trans = trans, direction = "backward")
)

harmonised_back <- bind_rows(
  step2$old,
  step1$old,
  step1$new,
  dummy_c2c(occup_2012, "code")
)
```

Validation: weighted counts should match the original counts within each year.

```{r validate-backward-counts}
harmonised_back %>%
  group_by(year) %>%
  summarise(weighted_n = round(sum(wei_freq_c2c)), .groups = "drop") %>%
  left_join(count(occup, year), by = "year")
```

### Forward chaining

```{r forward-chain}
trans_fwd <- rbind(
  trans,
  data.frame(old = "no_cat",
             new = setdiff(c(occup_2010$code, occup_2012$code), trans$new))
)

fwd1 <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans_fwd, direction = "forward")
)

fwd2 <- cat2cat(
  data = list(old = fwd1$new, new = occup_2012,
              cat_var_old = "g_new_c2c", cat_var_new = "code",
              time_var = "year"),
  mappings = list(trans = trans_fwd, direction = "forward")
)

harmonised_fwd <- bind_rows(
  dummy_c2c(occup_2006, "code"),
  fwd1$old,
  fwd1$new,
  fwd2$new
)
```

### Adding ML to the chain

```{r ml-chain, eval=FALSE}
step1_ml <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)

step2_ml <- cat2cat(
  data = list(old = occup_2006, new = step1_ml$old,
              cat_var_old = "code", cat_var_new = "g_new_c2c",
              time_var = "year"),
  mappings = list(trans = trans, direction = "backward"),
  ml = ml_setup
)
```

## Panel data with subject identifiers

If subjects have stable identifiers across waves, `id_var` can reduce unnecessary replication by directly matching returning subjects.

Use this module only when the identifier truly tracks the same subject across adjacent waves and short-run category changes are unlikely to represent genuine transitions rather than coding changes.

If you have a complete panel with every subject observed in both periods and no missing category values, you may not need probabilistic harmonisation at all. In that case, the target-period category can often be joined back to the earlier record by `id_var`, and the task is mostly a deterministic join. `cat2cat()` is more useful when the panel is incomplete, rotational, or mixed with new entrants and leavers, so some observations still need the mapping-table replication path.

```{r panel-load}
panel_old <- occup_panel[occup_panel$quarter == "2009Q4", ]
panel_new <- occup_panel[occup_panel$quarter == "2010Q1", ]
shared_ids <- intersect(panel_old$panel_id, panel_new$panel_id)
length(shared_ids)
```

```{r idvar-mapping}
result_id <- cat2cat(
  data = list(
    old = panel_old,
    new = panel_new,
    id_var = "panel_id",
    cat_var = "code",
    time_var = "quarter"
  ),
  mappings = list(trans = trans, direction = "backward")
)
```

How `id_var` works:

- direct match: workers observed in both periods receive `rep_c2c = 1` and weight 1,
- replication path: workers observed in only one period go through the standard mapping-table replication.

```{r idvar-inspect}
table(result_id$old$rep_c2c)
sum(result_id$old$wei_freq_c2c)
nrow(panel_old)
```

Compare with and without identifiers:

```{r idvar-compare}
result_no_id <- cat2cat(
  data = list(
    old = panel_old,
    new = panel_new,
    cat_var = "code",
    time_var = "quarter"
  ),
  mappings = list(trans = trans, direction = "backward")
)

cat("WITH id_var average replication:", round(mean(result_id$old$rep_c2c), 2), "\n")
cat("WITHOUT id_var average replication:", round(mean(result_no_id$old$rep_c2c), 2), "\n")
```

Use `id_var` when:

- identifiers are reliable,
- the time gap is short relative to true mobility,
- and direct matches are informative about the coding change.

## Aggregated data and special cases

When only aggregate counts are available, use `cat2cat_agg()` with mapping equations rather than micro-data replication.

Use this module when you do not have person-level data, or when the classification itself has a hierarchical code structure that can be exploited to build coarser mappings.

```{r agg-data}
agg_old <- verticals[verticals$v_date == "2020-04-01", ]
agg_new <- verticals[verticals$v_date == "2020-05-01", ]
```

```{r agg-mapping}
agg <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  Automotive %<% c(Automotive1, Automotive2),
  c(Kids1, Kids2) %>% c(Kids),
  Home %>% c(Home, Supermarket)
)
```

Inspect how categories were proportionally redistributed:

```{r agg-inspect}
agg$old[agg$old$vertical %in% c("Automotive1", "Automotive2"), ]
agg$new[agg$new$vertical %in% c("Kids1", "Kids2"), ]
```

Hierarchical codes can also be used to build coarser mapping tables when an official transition table is unavailable:

```{r hierarchical-codes}
trans_2digit <- data.frame(
  old = substr(trans$old, 1, 2),
  new = substr(trans$new, 1, 2)
)
trans_2digit <- unique(trans_2digit)

cat("2-digit mapping rows:", nrow(trans_2digit),
    "vs full mapping rows:", nrow(trans))
```

This works for classifications with stable prefix hierarchies such as ISCO, ICD, NACE, CPC, or HS codes.

## Regression on replicated data

The replication is neutral for regressions on non-mapped covariates because per-subject weights sum to one. Standard errors, however, must be corrected because replication inflates the row count.

Use this module when your end goal is estimation rather than descriptive harmonisation. The key issue is not coefficient bias for non-mapped regressors, but valid inference after replication.

### Building a 4-period harmonised repeated cross-section dataset

```{r build-panel}
cat2cat_data_back <- bind_rows(
  step2$old,
  step1$old,
  step1$new,
  dummy_c2c(occup_2012, "code")
)
```

### Neutral impact demonstration

```{r regression-neutral}
lms_orig <- lm(
  I(log(salary)) ~ age + sex + factor(edu) + parttime + exp,
  data = occup,
  weights = multiplier
)

lms_harmonised <- lm(
  I(log(salary)) ~ age + sex + factor(edu) + parttime + exp,
  data = cat2cat_data_back,
  weights = multiplier * wei_freq_c2c
)

summary_c2c(lms_harmonised, df_old = nrow(occup))
```

`summary_c2c()` scales naive standard errors by the replication factor:

$$\text{SE}_{\text{corrected}} = \text{SE}_{\text{naive}} \times \sqrt{\frac{n_{\text{rep}}}{n_{\text{orig}}}}$$

Report coefficient estimates with corrected standard errors and p-values from
`summary_c2c()`. Ordinary $R^2$ is preserved in this neutral setup because the
response and covariates do not vary across replicated copies and the weights for
each source observation sum to the original weight. Adjusted $R^2$, AIC, and BIC
depend on sample-size and degrees-of-freedom conventions, so do not report their
replicated-model values unless you recompute them on the intended
original-observation scale. If the harmonised category itself enters the model,
fit statistics are conditional on the chosen harmonisation weights.

### Fixed effects regression

```{r regression-fe}
harmonised_fe <- cat2cat_data_back %>%
  prune_c2c(method = "nonzero") %>%
  mutate(orig_obs_id = interaction(year, index_c2c, drop = TRUE, lex.order = TRUE)) %>%
  filter(!is.na(g_new_c2c), !is.na(salary), salary > 0)

fe_model_cluster <- feols(
  log(salary) ~ age + sex + factor(edu) + parttime + exp | g_new_c2c + year,
  data = harmonised_fe,
  weights = ~multiplier * wei_freq_c2c,
  cluster = ~orig_obs_id
)
summary(fe_model_cluster)
```

Do not cluster directly on `index_c2c` after binding multiple waves, because `index_c2c` is created separately inside each `cat2cat()` call and can repeat across years. 
Instead, build a wave-specific original-observation identifier such as `interaction(year, index_c2c)` and cluster on that. This treats all replications of the same source row as one cluster without incorrectly merging different people from different waves.

## Choosing the right advanced workflow

| Problem | Recommended tool |
|--------|-------------------|
| Need feature-informed weights | `cat2cat(..., ml = ...)` + `cat2cat_ml_run()` |
| Need 3+ wave harmonisation | iterative `cat2cat()` chaining |
| Stable subject identifiers across waves | `id_var` |
| Only aggregated counts available | `cat2cat_agg()` |
| Regression on replicated data | `summary_c2c()` or clustered inference |

## Next steps

- [Get Started](cat2cat.html) for core concepts and the two-period workflow
- [Choosing Weights and Validating ML](cat2cat_validation.html) for weight comparisons and robustness checks

---
title: "Choosing Weights and Validating ML"
author: "Maciej Nasinski"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Choosing Weights and Validating ML}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE, message = FALSE, warning = FALSE,
  collapse = TRUE, comment = "#>"
)
```

This vignette is a decision guide for choosing and checking weights in `cat2cat()`.

Read it when you want to answer one of these questions:

- Are naive and frequency weights telling the same story?
- Is ML worth trying at all?
- If ML is used, does it improve on the frequency baseline?
- What should I do when different weight methods disagree?
- How should I handle failed ML predictions?

If you only need the basic two-period workflow, go back to [Get Started](cat2cat.html). If you need multi-period, panel, aggregated, or regression workflows, continue to [Advanced Workflows](cat2cat_advanced.html).

```{r load-data}
library(cat2cat)
library(dplyr)
library(tidyr)
library(e1071)
library(randomForest)

data(occup, package = "cat2cat")
data(trans, package = "cat2cat")

occup_2008 <- occup[occup$year == 2008, ]
occup_2010 <- occup[occup$year == 2010, ]
occup_2012 <- occup[occup$year == 2012, ]
```

## Step 1: Understand the competing weight assumptions

`cat2cat` offers several ways to assign probability weights to replicated observations. Each method encodes a different **distributional assumption** about how ambiguous observations split across candidate categories. When a downstream estimand depends on the mapped category, this is the identifying assumption for that estimand - so always check sensitivity.

**Naive weights** (`wei_naive_c2c`) are always computed. Each replicated observation gets uniform probability $1/k$ where $k$ is the number of candidate categories.

- *Assumption*: All candidates equally likely (maximum entropy / uninformative prior)
- *Requires*: Only the mapping table - no data from either period
- *Use when*: No information favoring any candidate, or as a robustness lower bound

**Frequency-based weights** (`wei_freq_c2c`) are the default. They use category counts from the base period.

- *Assumption*: Ambiguous observations distribute like the base period population
- *Requires*: Observed counts in base period (falls back to naive if all zero)
- *Use when*: Base period is large and representative; ambiguous cases resemble the general population

**ML weights** (`wei_knn_c2c`, `wei_lda_c2c`, `wei_rf_c2c`, `wei_nb_c2c`) use individual features to predict category membership.

- *Assumption*: Features (age, education, etc.) predict true category: $P(j \mid X, g)$
- *Requires*: Training data with both category labels and predictive features
- *Use when*: Features are informative - verify with `cat2cat_ml_run()`

Available ML methods:

- **knn**: k-Nearest Neighbours. A non-parametric method that handles non-linear boundaries. Sensitive to the choice of `k`.
- **lda**: Linear Discriminant Analysis. Fast, interpretable. Assumes multivariate normality and equal covariance.
- **rf**: Random Forest. Handles interactions well. Slower, needs `ntree` tuning.
- **nb**: Naive Bayes via `e1071`. Fast, useful after numeric/logical/factor preprocessing. Assumes conditional independence of features.

ML features must be numeric, logical, or factor columns. Factor columns are
one-hot encoded automatically using levels observed in the training data and
the target period. Character columns are not encoded automatically; convert
them to factors first if they represent categories.

You can run multiple methods at once and compare or combine them:

```{r mixed-ml-weights}
occup_2_mix <- cat2cat(
  data = list(
    old = occup_2008, new = occup_2010,
    cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_2010,
    cat_var = "code",
    method = c("knn", "rf", "lda", "nb"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50),
    on_fail = "na"
  )
)
```

Correlations between weight methods:

```{r weight-correlations}
occup_2_mix$old %>%
  select(wei_knn_c2c, wei_rf_c2c, wei_lda_c2c, wei_nb_c2c, wei_freq_c2c, wei_naive_c2c) %>%
    cor(use = "pairwise.complete.obs")
```

### If ML fails on some rows: `on_fail` and `fail_warn`

Sometimes ML probabilities cannot be produced for a subset of replicated rows
(for example incomplete target features or method-specific prediction failures).
`cat2cat()` exposes explicit policy controls in `ml`:

- `on_fail = "freq"` (default): failed ML rows are filled with `wei_freq_c2c`.
- `on_fail = "naive"`: failed ML rows are filled with `wei_naive_c2c`.
- `on_fail = "na"`: failed ML rows are kept as `NA`.
- `on_fail = "error"`: stop immediately when failed rows are detected.
- `fail_warn = TRUE` (default): warn with affected rows/observations per method.
- `fail_warn = FALSE`: suppress these warnings.

Important: this failure accounting is specific to `cat2cat()` and the constructed
weight columns (`wei_*_c2c`). It is different from `cat2cat_ml_run()` "SKIPPED
GROUPS", which reports mapping groups that were not evaluated in holdout
diagnostics (single category, too few observations, or method fit/predict error
for that group).

```{r ml-failure-policy, eval=FALSE}
ml_setup <- list(
  data = bind_rows(occup_2010, occup_2012),
  cat_var = "code",
  method = c("knn", "rf", "lda"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 50),
  on_fail = "freq",   # default policy
  fail_warn = TRUE     # default reporting
)

# strict mode for QA pipelines
ml_strict <- ml_setup
ml_strict$on_fail <- "error"

# diagnostic mode to inspect failures directly
ml_diag <- ml_setup
ml_diag$on_fail <- "na"
ml_diag$fail_warn <- FALSE
```

Ensemble weights with `cross_c2c()` and pruning with `prune_c2c()`:

```{r ensemble-prune}
occup_old_mix <- occup_2_mix$old %>%
  cross_c2c(.) %>%
  prune_c2c(., column = "wei_cross_c2c", method = "nonzero")
```

## Step 2: Check whether conclusions are sensitive to the weight choice

Different weight methods affect regression coefficients when you filter to a specific occupation group and combine both periods. This is the proper sensitivity analysis: subjects from the base period (new, no replication) plus subjects from the target period (old, weighted by probability of belonging to this group).

### Compare weight methods on the same mapped data

Run backward mapping with all ML methods:

```{r sensitivity-result}
result_all <- cat2cat(
  data = list(old = occup_2008, new = occup_2010,
              cat_var = "code", time_var = "year"),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_2010, cat_var = "code",
    method = c("knn", "rf", "lda", "nb"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50)
  )
)
```

**Weighted counts per group** - compare how weight methods redistribute observations:
 
```{r sensitivity-counts}
weight_cols <- c("wei_naive_c2c", "wei_freq_c2c", "wei_knn_c2c", "wei_rf_c2c", "wei_lda_c2c", "wei_nb_c2c")

# Pick groups with high replication
top_groups <- result_all$old %>%
  filter(rep_c2c > 1) %>%
  count(g_new_c2c, sort = TRUE) %>%
  head(6) %>%
  pull(g_new_c2c)

# Weighted counts from OLD period (replicated)
old_counts <- lapply(weight_cols, function(wcol) {
  result_all$old %>%
    filter(g_new_c2c %in% top_groups) %>%
    group_by(g_new_c2c) %>%
    summarise(n = sum(.data[[wcol]]), .groups = "drop")
}) %>%
  setNames(gsub("wei_|_c2c", "", weight_cols)) %>%
  bind_rows(.id = "method") %>%
  tidyr::pivot_wider(names_from = method, values_from = n)

# Counts from NEW period (no replication, exact)
new_counts <- result_all$new %>%
  filter(code %in% top_groups) %>%
  count(code, name = "new_period") %>%
  rename(g_new_c2c = code)

# Combine for comparison
left_join(old_counts, new_counts, by = "g_new_c2c")
```

The `new_period` column shows the actual counts in 2010. The other columns show how the 2008 observations are redistributed under each weight method. `naive` assigns uniform probability (1/n candidates), `freq` uses base period frequencies, and ML methods (`knn`, `rf`, `lda`, `nb`) use predicted probabilities.

Pick a specific group for regression analysis:

```{r sensitivity-weights}
# New-period counts per category (no replication, so plain tally)
new_counts_all <- result_all$new %>%
  count(code, name = "n_new") %>%
  rename(g_new_c2c = code)

# Old-period weighted counts, joined to new-period counts
group_sizes <- result_all$old %>%
  group_by(g_new_c2c) %>%
  summarise(n_old = sum(wei_freq_c2c), .groups = "drop") %>%
  left_join(new_counts_all, by = "g_new_c2c") %>%
  filter(n_old >= 10, n_new >= 10) %>%
  arrange(desc(n_old))

# Pick a group for regression analysis
target_group <- group_sizes$g_new_c2c[1]
cat("Analysing occupation group:", target_group, "\n")
```

**Regression within a single occupation group** - combine both periods and compare coefficients:

```{r sensitivity-group-reg}
# Subset old period to target group (with weights)
old_subset <- result_all$old %>%
  filter(g_new_c2c == target_group)

# Subset new period to target group (no replication, weight = 1)
new_subset <- result_all$new %>%
  filter(code == target_group) %>%
  mutate(
    wei_naive_c2c = 1, wei_freq_c2c = 1, wei_knn_c2c = 1,
    wei_rf_c2c = 1, wei_lda_c2c = 1, wei_nb_c2c = 1
  )

# Combine both periods
d <- bind_rows(old_subset, new_subset)

# Compare all regression coefficients across weight methods
f <- I(log(salary)) ~ age + sex + factor(edu) + exp + parttime

coefs <- sapply(weight_cols, function(wcol) {
  d$w <- d$multiplier * d[[wcol]]
  coef(lm(f, data = d, weights = w))
})
colnames(coefs) <- gsub("wei_|_c2c", "", weight_cols)
round(coefs, 4)
```

All coefficients can vary because weight methods change which old-period subjects contribute to this occupation group.

### Compare pruning strategies only after comparing full weights

> **Note**: Pruning discards probability information and should be used only after analysis with full weights. 
Prefer `prune_c2c(method = "nonzero")` to remove impossible candidates while preserving the probability distribution. 
More aggressive pruning (`highest1`) is appropriate only for descriptive tables or when you need exactly one category per observation.

```{r sensitivity-pruning}
# Compare regression coefficients under different pruning strategies
prune_methods <- c("nonzero", "highest", "highest1")

prune_coefs <- sapply(prune_methods, function(pm) {
  old_pruned <- result_all$old %>%
    prune_c2c(method = pm) %>%
    filter(g_new_c2c == target_group)
  
  d <- bind_rows(old_pruned, new_subset)
  d$w <- d$multiplier * d$wei_freq_c2c
  coef(lm(f, data = d, weights = w))
})
round(prune_coefs, 4)
```

### Compare ensemble compositions when no single method dominates

`cross_c2c()` creates a weighted average of multiple weight columns. Vary the mix:

```{r sensitivity-ensemble}
configs <- list(
  equal      = c(1, 1) / 2,
  freq_heavy = c(3, 1) / 4,
  ml_heavy   = c(1, 3) / 4
)

ens_coefs <- sapply(names(configs), function(nm) {
  old_ens <- result_all$old %>%
    cross_c2c(c("wei_freq_c2c", "wei_knn_c2c"), configs[[nm]]) %>%
    filter(g_new_c2c == target_group)
  
  new_ens <- new_subset %>% mutate(wei_cross_c2c = 1)
  d <- bind_rows(old_ens, new_ens)
  d$w <- d$multiplier * d$wei_cross_c2c
  coef(lm(f, data = d, weights = w))
})
round(ens_coefs, 4)
```

When regression coefficients are stable across weight methods, pruning strategies, and ensemble compositions, report with confidence. When they diverge, the mapping introduces uncertainty - report the range or investigate the source.

## Step 3: Validate whether ML actually improves on simpler baselines

The `ml` argument in `cat2cat()` adds ML-based probability weights, but ML is not guaranteed to improve over simpler baselines. `cat2cat_ml_run()` provides per-group holdout (single train/test split) diagnostics to answer this question *before* committing to a method.

### What `cat2cat_ml_run()` is doing

For each mapping group (set of candidate categories linked by the transition table) `cat2cat_ml_run()`:

1. Collects all observations from `ml$data` whose category belongs to the group.
2. Randomly splits them into training (`1 - test_prop`) and test (`test_prop`) sets.
3. Computes two baselines on the test set:
   - **naive** - accuracy of a random guess ($1 / k$ where $k$ is the number of candidate categories).
   - **freq** - accuracy of always predicting the most frequent training-set category.
4. Trains each specified ML method on the training set and records test-set model performance.

Groups with fewer than 5 observations or only one candidate category are skipped.
Also note that `cat2cat_ml_run()` does not use `on_fail`; it is a diagnostic
tool and reports skipped groups instead of applying row-level fallback weights.

### Minimal validation workflow

```{r cv-basic}
cv_knn <- cat2cat_ml_run(
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = bind_rows(occup_2010, occup_2012),
    cat_var = "code",
    method = "knn",
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)
print(cv_knn)
```

The `print()` summary reports:

- **ACCURACY** - average held-out classification accuracy across non-skipped groups. `naive (1/k)` is the random-guess baseline, `freq` is the majority-class baseline, and each ML line reports top-class accuracy for that method.
- **BRIER SCORE** - average full-vector probability error across non-skipped groups. Lower is better. This matters because `cat2cat` ultimately uses probability weights, not just hard classifications.
- **MEAN P(TRUE CLASS)** - average probability assigned to the true category. Higher is better. This is often the most directly relevant metric for `cat2cat`, because it measures the quality of the probability weights themselves.
- **ACCURACY: ML vs BASELINES** - the share of groups in which the ML method beats `naive` or beats `freq` on accuracy. This is a win-rate summary, not an average accuracy gap.
- **SKIPPED GROUPS** - the percentage of mapping groups for which that ML method has no reported result because the group had only one candidate category, fewer than 5 observations, or the model could not be fit for that group.

So for output like:

- `knn > naive: 87.7%`
- `knn > freq: 18.0%`
- `knn: accuracy = 0.5108` vs `freq (most common): 0.5366`

the right reading is: kNN clearly beats the naive baseline, but it does **not** beat the frequency baseline on top-class accuracy overall. In that case, `wei_freq_c2c` remains the default choice if your only goal is classification accuracy.

At the same time, if kNN has a slightly lower Brier score and a higher mean P(true class) than `freq`, then it may still be producing better-calibrated probability weights even though its top prediction is less often correct. That distinction matters in `cat2cat`, because the mapped weights are probabilities distributed across candidate categories rather than single-class assignments.

### Compare multiple ML methods in one run

```{r cv-multiple}
cv_all <- cat2cat_ml_run(
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = bind_rows(occup_2010, occup_2012),
    cat_var = "code",
    method = c("knn", "lda", "rf", "nb"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 50)
  )
)
print(cv_all)
```

Interpretation tip for mixed outputs:

- It is possible for a method to have 0 failed rows in `cat2cat()` but a non-zero
  skipped-group rate in `cat2cat_ml_run()`.
- This is not a contradiction: the first is row-level weight construction, the
  second is group-level holdout evaluation coverage.

### Inspect per-group diagnostics when methods disagree

The returned object is a named list. Each element corresponds to one mapping group:

```{r cv-inspect}
# Pick a group with multiple candidates
group_names <- names(cv_all)
example_group <- group_names[
  which(vapply(cv_all, function(g) !is.na(g$freq) && g$naive < 1, logical(1)))[1]
]
cv_all[[example_group]]
```

Each group entry contains the group-level diagnostics behind the printed summary:

- `$naive` - $1/k$ random-guess accuracy for that group.
- `$freq` - majority-class accuracy for that group.
- `$acc` - named numeric vector with ML accuracy by method.
- `$naive_brier` and `$freq_brier` - baseline Brier scores.
- `$brier` - named numeric vector with ML Brier scores by method.
- `$naive_mean_prob` and `$freq_mean_prob` - baseline mean P(true class).
- `$mean_prob` - named numeric vector with ML mean P(true class) by method.

### Decision rules for interpreting the output

**Understanding model performance in context**: This is **multi-class classification** - each mapping group can have 3-10+ candidate categories. 
A naive random guess yields only ~18% accuracy (1/k where k is the number of candidates). 
Achieving 50%+ is substantial improvement over random - do not compare these numbers to binary classification benchmarks where 80%+ is typical. 
The key question is whether ML beats the *frequency* baseline, not whether it reaches some absolute threshold.

| Scenario | Recommendation |
|----------|---------------|
| ML model performance >> freq across most groups | ML weights add genuine signal; use them |
| ML model performance $\approx$ freq | ML is no better than frequency; prefer `wei_freq_c2c` (simpler, faster) |
| ML model performance < freq for many groups | ML is adding noise; do **not** use ML weights |
| High skipped-group rate (>20%) | Features may have too many missing values, groups are too small, or method fitting is unstable |

Because the train/test split is random, results vary between runs. For more stable estimates, pool more data into `ml$data` (e.g. multiple survey waves) or run `cat2cat_ml_run()` several times and average the summaries.

> **Caveat**: high `cat2cat_ml_run()` model performance means the model discriminates well *within* mapping groups. 
It does not validate the mapping table itself. A perfect model with a wrong transition table will still produce wrong results.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapping.R
\name{cat_apply_freq}
\alias{cat_apply_freq}
\title{Applying frequencies to the object returned by the `get_mappings` function}
\usage{
cat_apply_freq(to_x, freqs)
}
\arguments{
\item{to_x}{`list` object returned by `get_mappings`.}

\item{freqs}{`data.frame` object like the one returned by
the `get_freqs` function.}
}
\value{
a `list` with a structure like `to_x` object but with probabilities
for each category.
}
\description{
applying frequencies to the object returned by
the `get_mappings` function.
We will get a symmetric object to the one returned by
the `get_mappings` function, nevertheless categories are replaced
with frequencies.
Frequencies for each category/key are sum to 1, so could be interpreted
as probabilities.
}
\note{
`freqs` arg first column (keys) and the to_x arg values have to be of
the same type.
The uniform distribution (outcomes are equally likely) is assumed
for no match for all possible categories.
}
\examples{
data("trans", package = "cat2cat")
data("occup", package = "cat2cat")

mappings <- get_mappings(trans)

mappings$to_old[1:4]
mappings$to_new[1:4]

mapp_p <- cat_apply_freq(
  mappings$to_old,
  get_freqs(
    occup$code[occup$year == "2008"],
    occup$multiplier[occup$year == "2008"]
  )
)
head(data.frame(I(mappings$to_old), I(mapp_p)))
mapp_p <- cat_apply_freq(
  mappings$to_new,
  get_freqs(
    occup$code[occup$year == "2010"],
    occup$multiplier[occup$year == "2010"]
  )
)
head(data.frame(I(mappings$to_new), I(mapp_p)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{trans}
\alias{trans}
\title{trans dataset containing mappings (transitions) between
old (2008) and new (2010) occupational codes.
This table could be used to map encodings in both directions.}
\format{
A data frame with 2693 observations and 2 variables.
\describe{
\item{old}{	character an old encoding of a certain occupation}
\item{new}{	character a new encoding of a certain occupation}
}
}
\usage{
trans
}
\description{
trans dataset containing mappings (transitions) between
old (2008) and new (2010) occupational codes.
This table could be used to map encodings in both directions.
}
\details{
mapping (transition) table for occupations where first column
contains old encodings and second one a new encoding
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat_ml.R
\name{cat2cat_ml}
\alias{cat2cat_ml}
\title{The internal function used in the cat2cat one}
\usage{
cat2cat_ml(ml, mapp, target_data, cat_var_target)
}
\arguments{
\item{ml}{`list` the same `ml` argument as provided to `cat2cat` function.}

\item{mapp}{`list` a mapping table}

\item{target_data}{`data.frame`}

\item{cat_var_target}{`character(1)` name of the categorical variable
in the target period.}
}
\description{
apply the ml models to the cat2cat data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{verticals2}
\alias{verticals2}
\title{verticals2 dataset}
\format{
A data frame with 202 observations and 4 variables.
\describe{
\item{ean}{ product ean}
\item{vertical}{	character an certain sales vertical}
\item{sales}{	numeric a size of sale}
\item{v_date}{	character Date}
}
}
\usage{
verticals2
}
\description{
verticals2 dataset
}
\details{
random data - single products sales across e-commerce verticals
}
\examples{
set.seed(1234)
vert_old <- data.frame(
  ean = 90000001:90000020,
  vertical = sample(c(
    "Electronics", "Kids1", "Kids2", "Automotive", "Books",
    "Clothes", "Home", "Fashion", "Health", "Sport"
  ), 20, replace = TRUE),
  sales = rnorm(20, 100, 10),
  v_date = rep("2020-04-01", 20), stringsAsFactors = FALSE
)

vert_old2 <- data.frame(
  ean = 90000021:90000100,
  vertical = sample(c(
    "Electronics", "Kids1", "Kids2", "Automotive", "Books",
    "Clothes", "Home", "Fashion", "Health", "Sport"
  ), 80, replace = TRUE),
  sales = rnorm(80, 100, 10),
  v_date = rep("2020-04-01", 80), stringsAsFactors = FALSE
)

vert_new <- vert_old2
vert_new$sales <- rnorm(nrow(vert_new), 80, 10)
vert_new$v_date <- "2020-05-01"
vert_new$vertical[vert_new$vertical \%in\% c("Kids1", "Kids2")] <- "Kids"
vert_new$vertical[vert_new$vertical \%in\% c("Automotive")] <-
  sample(
    c("Automotive1", "Automotive2"),
    sum(vert_new$vertical \%in\% c("Automotive")),
    replace = TRUE
  )
vert_new$vertical[vert_new$vertical \%in\% c("Home")] <-
  sample(
    c("Home", "Supermarket"),
    sum(vert_new$vertical \%in\% c("Home")),
    replace = TRUE
  )

vert_new2 <- data.frame(
  ean = 90000101:90000120,
  vertical = sample(
    c(
      "Electronics", "Supermarket", "Kids", "Automotive1",
      "Automotive2", "Books", "Clothes", "Home",
      "Fashion", "Health", "Sport"
    ), 20,
    replace = TRUE
  ),
  sales = rnorm(20, 100, 10),
  v_date = rep("2020-05-01", 20), stringsAsFactors = FALSE
)

verticals2 <- rbind(
  rbind(vert_old, vert_old2),
  rbind(vert_new, vert_new2)
)
verticals2$vertical <- as.character(verticals2$vertical)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_cat2cat.R
\name{plot_c2c}
\alias{plot_c2c}
\title{Summary plots for cat2cat results}
\usage{
plot_c2c(data, weis = "wei_freq_c2c", type = c("both", "hist", "bar"))
}
\arguments{
\item{data}{`data.frame` - one of the data.frames returned by
the `cat2cat` function.}

\item{weis}{`character(1)` - name of a certain wei_*_c2c column,
added by cat2cat function. Default `wei_freq_c2c`.}

\item{type}{`character(1)` - one of 3 types `"both"`, `"hist"`, `"bar"`.}
}
\value{
base plot graphics
}
\description{
This function help to understand properties of cat2cat results.
It is recommended to run it before further processing, like next iterations.
}
\note{
It will work only for data.frame produced by cat2cat function.
}
\examples{
data("occup_small", package = "cat2cat")
occup_old <- occup_small[occup_small$year == 2008, ]
occup_new <- occup_small[occup_small$year == 2010, ]

occup_2 <- cat2cat(
  data = list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward")
)

plot_c2c(occup_2$old, type = c("both"))
plot_c2c(occup_2$old, type = c("hist"))
plot_c2c(occup_2$old, type = c("bar"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat.R
\name{resolve_frequencies}
\alias{resolve_frequencies}
\title{Resolve the frequencies}
\usage{
resolve_frequencies(cat_base_year, cat_var_base, freqs_df, multiplier_var)
}
\arguments{
\item{cat_base_year}{`data.frame` a base period dataset.}

\item{cat_var_base}{`character(1)` name of the base period
categorical variable.}

\item{freqs_df}{`data.frame` with 2 columns, the first one with category
and second with counts.}

\item{multiplier_var}{`character(1)` name of the multiplier variable.}
}
\description{
Resolve the frequencies
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat_utils.R
\name{cross_c2c}
\alias{cross_c2c}
\title{Make a combination of weights from different methods}
\usage{
cross_c2c(
  df,
  cols = colnames(df)[grepl("^wei_.*_c2c$", colnames(df))],
  weis = rep(1/length(cols), length(cols)),
  na.rm = TRUE
)
}
\arguments{
\item{df}{`data.frame` like result of the `cat2cat` function
for a specific period.}

\item{cols}{`character` vector default all columns under
the regex "wei_.*_c2c".}

\item{weis}{`numeric` vector weighs for columns in the `cols` argument.
By default a vector of the same length as `cols` argument and with equally
spaced probability (summing to 1).}

\item{na.rm}{`logical(1)` if `NA` values should be omitted, default TRUE.}
}
\value{
`data.frame` with the additional column `wei_cross_c2c`.
}
\description{
adding the additional column which is a mix of weights columns
by each row.
Ensemble of a few methods usually produces more accurate solutions
than a single model would.
}
\examples{
\dontrun{
data("occup_small", package = "cat2cat")
data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup_small[occup_small$year == 2008, ]
occup_new <- occup_small[occup_small$year == 2010, ]

# mix of methods - forward direction, try out backward too
occup_mix <- cat2cat(
  data = list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_new,
    cat_var = "code",
    method = c("knn"),
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10, ntree = 20)
  )
)
# correlation between ml model
occup_mix_old <- occup_mix$old
cor(
  occup_mix_old[occup_mix_old$rep_c2c != 1, c("wei_knn_c2c", "wei_freq_c2c")]
)
# cross all methods and subset one highest probability category for each obs
occup_old_highest1_mix <- prune_c2c(cross_c2c(occup_mix$old),
  column = "wei_cross_c2c", method = "highest1"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{verticals}
\alias{verticals}
\title{verticals dataset}
\format{
A data frame with 21 observations and 4 variables.
\describe{
\item{vertical}{	character an certain sales vertical}
\item{sales}{	numeric a size of sale}
\item{counts}{	integer counts size}
\item{v_date}{	character Date}
}
}
\usage{
verticals
}
\description{
verticals dataset
}
\details{
random data - aggregate sales across e-commerce verticals
}
\examples{
set.seed(1234)
agg_old <- data.frame(
  vertical = c(
    "Electronics", "Kids1", "Kids2", "Automotive", "Books",
    "Clothes", "Home", "Fashion", "Health", "Sport"
  ),
  sales = rnorm(10, 100, 10),
  counts = rgeom(10, 0.0001),
  v_date = rep("2020-04-01", 10), stringsAsFactors = FALSE
)

agg_new <- data.frame(
  vertical = c(
    "Electronics", "Supermarket", "Kids", "Automotive1",
    "Automotive2", "Books", "Clothes", "Home", "Fashion", "Health", "Sport"
  ),
  sales = rnorm(11, 100, 10),
  counts = rgeom(11, 0.0001),
  v_date = rep("2020-05-01", 11), stringsAsFactors = FALSE
)
verticals <- rbind(agg_old, agg_new)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat.R
\name{validate_data}
\alias{validate_data}
\title{Validate cat2cat data}
\usage{
validate_data(data)
}
\description{
Validate cat2cat data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat.R
\name{validate_cover_cats}
\alias{validate_cover_cats}
\title{Validate if the trans table contains all proper mappings}
\usage{
validate_cover_cats(u_cats_target, mapp)
}
\arguments{
\item{u_cats_target}{vector of unique target period categories}

\item{mapp}{transition (mapping) table process with `get_mappings`,
the "to_base" direction is taken.}
}
\description{
Validate if the trans table contains all proper mappings
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat_utils.R
\name{prune_c2c}
\alias{prune_c2c}
\title{Pruning which could be useful after the mapping process}
\usage{
prune_c2c(
  df,
  index = "index_c2c",
  column = "wei_freq_c2c",
  method = "nonzero",
  percent = 50
)
}
\arguments{
\item{df}{`data.frame` like result of the `cat2cat` function
for a specific period.}

\item{index}{`character(1)` a column name with the `cat2cat` identifier.
Should not be updated in most cases. Default `index_c2c`.}

\item{column}{`character(1)` a column name with weights,
default `wei_freq_c2c`.}

\item{method}{`character(1)` one of four available methods:
"nonzero" (default), "highest", "highest1" or "morethan".}

\item{percent}{`integer(1)` from 0 to 99}
}
\value{
`data.frame` with the same structure and possibly reduced
number of rows
}
\description{
user could specify one of four methods to prune replications
created in the cat2cat procedure.
}
\details{
method - specify a method to reduce number of replications
\describe{
 \item{"nonzero"}{ remove nonzero probabilities}
 \item{"highest"}{
 leave only highest probabilities for each subject- accepting ties
 }
 \item{"highest1"}{
 leave only highest probabilities for each subject -
 not accepting ties so always one is returned
 }
 \item{"morethan"}{
 leave rows where a probability is higher than value specify
 by percent argument
 }
}
}
\examples{
\dontrun{
data("occup_small", package = "cat2cat")
data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup_small[occup_small$year == 2008, ]
occup_new <- occup_small[occup_small$year == 2010, ]

occup_ml <- cat2cat(
  data = list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_new,
    cat_var = "code",
    method = "knn",
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)

prune_c2c(occup_ml$old, method = "nonzero")
prune_c2c(occup_ml$old, method = "highest")
prune_c2c(occup_ml$old, method = "highest1")
prune_c2c(occup_ml$old, method = "morethan", percent = 90)

prune_c2c(occup_ml$old, column = "wei_knn_c2c", method = "nonzero")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapping.R
\name{get_mappings}
\alias{get_mappings}
\title{Transforming a mapping (transition) table to two associative lists}
\usage{
get_mappings(x = data.frame())
}
\arguments{
\item{x}{`data.frame` or `matrix` - mapping (transition) table with 2 columns
where first column is assumed to be the older encoding.}
}
\value{
a list with 2 named lists `to_old` and `to_new`.
}
\description{
to rearrange the one classification encoding into another,
 an associative list that maps keys to values is used.
More precisely, an association list is used which is a linked list in which
each list element consists of a key and value or values.
An association list where unique categories codes are keys and matching
categories from next or previous time point are values.
A mapping (transition) table is used to build such associative lists.
}
\examples{
data("trans", package = "cat2cat")

mappings <- get_mappings(trans)
mappings$to_old[1:4]
mappings$to_new[1:4]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lms.R
\name{summary_c2c}
\alias{summary_c2c}
\title{Adjusted summary for regressions on replicated datasets}
\usage{
summary_c2c(x, df_old, df_new = x$df.residual)
}
\arguments{
\item{x}{lm or glm object}

\item{df_old}{integer number of d.f in original dataset. For bigger datasets
`nrow` should be sufficient.}

\item{df_new}{integer number of d.f in dataset with replicated rows,
Default: x$df.residual}
}
\value{
data.frame with additional columns over a regular summary output,
like correct and statistics adjusted by it.
}
\description{
adjusting lm/glm object results according to original number of
degree of freedom.
The standard errors, test statistics and p values have to be adjusted because of
replicated observations.
}
\details{
The replication step in \code{cat2cat} inflates the nominal sample
size: the model sees \code{n_rep} rows but only \code{n_orig} are independent
observations. Naive OLS therefore under-estimates standard errors.

The correction factor is \code{sqrt(df_new / df_old)}, where \code{df_new} is
the residual d.f. from the replicated model and \code{df_old} the residual
d.f. from the original dataset. Standard errors are multiplied by this
factor, test statistics divided by it, and p-values recomputed from the
appropriate reference distribution: \code{t(df_old)} for t-based summaries
and standard normal for z-based summaries.

This is a pragmatic d.f. adjustment. It works well when per-subject
weights sum to one and no extreme weights dominate.

Note: Ordinary R-squared can be interpreted in neutral replication cases
where the response and covariates do not vary across replicated copies and
per-observation weights sum to the original observation weight. Adjusted
R-squared, AIC, and BIC depend on sample-size and degrees-of-freedom
conventions, so the values from the replicated model should not be reported
without recomputing them on the intended original-observation scale. If the
harmonised category itself enters the model, all fit statistics are
conditional on the chosen harmonisation weights.
}
\examples{
data("occup_small", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup_small[occup_small$year == 2008, ]
occup_new <- occup_small[occup_small$year == 2010, ]

occup_2 <- cat2cat(
  data = list(
    old = occup_old,
    new = occup_new,
    cat_var = "code",
    time_var = "year"
  ),
  mappings = list(trans = trans, direction = "backward"),
  ml = list(
    data = occup_new,
    cat_var = "code",
    method = "knn",
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10)
  )
)

# Regression
# we have to adjust size of std as we artificialy enlarge degrees of freedom
lms <- lm(
  formula = I(log(salary)) ~ age + sex + factor(edu) + parttime + exp,
  data = occup_2$old,
  weights = multiplier * wei_freq_c2c
)

summary_c2c(lms, df_old = nrow(occup_old))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat_agg.R
\name{cat2cat_agg}
\alias{cat2cat_agg}
\title{Manual mapping for an aggregated panel dataset}
\usage{
cat2cat_agg(
  data = list(old = NULL, new = NULL, cat_var_old = NULL, cat_var_new = NULL, time_var =
    NULL, freq_var = NULL),
  ...
)
}
\arguments{
\item{data}{Named list with 5-6 fields describing the datasets:
\describe{
  \item{\code{old}}{\code{data.frame} - older time period (one row per category).}
  \item{\code{new}}{\code{data.frame} - newer time period (one row per category).}
  \item{\code{cat_var}}{\code{character(1)} - (deprecated) category variable name
    if identical in both periods. Use \code{cat_var_old}/\code{cat_var_new} instead.}
  \item{\code{cat_var_old}}{\code{character(1)} - category variable name in old period.}
  \item{\code{cat_var_new}}{\code{character(1)} - category variable name in new period.}
  \item{\code{time_var}}{\code{character(1)} - time variable name.}
  \item{\code{freq_var}}{\code{character(1)} - frequency/count variable used to
    compute proportions when splitting one category into many.}
}}

\item{...}{Mapping equations specifying how categories relate across periods.
See **Equation syntax** in Details.}
}
\value{
A named list with two elements:
\describe{
  \item{\code{$old}}{\code{data.frame} - old period with \code{prop_c2c} column added.
    Categories may be replicated if split by backward equations.}
  \item{\code{$new}}{\code{data.frame} - new period with \code{prop_c2c} column added.
    Categories may be replicated if split by forward equations.}
}
The \code{prop_c2c} column contains the proportion (0-1) to apply when
aggregating. For rows not affected by any equation, \code{prop_c2c = 1}.
For split categories, proportions sum to 1 within the original category.
}
\description{
Applies user-defined mapping equations to redistribute aggregated counts
(and other numeric columns) when categorical encodings change between
time periods. Unlike \code{\link{cat2cat}}, which works on micro-data,
this function operates on pre-aggregated data where each row represents
a category with associated counts/totals.
}
\details{
\strong{Equation syntax}

Each equation has the form:
\preformatted{OLD_SIDE  DIRECTION  NEW_SIDE}

where:
\itemize{
  \item \strong{OLD_SIDE} - one or more old-period category names
    (use \code{c(A, B)} for multiple)
  \item \strong{DIRECTION} - one of:
    \itemize{
      \item \code{\%>\%} or \code{>} - \strong{forward}: replicates the NEW period,
        renaming/splitting new categories to match old encoding
      \item \code{\%<\%} or \code{<} - \strong{backward}: replicates the OLD period,
        renaming/splitting old categories to match new encoding
    }
  \item \strong{NEW_SIDE} - one or more new-period category names
}

\strong{How proportions are calculated}

When one category maps to multiple:
\itemize{
  \item \strong{Backward} (\code{\%<\%}): proportions come from \code{freq_var}
    in the \strong{new} period (the target encoding)
  \item \strong{Forward} (\code{\%>\%}): proportions come from \code{freq_var}
    in the \strong{old} period (the target encoding)
}

\strong{Examples of valid equations}

\describe{
  \item{\code{Automotive \%<\% c(Automotive1, Automotive2)}}{
    Backward: the old "Automotive" row is split into two rows
    ("Automotive1", "Automotive2") with proportions from new-period counts.}
  \item{\code{c(Kids1, Kids2) \%>\% c(Kids)}}{
    Forward: the new "Kids" row is split into two rows
    ("Kids1", "Kids2") with proportions from old-period counts.}
  \item{\code{Home \%>\% c(Home, Supermarket)}}{
    Forward: the new "Home" and "Supermarket" rows are each renamed to
    "Home" (1-to-many from new perspective; after aggregation they merge).}
}

\strong{Typical workflow}

1. Call \code{cat2cat_agg()} with all mapping equations.
2. Bind \code{$old} and \code{$new} together.
3. Group by time and the (now unified) category variable.
4. Summarise numeric columns as \code{sum(value * prop_c2c)}.
}
\note{
\itemize{
  \item All equations must be valid - unknown category names cause an error.
  \item Each equation must have exactly one category on one side
    (the "one" in one-to-many). Many-to-many within a single equation
    is not allowed.
  \item Each category in \code{old} and \code{new} must appear exactly once
    (no duplicates allowed before mapping).
}
}
\examples{
data("verticals", package = "cat2cat")
agg_old <- verticals[verticals$v_date == "2020-04-01", ]
agg_new <- verticals[verticals$v_date == "2020-05-01", ]

# cat2cat_agg - can map in both directions at once
# although usually we want to have the old or the new representation

agg <- cat2cat_agg(
  data = list(
    old = agg_old,
    new = agg_new,
    cat_var_old = "vertical",
    cat_var_new = "vertical",
    time_var = "v_date",
    freq_var = "counts"
  ),
  Automotive \%<\% c(Automotive1, Automotive2),
  c(Kids1, Kids2) \%>\% c(Kids),
  Home \%>\% c(Home, Supermarket)
)

## possible processing
library("dplyr")
agg \%>\%
  bind_rows() \%>\%
  group_by(v_date, vertical) \%>\%
  summarise(
    sales = sum(sales * prop_c2c),
    counts = sum(counts * prop_c2c),
    v_date = first(v_date)
  )
}
\seealso{
\code{vignette("cat2cat_advanced")} for a complete workflow example.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{occup_small}
\alias{occup_small}
\title{Occupational dataset - small one}
\format{
A data frame with around 8000 observations and 12 variables.
\describe{
\item{id}{	integer	id}
\item{age}{ numeric age of a subject}
\item{sex}{ numeric sex of a subject}
\item{edu}{
  integer edu level of education of a subject where lower means
  higher - 1 for at least master degree
}
\item{exp}{ numeric exp number of experience years for a subject}
\item{district}{ integer district}
\item{parttime}{
  numeric contract type regards time where 1 mean full-time
  (work a whole week)
}
\item{salary}{ numeric salary per year}
\item{code}{ character code - occupational code}
\item{multiplier}{
  numeric multiplier for the subject to reproduce a population -
  how many of such subjects in population
}
\item{year}{integer year}
\item{code4}{ character code - occupational code - first 4 digits}
}
}
\usage{
occup_small
}
\description{
Occupational dataset - small one
}
\details{
occup dataset is an example of unbalance panel dataset.
This is a simulated data although there are applied a real world
characteristics from national statistical office survey.
The original survey is anonymous and take place every two years.
It is presenting a characteristics from randomly selected company and
then using k step procedure employees are chosen.

occupational dataset
}
\examples{
set.seed(1234)
data("occup", package = "cat2cat")
occup_small <- occup[sort(sample(nrow(occup), 8000)), ]
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{occup_panel}
\alias{occup_panel}
\title{Occupational panel dataset with BAEL-style quarterly rotation}
\format{
A data frame with approximately 3900 observations and 15 variables.
\describe{
\item{id}{integer original row identifier from \code{occup}}
\item{panel_id}{integer consistent person identifier across quarters}
\item{cohort}{character entry quarter (e.g., "2009Q1")}
\item{age}{numeric age of the worker}
\item{sex}{numeric sex of the worker}
\item{edu}{integer education level (1 = highest)}
\item{exp}{numeric years of experience}
\item{district}{integer district code}
\item{parttime}{numeric contract type (1 = full-time)}
\item{salary}{numeric annual salary}
\item{code}{character occupational code (4-digit pre-2010, 6-digit post-2010)}
\item{multiplier}{numeric survey weight}
\item{quarter}{character survey quarter (e.g., "2009Q1", "2010Q2")}
\item{code4}{character first 4 digits of occupational code}
\item{year}{integer year extracted from quarter}
\item{quarter_num}{integer quarter number (1-4)}
}
}
\usage{
occup_panel
}
\description{
Occupational panel dataset with BAEL-style quarterly rotation
}
\details{
A simulated rotational panel derived from \code{occup}, inspired by
the Polish BAEL (Badanie Aktywnosci Ekonomicznej Ludnosci - Labour Force
Survey). Unlike the main \code{occup} dataset (repeated cross-sections), this
dataset includes workers observed for 4 consecutive quarters, enabling
demonstration of the \code{id_var} feature in \code{cat2cat()}.

Panel design:
\itemize{
  \item 8 quarters: 2009Q1 through 2010Q4
  \item Encoding change between 2009Q4 and 2010Q1
  \item Each cohort enters quarterly and stays for 4 consecutive quarters
  \item ~150 new subjects enter each quarter (1/4 rotation)
  \item ~450 subjects observed across the encoding change
}

For subjects across quarters:
\itemize{
  \item \code{panel_id} is consistent across quarters
  \item Occupation code is preserved (or mapped via \code{trans} at encoding change)
  \item Age and experience increase annually (every 4 quarters)
  \item Salary varies slightly between quarters (-1\% to +2\%)
}
}
\examples{
data("occup_panel", package = "cat2cat")

# Check panel structure
table(occup_panel$quarter)
table(table(occup_panel$panel_id))  # appearances per subject (target: 4)

# Subjects observed across the encoding change (2009Q4 -> 2010Q1)
panel_2009Q4 <- occup_panel[occup_panel$quarter == "2009Q4", ]
panel_2010Q1 <- occup_panel[occup_panel$quarter == "2010Q1", ]
length(intersect(panel_2009Q4$panel_id, panel_2010Q1$panel_id))
}
\seealso{
\code{\link{occup}} for the full repeated cross-section dataset,
  \code{\link{trans}} for the transition table
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat_utils.R
\name{dummy_c2c}
\alias{dummy_c2c}
\title{Add default cat2cat columns to a `data.frame`}
\usage{
dummy_c2c(df, cat_var, ml = NULL)
}
\arguments{
\item{df}{`data.frame`.}

\item{cat_var}{`character(1)` a categorical variable name.}

\item{ml}{`character` vector of ml models applied,
any of `c("knn", "rf", "lda", "nb")`.}
}
\value{
the provided `data.frame` with additional `cat2cat` like columns.
}
\description{
a utils function to add default cat2cat columns
to a `data.frame`.
It will be useful e.g. for a boarder periods which will not have additional
 `cat2cat` columns.
}
\examples{
\dontrun{
dummy_c2c(airquality, "Month")

data("occup_small", package = "cat2cat")
occup_old <- occup_small[occup_small$year == 2008, ]
dummy_c2c(occup_old, "code")
dummy_c2c(occup_old, "code", "knn")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapping.R
\name{get_freqs}
\alias{get_freqs}
\title{Getting frequencies from a vector with an optional multiplier}
\usage{
get_freqs(x, multiplier = NULL)
}
\arguments{
\item{x}{`vector` categorical variable to summarize.}

\item{multiplier}{`numeric` vector how many times to repeat certain value,
additional weights.}
}
\value{
`data.frame` with two columns `input` `Freq`
}
\description{
getting frequencies for a vector with an optional multiplier.
}
\note{
without multiplier variable it is a basic `table` function wrapped with
 the `as.data.frame` function.
The `table` function is used with the `useNA = "ifany"` argument.
}
\examples{
data("occup", package = "cat2cat")

head(get_freqs(occup$code[occup$year == "2008"]))
head(get_freqs(occup$code[occup$year == "2010"]))

head(
  get_freqs(
    occup$code[occup$year == "2008"],
    occup$multiplier[occup$year == "2008"]
  )
)
head(
  get_freqs(
    occup$code[occup$year == "2010"],
    occup$multiplier[occup$year == "2010"]
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat.R
\name{cat2cat}
\alias{cat2cat}
\title{Automatic mapping in a panel dataset}
\usage{
cat2cat(
  data = list(old = NULL, new = NULL, time_var = NULL, cat_var = NULL, cat_var_old =
    NULL, cat_var_new = NULL, id_var = NULL, multiplier_var = NULL),
  mappings = list(trans = NULL, direction = NULL, freqs_df = NULL),
  ml = list(data = NULL, cat_var = NULL, method = NULL, features = NULL, args = NULL)
)
}
\arguments{
\item{data}{`named list` with fields `old`, `new`,
`cat_var` (shorthand for `cat_var_old = cat_var_new = cat_var`),
`time_var` and optional `id_var`, `multiplier_var`.}

\item{mappings}{`named list` with 3 fields `trans`, `direction` and
optional `freqs_df`.}

\item{ml}{`named list` (optional) with fields
`data`, `cat_var`, `method`, `features` and optional `args`,
`on_fail`, `fail_warn`.}
}
\value{
`named list` with 2 fields old and new - 2 data.frames.
There will be added additional columns like
index_c2c, g_new_c2c, wei_freq_c2c, rep_c2c, wei_(ml method name)_c2c.
Additional columns will be informative only for a one data.frame
as we always make the changes to one direction.
The new columns are added instead of the additional metadata as
we are working with new datasets
where observations could be replicated.
For the transparency the probability and number of replications are part
 of each observation in the `data.frame`.
}
\description{
Unifies an inconsistently coded categorical variable in a panel dataset
according to a mapping (transition) table.
Apply iteratively for each pair of neighboring periods.
Use \code{\link{prune_c2c}} to limit growing replications across steps.
}
\details{
data args
\describe{
 \item{"old"}{ data.frame older time point in a panel}
 \item{"new"}{ data.frame more recent time point in a panel}
 \item{"time_var"}{ character(1) name of the time variable.}
 \item{"cat_var"}{ character(1) name of the categorical variable.
  Shorthand: sets both \code{cat_var_old} and \code{cat_var_new} to the same value.
 }
 \item{"cat_var_old"}{
 Optional character(1) name of the categorical variable
 in the older time point. Default `cat_var`.
 }
 \item{"cat_var_new"}{
 Optional character(1) name of the categorical variable
 in the newer time point. Default `cat_var`.
 }
 \item{"id_var"}{Optional character(1) name of the unique identifier variable.
  When specified, subjects observed in both periods are mapped 1-to-1
  directly (no replication, \code{wei_freq_c2c = 1, rep_c2c = 1}).
  Only subjects absent from the base period enter the replication path.
  This assumes a subject's true category does not change between adjacent
  waves. See \code{vignette("cat2cat_advanced")} for details on when
  this assumption is and is not satisfied.
 }
 \item{"multiplier_var"}{
 Optional character(1) name of the multiplier variable -
 number of replication needed to reproduce the population
 }
 \item{"freqs_df"}{
 Deprecated - use \code{mappings$freqs_df} instead.
 }
}
mappings args
\describe{
 \item{"trans"}{ data.frame with 2 columns - mapping (transition) table -
  all categories for cat_var in old and new datasets have to be included.
  First column contains an old encoding and second a new one.
  The mapping (transition) table should include a candidate for each category
  in the period being harmonised.
}
 \item{"direction"}{ character(1) direction - "backward" or "forward"}
 \item{"freqs_df"}{
 Optional - data.frame with 2 columns where first one
 is category name (base period) and second counts.
 If It is not provided then is assessed automatically.
 Artificial counts for each variable level in the base period.
 It is optional nevertheless will be often needed, as gives more control.
 It will be used to assess the probabilities.
 The multiplier variable is omitted so sb has to apply it in this table.
 }
}
Optional ml args
\describe{
 \item{"data"}{ data.frame - dataset with features and the `cat_var`.}
 \item{"cat_var"}{ character(1) - the dependent variable name.}
 \item{"method"}{
 character vector - one or a few from
 "knn", "rf", "lda" and "nb" methods - "knn" k-NearestNeighbors,
 "lda" Linear Discriminant Analysis, "rf" Random Forest,
 "nb" Naive Bayes
 }
 \item{"features"}{
 character vector of features names where all
 have to be numeric, logical or factor.
 Factor features are automatically one-hot encoded using the union of
 levels observed in \code{ml$data} and the target period.
 }
 \item{"args"}{ optional - list parameters: knn: k ; rf: ntree  }
 \item{"on_fail"}{ optional character(1) controlling failed ML weights:
 \code{"freq"} (default) uses \code{wei_freq_c2c}, \code{"naive"} uses
 \code{wei_naive_c2c}, \code{"na"} leaves failed weights as \code{NA},
 and \code{"error"} stops when failed weights are detected.}
 \item{"fail_warn"}{ optional logical(1), default \code{TRUE}; warn when
 failed ML weights are replaced or retained as \code{NA}.}
}

Without the \code{ml} argument, only frequency-based weights are computed.
If an ML model fails, \code{ml$on_fail} controls whether frequency weights,
naive weights, \code{NA}, or an error are used.
The \code{knn} method is recommended for smaller datasets.
}
\note{
\code{trans} columns and \code{cat_var} must be of the same type.
The mapping table must include a candidate for every category in the target
period. Observations without a matched candidate are dropped; add a
\code{c(NA, NA)} row to \code{trans} to retain them as \code{NA}.
}
\examples{
\dontrun{
data("occup_small", package = "cat2cat")
data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_old <- occup_small[occup_small$year == 2008, ]
occup_new <- occup_small[occup_small$year == 2010, ]

# Adding the dummy level to the mapping table for levels without a candidate
# The best to fill them manually with proper candidates, if possible
# In this case it is only needed for forward mapping, to suppress warnings
trans2 <- rbind(
  trans,
  data.frame(
    old = "no_cat",
    new = setdiff(c(occup_new$code), trans$new)
  )
)

# default only simple frequencies
occup_simple <- cat2cat(
  data = list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  ),
  mappings = list(trans = trans2, direction = "forward")
)

mappings <- list(trans = trans, direction = "backward")

ml_setup <- list(
    data = occup_small[occup_small$year >= 2010, ],
    cat_var = "code",
    method = "knn",
    features = c("age", "sex", "edu", "exp", "parttime", "salary"),
    args = list(k = 10),
    # defaults for failed ML weights
    on_fail = "freq",
    fail_warn = TRUE
)

# ml model performance check
print(cat2cat_ml_run(mappings, ml_setup))

# additional probabilities from knn
occup_ml <- cat2cat(
  data = list(
    old = occup_old, new = occup_new, cat_var = "code", time_var = "year"
  ),
  mappings = mappings,
  ml = ml_setup
)

# strict mode: stop when any ML weights cannot be produced
ml_setup_strict <- ml_setup
ml_setup_strict$on_fail <- "error"

# diagnostic mode: keep failed ML weights as NA and silence warnings
ml_setup_diag <- ml_setup
ml_setup_diag$on_fail <- "na"
ml_setup_diag$fail_warn <- FALSE
}

}
\seealso{
\itemize{
  \item \code{\link{cat2cat_ml_run}} - validate ML performance before use
  \item \code{\link{prune_c2c}}, \code{\link{cross_c2c}} - post-processing
  \item \code{\link{summary_c2c}}, \code{\link{dummy_c2c}} - helpers
  \item \code{\link{cat2cat_agg}} - for pre-aggregated data
  \item \code{vignette("cat2cat")} - identification assumptions and worked examples
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat2cat_ml.R
\name{cat2cat_ml_run}
\alias{cat2cat_ml_run}
\alias{print.cat2cat_ml_run}
\title{Cross-validation diagnostics for cat2cat ML models}
\usage{
cat2cat_ml_run(mappings, ml, ...)

\method{print}{cat2cat_ml_run}(x, ...)
}
\arguments{
\item{mappings}{`named list` with 3 fields `trans`, `direction` and
optional `freqs_df`.}

\item{ml}{`named list` (optional) with fields
`data`, `cat_var`, `method`, `features` and optional `args`,
`on_fail`, `fail_warn`.}

\item{...}{other arguments (currently unused).}

\item{x}{\code{cat2cat_ml_run} instance created with
\code{\link{cat2cat_ml_run}}.}
}
\value{
An object of class \code{"cat2cat_ml_run"} (a named list). Each
  element corresponds to one mapping group and contains:
  \describe{
    \item{\code{naive}}{\code{numeric(1)} --- random-guess baseline accuracy
      (1/k where k is number of categories). Theoretical lower bound.}
    \item{\code{freq}}{\code{numeric(1)} --- most-frequent-category baseline
      accuracy. A simple but often strong baseline.}
    \item{\code{acc}}{Named \code{numeric} vector --- test-set accuracy for
      each ML method. Higher is better; compare to \code{freq}.}
    \item{\code{brier}}{Named \code{numeric} vector --- Brier score
      for each ML method, computed from the full probability vector.
      Lower is better; range is [0, 1].}
    \item{\code{mean_prob}}{Named \code{numeric} vector --- average probability
      assigned to the true class. Higher is better. This directly measures
      the quality of probability weights used by cat2cat.}
    \item{\code{naive_brier}}{\code{numeric(1)} --- Brier score for uniform
      baseline (= (1 - 1/k) / 2). Serves as a calibration reference.}
    \item{\code{naive_mean_prob}}{\code{numeric(1)} --- mean P(true) for
      uniform baseline (= 1/k). Equals \code{naive} by definition.}
    \item{\code{freq_brier}}{\code{numeric(1)} --- Brier score using training
      set category frequencies as probabilities.}
    \item{\code{freq_mean_prob}}{\code{numeric(1)} --- mean P(true) using
      training set category frequencies.}
  }
  The object also carries an \code{"ml_models"} attribute listing the
  methods evaluated. Use \code{print()} for a human-readable summary.

\code{print} returns \code{x} invisibly, after printing a summary
  with average accuracy per method, baseline comparisons, and failure rates.
}
\description{
Evaluates whether machine-learning models used in the \code{ml} argument of
\code{\link{cat2cat}} actually improve category assignment over simpler
baselines. The function runs a per-group train/test split across every
mapping group defined by the transition table.
}
\details{
For each mapping group (set of candidate categories linked by the transition
table) the function:
\enumerate{
  \item Collects all observations from \code{ml$data} whose category belongs
        to the group.
  \item Randomly splits them into training (\code{1 - test_prop}) and test
        (\code{test_prop}) sets.
  \item Computes two baselines on the test set:
        \itemize{
          \item \strong{naive}: accuracy of a random guess (\eqn{1 / k}
                where \eqn{k} is the number of candidate categories).
          \item \strong{freq}: accuracy of always predicting the most
                frequent category in the training set.
        }
  \item Trains each ML model specified in \code{ml$method} on the training
        set and records its classification accuracy on the test set.
}

Groups with fewer than 5 observations or only one candidate category are
skipped (their accuracy is recorded as \code{NA}).

\subsection{Baseline-Only Diagnostics}{
To inspect only baseline diagnostics (\code{naive}, \code{freq}, and their
Brier/mean-probability variants), pass empty model and feature vectors:
\code{ml$method = character(0)} and \code{ml$features = character(0)}.
In this mode, no ML models are trained, but baseline diagnostics are still
computed for each mapping group.
}

\subsection{Understanding the Metrics}{
Three complementary metrics evaluate model quality:

\strong{Accuracy} measures how often the model's top prediction matches
the true category. Use this when you only care about the single most likely
assignment. Higher is better; theoretical maximum is 1.0.

\strong{Mean P(true class)} measures the average probability the model
assigns to the correct category. This evaluates the full probability
distribution, not just the top prediction. For cat2cat, where weights ARE
probabilities distributed across candidates, this metric directly measures
weight quality. Higher is better; range is \eqn{[0, 1]}.

\strong{Brier score} measures the squared error between predicted
probabilities and the one-hot encoded true outcome, normalized to
\eqn{[0, 1]}. Unlike log-loss, Brier score is bounded and does not explode
when P(true) is near zero. Lower is better; 0 means perfect prediction. For k
categories, the naive baseline (uniform 1/k) gives Brier = \eqn{(1 - 1/k) / 2}.
}

\subsection{Choosing a Method}{
\itemize{
  \item If accuracy and mean P(true) are similar across methods, prefer
        simpler methods (freq, lda) over complex ones (rf, knn).
  \item If ML methods rarely beat the frequency baseline, use
        \code{wei_freq_c2c} --- ML adds complexity without benefit.
  \item If Brier score for ML is similar to or worse than naive, the model
        is not well-calibrated. Consider \code{wei_freq_c2c} or a different ML method.
  \item Use \code{\link{cross_c2c}} to ensemble multiple methods if no
        single method dominates.
}
}

Because the split is random, results will vary between runs. For more
stable estimates, call the function several times or use a larger
\code{ml$data} set (e.g. pool multiple survey waves).
}
\examples{
\donttest{
library("cat2cat")
data("occup", package = "cat2cat")
data("trans", package = "cat2cat")

occup_2006 <- occup[occup$year == 2006, ]
occup_2008 <- occup[occup$year == 2008, ]

# Forward direction: old encoding -> new encoding
# Use data from OLD encoding periods (2006, 2008)
ml_setup <- list(
  data = rbind(occup_2006, occup_2008),
  cat_var = "code",
  method = c("knn", "rf", "lda", "nb"),
  features = c("age", "sex", "edu", "exp", "parttime", "salary"),
  args = list(k = 10, ntree = 50)
)
mappings <- list(trans = trans, direction = "forward")

set.seed(1234)
res <- cat2cat_ml_run(mappings, ml_setup, test_prop = 0.2)
print(res)

# Typical good results show:
# - ML accuracy > freq baseline (ML adds value)
# - ML Brier < naive (well-calibrated probabilities)
# - ML mean P(true) > freq (better probability weights)
#
# If Brier(ML) >= Brier(naive), the model is poorly calibrated
# and wei_freq_c2c may be safer. Use cross_c2c() to ensemble.

# High failure rate is normal - most groups have <5 observations

# Baseline-only diagnostics (no ML models):
ml_baseline <- list(
  data = rbind(occup_2006, occup_2008),
  cat_var = "code",
  method = character(0),
  features = character(0)
)
baseline_cv <- cat2cat_ml_run(mappings, ml_baseline)
print(baseline_cv)
}

}
\seealso{
\code{\link{cat2cat}} for the main mapping function,
  \code{\link{cross_c2c}} for ensembling weights from multiple methods.
}
